/* iris_gl.c
   Copyright (C) 1992, 1993, 1994, 1995, 1997, 1999  J Dana Eckart
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with CELLULAR; see the file COPYING.  If not, write to the 
   Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>

#if IRIS_GL

#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <gl/gl.h>
#include <fmclient.h>
#include <gl/device.h>
#include "view.h"

static int BIG_ENDIAN;

/* Font (and its height) for displaying the time in the time window. */
#define FONT  		0
#define FONT_HEIGHT  	17

/* Constants describing screen/monitor size. */
#define MAX_X_PIXELS	1280
#define MAX_Y_PIXELS	1024

static struct {
	unsigned int red, green, blue;
	unsigned long int composite;
} pixels[MAX_MAP_SIZE];

static unsigned long int Black_Pixel, White_Pixel;

static unsigned long int *cell_pixmap;

static long CurrentWindow = NULL;

static char *iris_gl_title;

/* Variables for controlling the main viewing window. */
static long mainWindow = NULL;
static int			cell_side_x = MAX_CELL_SIDE;
static int			cell_side_y = MAX_CELL_SIDE;
static float			cell_side_ratio = 1.0;
static int 			time_window_height = 0;
static int			mainWidth = 0, mainHeight = 0;

/* Variables for controlling the neighborhood sub-window. */
static long     neighborWindow = NULL;
static int 	neighborhood = 1;		/* size of the neighborhood. */
static 		int neighbor_x, neighbor_y;	/* x, y neighborhood viewing 
								co-ordinates. */
static 		int neighborWidth = 0, neighborHeight = 0;

#if COMBINED

/* Variables for controlling the neighborhood sub-window. */
static long     examineWindow = NULL;
static int examineWidth = 0, examineHeight = 0;

#endif

/* Clears the buffer of the currently active window. */
static void clear_buffer(void) {
	cpack(Black_Pixel);
	clear();
}

#if COMBINED

/* Display the values of the cell fields. */
static void display_cell_field_values(void) {
	int i;
	char coordinate_string[MAX_STRING_SIZE];

	winset(examineWindow);

	/* Write the cooridnates. */
	cpack(White_Pixel);
	rectf(0, 0, examineWidth, time_window_height-1);

	if (range_dim_given > 1)
		sprintf(coordinate_string, "< %d, %d >",
			neighbor_x + lower[range_dim_1_index],
			neighbor_y + lower[range_dim_2_index]);
	else
		sprintf(coordinate_string, "< %d >",
			neighbor_x + lower[range_dim_1_index]);

	cpack(Black_Pixel);
	cmov2i(time_window_height/2, (int) (0.4 * time_window_height));
	charstr(coordinate_string);

	/* Clear the contents of the window. */
	cpack(Black_Pixel);
	rectf(0, time_window_height, examineWidth, examineHeight);

	cpack(White_Pixel);

	/* Draw the horizontal lines. */
	for (i = field_count; i > 0; i--) {
		move2i(0, i*FONT_HEIGHT*2 + time_window_height);
		draw2i(examineWidth, i*FONT_HEIGHT*2 + time_window_height);
	}

	/* Display the cell field values. */
	for (i = 0; i < field_count; i++) {
		char string[MAX_STRING_SIZE];
		char field_name[MAX_STRING_SIZE];
		char field_name_value[MAX_STRING_SIZE];

		{
			int x, y, z;

			if (range[0] == 1) x = lower[0];
			else x = neighbor_x;

			if (range[1] == 1) y = lower[1];
			else if (range[0] == 1) y = neighbor_x;
			else y = neighbor_y;

			if (range[2] == 1) z = lower[2];
			else z = neighbor_y;

			if (use_floats)
				sprintf(string, "%f",
					cell_field_float_value(i, x, y, z));
			else
				get_string_for_value(string,
						     cell_field_value(i,
								      x, y, z),
						     false);
		}
		strncpy(field_name, &field_names[i*field_name_length],
			field_name_length);
		field_name[field_name_length] = '\0';
		sprintf(field_name_value, "%s: %s", field_name, string);

		cmov2i(FONT_HEIGHT*2/3, examineHeight - (i*2+1)*FONT_HEIGHT);
		charstr(field_name_value);
	}
	winset(mainWindow);
}

/* Create the cell field examination window. */
static void CreateExamineWindow(void) {

	examineWidth = FONT_HEIGHT * (field_name_length + MAX_VALUE_WIDTH + 6)/2;
	examineHeight = FONT_HEIGHT * field_count * 2;

	examineHeight += time_window_height;	/* For the coordinates. */

	prefsize(examineWidth, examineHeight);
	examineWindow = winopen(iris_gl_title);
	winconstraints();
	winconstraints();
	RGBmode();
	gconfig();

	/* Indicate which events to queue. */
	qdevice(REDRAW);

	font(FONT);

	clear_buffer();

	winset(mainWindow);

	if (!examineWindow) {
		error("Unable to open a cell field examination window",
		      (char*) NULL);
		return;
	}
	else
		display_cell_field_values();
}
#endif

/* Display the values of the appropriate neighborhood. */
static void display_neighbor_values(void) {
	int i, j;
	char coordinate_string[MAX_STRING_SIZE];

	winset(neighborWindow);
	clear_buffer();

	/* Write the cooridnates. */
	cpack(White_Pixel);
	rectf(0, 0, neighborWidth, time_window_height-1);

	if (range_dim_given > 1)
		sprintf(coordinate_string, "< %d, %d >",
			neighbor_x + lower[range_dim_1_index],
			neighbor_y + lower[range_dim_2_index]);
	else
		sprintf(coordinate_string, "< %d >",
			neighbor_x + lower[range_dim_1_index]);

	cpack(Black_Pixel);
	cmov2i(time_window_height/2, (int) (0.4 * time_window_height));
	charstr(coordinate_string);

	/* Clear the contents of the window. */
	cpack(Black_Pixel);
	rectf(0, time_window_height, neighborWidth, neighborHeight);

	cpack(White_Pixel);

	/* Draw the horizontal lines (if need be). */
	if (range_dim_given > 1) for (i = 2*neighborhood; i > 0; i--) {
		move2i(0, i*FONT_HEIGHT*MAX_FIELD_VALUE_WIDTH + time_window_height);
		draw2i(neighborWidth, i*FONT_HEIGHT*MAX_FIELD_VALUE_WIDTH + time_window_height);
	}

	/* Draw the vertical lines. */
	for (i = 2*neighborhood; i > 0; i--) {
		move2i(i*FONT_HEIGHT*MAX_FIELD_VALUE_WIDTH, time_window_height);
		draw2i(i*FONT_HEIGHT*MAX_FIELD_VALUE_WIDTH, neighborHeight);
	}

	/* Display the neighbor values. */
	if (range_dim_given == 1)
		for (i = -neighborhood; i <= neighborhood; i++) {
			int x = (neighbor_x + i + range[range_dim_1_index]) %
						range[range_dim_1_index];
			char string[MAX_STRING_SIZE];

			if (use_floats)
				sprintf(string, "%f", cell_float_vals[x]);
			else
				get_string_for_value(string,
						     cell_vals[x], true);

			cmov2i((neighborhood+i)*FONT_HEIGHT*MAX_FIELD_VALUE_WIDTH +
					FONT_HEIGHT*0.4,
				MAX_FIELD_VALUE_WIDTH*FONT_HEIGHT*0.4 +
					time_window_height);
			cpack(pixels[cell_map_vals[x]].composite);
			charstr(string);
	}
	for (i = -neighborhood; i <= neighborhood; i++)
		for (j = -neighborhood; j <= neighborhood; j++) {
			int x = (neighbor_x + i + range[range_dim_1_index]) %
						range[range_dim_1_index];
			int y = (neighbor_y + j + range[range_dim_2_index]) %
						range[range_dim_2_index];
			char string[MAX_STRING_SIZE];
			int index = x * range[range_dim_2_index] + y;

			if (use_floats)
				sprintf(string, "%f", cell_float_vals[index]);
			else
				get_string_for_value(string,
						     cell_vals[index], true);

			cmov2i((neighborhood+i)*FONT_HEIGHT*MAX_FIELD_VALUE_WIDTH +
					FONT_HEIGHT*0.4,
		    	       (neighborhood+j)*FONT_HEIGHT*MAX_FIELD_VALUE_WIDTH +
					MAX_FIELD_VALUE_WIDTH*FONT_HEIGHT*0.4 +
					time_window_height);
			cpack(pixels[cell_map_vals[x*range[range_dim_2_index]+y]].composite);
			charstr(string);
		}
	winset(mainWindow);
}

/* Create the sub-window dedicated to showing the numeric values of a
   neighborhood.
*/
static void CreateNeighborWindow(void) {

	neighborWidth = (2*neighborhood+1) * FONT_HEIGHT * MAX_FIELD_VALUE_WIDTH;
	neighborHeight = (range_dim_given > 1 ? neighborWidth :
					        FONT_HEIGHT*MAX_FIELD_VALUE_WIDTH);

	neighborHeight += time_window_height;	/* For the coordinates. */

	prefsize(neighborWidth, neighborHeight);
	neighborWindow = winopen(iris_gl_title);
	winconstraints();
	winconstraints();
	RGBmode();
	gconfig();

	/* Indicate which events to queue. */
	qdevice(REDRAW);
        qdevice(KEYBD);

	font(FONT);

	clear_buffer();

	winset(mainWindow);

	if (!neighborWindow) {
		error("Unable to open a neighborhood window", (char*) NULL);
		return;
	}
	else
		display_neighbor_values();
}

/* Resize the neighborhood sub-window. */
static void resize_neighborhood(void) {
	long int x_origin, y_origin;

	winset(neighborWindow);

	neighborWidth = (2*neighborhood+1) * FONT_HEIGHT * MAX_FIELD_VALUE_WIDTH;
	neighborHeight = (range_dim_given > 1 ? neighborWidth :
					        FONT_HEIGHT*MAX_FIELD_VALUE_WIDTH);

	neighborHeight += time_window_height;	/* For the coordinates. */

	getorigin(&x_origin, &y_origin);
	winposition(x_origin, x_origin + neighborWidth, 
		    y_origin, y_origin + neighborHeight);
	ortho2(0, neighborWidth, 0, neighborHeight);
	reshapeviewport();

	clear_buffer();

	winset(mainWindow);
}

/* Do any cleanup for iris_gl windows. */
void finish_iris_gl(void) {
	if (mainWindow) winclose(mainWindow);
	if (neighborWindow) winclose(neighborWindow);
}

/* Display the current time step at the bottom of the window. */
static void display_time(void) {
	char time_string[MAX_STRING_SIZE];

	/* Clear the time display part of the window. */
	cpack(White_Pixel);
	rectf(0, 0, mainWidth, time_window_height-1);

	/* Display the new current time step. */
	sprintf(time_string, "time = %d", _time_step);
	cpack(Black_Pixel);
	cmov2i(time_window_height/2, (int) (0.4 * time_window_height));
	charstr(time_string);
}

/* Display all the cell_vals in the universe and update the neighborhood
   window if it is being used.
*/
void display_iris_gl_cells(void) {
	rectzoom(cell_side_x, cell_side_y);
	lrectwrite(0, time_window_height,
		   range[range_dim_1_index]-1, 
		   range[range_dim_2_index]-1 + time_window_height,
		   cell_pixmap);

        /* May need to update other windows too. */
        if (neighborWindow) display_neighbor_values();
#if COMBINED
        if (examineWindow) display_cell_field_values();
#endif
}

/* Update (display) the color of a single cell. */
void update_iris_gl(long int x, long int y, celltype color_index) {
        long int i = x - lower[range_dim_1_index];
        long int j = y - lower[range_dim_2_index];

	/* The index to cell_pixmap seems to be "backwards", but it works! */
	cell_pixmap[j*range[range_dim_1_index]+i] = 
						pixels[color_index].composite;
}

/* Redisplay the cell "colors" in the window. */
static void redisplay_cells(void) {
	register long int i, j;

	for (i = 0; i < range[range_dim_1_index]; i++)
		for (j = 0; j < range[range_dim_2_index]; j++)
			update_iris_gl(i, j, 
				cell_map_vals[i*range[range_dim_2_index]+j]);

	display_iris_gl_cells();
}

/* Handle the events for the neighborhood sub-window. */
static void sub_window_event(short int device, short int value) {

	switch(device) {
	case REDRAW:
		display_neighbor_values();
       		break;

	case KEYBD:
               	switch (value) {
                case '0':
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
			neighborhood = value-'0';
			resize_neighborhood();
			display_neighbor_values();
			break;

		default:
			break;
       	       	}
       	       	break;

	default:
		break;
	}
}

/* Handle the events for the main window. */
static void main_window_event(int *stopped, short int device, short int value) {

	switch(device) {
       	case REDRAW:
	{
		long int x_origin, y_origin;
		long int width, height;

		/* Don't resize the window if both the width and
		   the height are the same.
		*/
		getsize(&width, &height);
		if (width == mainWidth && height == mainHeight) {
			redisplay_cells();
			display_time();
			break;
		}

		/* The window needs to be resized. */
		height -= time_window_height;
		cell_side_x = (width / range[range_dim_1_index] > 
				height / range[range_dim_2_index] ?
				height / range[range_dim_2_index] : 
				width / range[range_dim_1_index]);

		if (cell_side_ratio < 1) {
			if (cell_side_x > MAX_CELL_SIDE) 
				cell_side_x = MAX_CELL_SIDE;

			/* Make sure cell_vals are "visible". */
			if (cell_side_x <= 0) cell_side_x = 1;

			cell_side_y = cell_side_x * cell_side_ratio;
		}
		else {
			cell_side_y = cell_side_x * cell_side_ratio;

			if (cell_side_y > MAX_CELL_SIDE) 
				cell_side_y = MAX_CELL_SIDE;

			/* Make sure cell_vals are "visible". */
			if (cell_side_y <= 0) cell_side_y = 1;

			cell_side_x = cell_side_y / cell_side_ratio;
		}

		/* Since the size just chosen for the window may
		   not yield an integer number of pixels for the
		   sides of cells, reset the window size to just
		   fit the new cell_side_x and cell_side_y values.
		*/
		mainWidth = range[range_dim_1_index]*cell_side_x-1;
		mainHeight = range[range_dim_2_index]*cell_side_y-1 + 
			      	time_window_height;

		getorigin(&x_origin, &y_origin);
		winposition(x_origin, x_origin + mainWidth, 
			    y_origin, y_origin + mainHeight);
		ortho2(0, mainWidth, 0, mainHeight);
		reshapeviewport();

		clear_buffer();

		redisplay_cells();
		display_time();
       		break;
	}

	case LEFTMOUSE:
	case MIDDLEMOUSE:
	case RIGHTMOUSE:
#if COMBINED
		if (neighborWindow || examineWindow) {
#else
		if (neighborWindow) {
#endif
			long int mouse_x = getvaluator(MOUSEX);
			long int mouse_y = getvaluator(MOUSEY);
			long int origin_x, origin_y;
			int old_x = neighbor_x;
			int old_y = neighbor_y;

			getorigin(&origin_x, &origin_y);
			mouse_x -= origin_x;
			mouse_y -= origin_y;
			mouse_y -= time_window_height;

			neighbor_x = (mouse_x/cell_side_x + 
					range[range_dim_1_index]) %
						range[range_dim_1_index];

			if (range_dim_given > 1)
				neighbor_y = (mouse_y/cell_side_y + 
						range[range_dim_2_index]) %
						range[range_dim_2_index];

			if (neighbor_x < 0 ||
				neighbor_x >= range[range_dim_1_index] ||
				(range_dim_given > 1 &&
					(neighbor_y < 0 || 
                                	 neighbor_y >= range[range_dim_2_index]))) {
				neighbor_x = old_x;
				neighbor_y = old_y;
			}
		}
#if COMBINED
		if (examineWindow) display_cell_field_values();
#endif
		if (neighborWindow) display_neighbor_values();
		break;

	case KEYBD:
               	switch (value) {
#if COMBINED
              	case 'e':
              	case 'E':
			/* Cell field examination window */
			if (field_count == 1)
				;
			else if (!examineWindow) {
				CreateExamineWindow();
				display_cell_field_values();
			} else {
				winclose(examineWindow);
				examineWindow = NULL;
			}
			break;
#endif
              	case 'n':
              	case 'N':
			/* Neighborhood window */
			if (!neighborWindow) {
				CreateNeighborWindow();
				display_neighbor_values();
			} else {
				winclose(neighborWindow);
				neighborWindow = NULL;
			}
			break;
               	case 'q':
               	case 'Q':
			/* Quit */
			exit_signaled(0);
               	case 's':
               	case 'S':
			/* Stop */
                       	*stopped = true;
			break;
               	case 'c':
               	case 'C':
			/* Continue */
                       	*stopped = false;
			break;
		default:
			break;
                }
       	       	break;

	default:
		break;
	}
}

/* Check for, and perform events as they arise. */
static void do_event(void) {
	boolean stopped = false;

	/* If there are events, handled them. */
	while (qtest() || stopped) {
		short int value;
		short int device = qread(&value);

		if (device == INPUTCHANGE)
			CurrentWindow = (long) value;
		else if (CurrentWindow)
			if (CurrentWindow == neighborWindow)
				sub_window_event(device, value);
#if COMBINED
			else if (CurrentWindow == examineWindow)
				/* Assume the event is a REDRAW. */
				display_cell_field_values();
#endif
			else if (CurrentWindow == mainWindow)
				main_window_event(&stopped, device, value);
	}
}

/* Stores the red, green and blue portions that make up a color. */
celltype setcolor_iris_gl(unsigned int red,
			  unsigned int green,
			  unsigned int blue,
			  unsigned int alpha,
			  int index) {
	pixels[index].red = (unsigned int) (red*255/65535);
	unsigned char new_alpha = (unsigned char) (alpha*255/65535);

	if (new_alpha < 255) do_alpha = true;
	pixels[index].green = (unsigned int) (green*255/65535);
	pixels[index].blue = (unsigned int) (blue*255/65535);

	if (BIG_ENDIAN)
		pixels[index].composite = (unsigned long int) (
					new_alpha |
					(pixels[index].blue << 8) |
					(pixels[index].green << 16) |
					(pixels[index].red << 24)
					);
	else
		pixels[index].composite = (unsigned long int) (
					(new_alpha << 24) |
					(pixels[index].blue << 16) |
					(pixels[index].green << 8) |
					pixels[index].red
					);
	return index;
}

/* Set the values for Black_Pixel and White_Pixel. */
static void Black_and_White_Pixels(int max_colors) {
	if (max_colors <= 2) {
		White_Pixel = 0xFFFFFF;
		Black_Pixel = 0;
	}
	else {
		/* Assume that the darkest pixels are given first and
		   the lightest in color are given last.  In either
		   case, both should be sufficiently different in color
		   to be readable.
		*/
		White_Pixel = pixels[max_colors-1].composite;
		Black_Pixel = pixels[0].composite;
	}
}

/* Do all the setup and housekeeping for starting an iris gl window. */
void setup_iris_gl(char *map_name, 
			int first_cell_side_x, int first_cell_side_y,
			char *title) {
	int max_x_pixels, max_y_pixels, ncolors;
	char *default_file;

	BIG_ENDIAN = big_endian();

	if (BIG_ENDIAN < 0)
		error("Unknown architecture (neither big or little endian)",
			(char*) NULL);

	iris_gl_title = title;

	time_window_height = 1.5*FONT_HEIGHT;

	/* Get the height and width of the screen in pixels. */
	max_x_pixels = MAX_X_PIXELS;
	max_y_pixels = MAX_Y_PIXELS - time_window_height;

	cell_side_ratio = (float) first_cell_side_y / (float) first_cell_side_x;

	/* Determine the size of the sides of each displayed cell. */
	cell_side_x = (max_x_pixels / range[range_dim_1_index] > 
				max_y_pixels / range[range_dim_2_index] ?
			max_y_pixels / range[range_dim_2_index] : 
			max_x_pixels / range[range_dim_1_index]);

	if (cell_side_ratio < 1) {
		if (cell_side_x > MAX_CELL_SIDE) 
			cell_side_x = MAX_CELL_SIDE;

		/* Make sure cell_vals are "visible". */
		if (cell_side_x <= 0)
			error("Too many cells to display on screen", 
				(char*) NULL);

		cell_side_y = cell_side_x * cell_side_ratio;
	}
	else {
		cell_side_y = cell_side_x * cell_side_ratio;

		if (cell_side_y > MAX_CELL_SIDE) 
			cell_side_y = MAX_CELL_SIDE;

		/* Make sure cell_vals are "visible". */
		if (cell_side_y <= 0)
			error("Too many cells to display on screen", 
				(char*) NULL);

		cell_side_x = cell_side_y / cell_side_ratio;
	}

	/* Go with the prefered size if possible. */
	if (cell_side_x > first_cell_side_x) cell_side_x = first_cell_side_x;
	if (cell_side_y > first_cell_side_y) cell_side_y = first_cell_side_y;

	mainWidth = range[range_dim_1_index]*cell_side_x-1;
	mainHeight = range[range_dim_2_index]*cell_side_y-1 + 
						time_window_height;

	/* If a map file was given, read it; otherwise use the default. */
	if (!map_name) {
		default_file = "c-map";

		map_name = (char*) malloc((unsigned int) (strlen(LIB_DIR) + 
						strlen(default_file) + 2));
		if (!map_name)
			error("Out of memory", (char*) NULL);
		sprintf(map_name, "%s%s", LIB_DIR, default_file);
	}
	ncolors = file_size(map_name);

	if (ncolors > 256*256*256)
		error("Too many colors requested", (char*) NULL);

	fprintf(stderr, "Using %d colors...\n", ncolors);
	read_map(map_name, ncolors <= 2);
	Black_and_White_Pixels(ncolors);

	/* Create the window. */
	foreground();
	prefsize(mainWidth, mainHeight);
	mainWindow = winopen(iris_gl_title);
	winconstraints();
	winconstraints();
	RGBmode();
	gconfig();

	clear_buffer();

	/* Create a pixel map. */
	cell_pixmap = (unsigned long int*)
			calloc(range[range_dim_1_index]*range[range_dim_2_index],
			       sizeof(unsigned long int));

	/* Indicate which events to queue. */
	qdevice(REDRAW);
        qdevice(LEFTMOUSE);
        qdevice(MIDDLEMOUSE);
        qdevice(RIGHTMOUSE);
        qdevice(KEYBD);

	font(FONT);

	/* Get a more accurate looking cursor. */
	{
		const unsigned short int cursor[16] = {
			0x0100,	0x0100,	0x0100,	0x0100,
			0x0100,	0x0100,	0x0100,	0x0100,
			0xffff,	0x0100,	0x0100,	0x0100,
			0x0100,	0x0100,	0x0100,	0x0100
		};
		curstype(C16X1);
		drawmode(CURSORDRAW);
		mapcolor(1, 255, 255, 255);
		defcursor(1, cursor);
		curorigin(1, 7, 9);
		setcursor(1, 0, 0);
		drawmode(NORMALDRAW);
	}

	/* Assign intial values for the cursor location to be used for
	   neighborhood value viewing within a sub-window.
	*/
	neighbor_x = range[range_dim_1_index]/2;
	if (range_dim_given > 1)
		neighbor_y = range[range_dim_2_index]/2;

	CurrentWindow = mainWindow;

	/* On an SGI, the window header bar can only get so small, and
	   this prevents the data from taking less room than the bar does.
	*/
	if (MAX_FIELD_VALUE_WIDTH == 1) MAX_FIELD_VALUE_WIDTH++;

	/* Wait for the first event. */
	while (!qtest());
}

/* Start displaying the cell_vals. */
void display_iris_gl(int field) {

	redisplay_cells();

	/* Display cell_vals for the rest of time. */
	while (1) {
		do_event();
#if !COMBINED
		if (next_time >= 0) 
#endif
		{
#if COMBINED
#if MAX_CPUS > 1
			cellang_main(output, view, filter, field, cpus-1);
#else
			cellang_main(output, view, filter, field);
#endif
#else
			read_cell_block();
#endif

#if COMBINED
                        if (1 == report_time_step ||
                            _time_step == next_report_time ||
                            _time_step == next_read_time)
#endif
			{
				display_iris_gl_cells();
				display_time();
			}

#if COMBINED
			if (_time_step == next_read_time)
				get_report_time(time_file);
			else if (report_time_step > 0 && _time_step >= next_report_time)
				next_report_time = _time_step + report_time_step;
			_time_step++;
#else
			if (next_time >= 0) _time_step = next_time;
#endif
	
			/* Pause times of 0 or less are ignored. */
			if (pause_time > 0) sleep((unsigned int) pause_time);
		}
	}
}

#endif


