/* semantic.c
   Copyright (C) 1992, 1993, 1994, 1995, 1997, 1998, 1999  J Dana Eckart
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with CELLULAR; see the file COPYING.  If not, write to the 
   Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* This file contains the semantic routines (and their respective
   support functions) for checking static semantics and generating 
   the intermediate representation (C code).

   The names of variables are preceeded by an underscore ("_") in the C 
   code for the rule set of the cellular automata.  All other variables 
   generated by the compiler (and not appearing in the rule set of the 
   input) are NOT preceeded by a "_" (underscore); thus guaranteeing no 
   name conflicts (defines are given in capitals to prevent name conflicts).

   NOTE: The comments given for semantic routines give the items that
	 the routine expects and leaves on the semantic stack.  In both
	 cases, the top of the stack is the leftmost item given.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "boolean.h"
#include "error.h"
#include "semantic.h"
#include "attr.h"
#include "symtable.h"
#include "stack.h"

FILE *file_hh = NULL;		/* Denotes the C pre-header output stream. */
FILE *file_h = NULL;		/* Denotes the C header output stream. */
FILE *file_c = NULL;		/* Denotes the C code output stream. */

/* This variable is used to communicate which of two types of numeric
   literals has just been read by the scanner.
*/
boolean NumericIsFloat;

boolean semantic_error = false;	/* Indicates if a semantic error has occured. */

/* Holds the largest relative index absolute value. */
int max_rel_index = 0;

/* Declared in scanner.c and used to implement the trace option. */
extern int linenumber;

/************************************************************************/

#define MAX_STACK_SIZE 	64	/* Size of the semantic stack. */

stack_entry stack[MAX_STACK_SIZE];	/* The semantic stack. */
long int top = -1;			/* Current top of semantic stack. */

/* SUPPORT ROUTINE
      ACTION - Increments top.  This provides checking so that pushes
	       onto the semantic stack are checked to insure that an
	       overflow does not occur.
*/
void new_entry(void) {
	if (top == MAX_STACK_SIZE - 1) {
		fprintf(stderr, "INTERNAL ERROR: Semantic stack overflow.\n");
		exit(1);
	}
	top++;
}

/************************************************************************/

/* SUPPORT ROUTINE
      ACTION - Prints an error message using the provided format
	       and input string, preceeded by the current line
	       number of the input stream.  A flag specifing that 
	       a semantic error was encountered is also set to true.
*/
void error(char *format, char *string) {
	semantic_error = true;
	error_prefix();
	if (string)
		fprintf(stderr, format, string);
	else
		fprintf(stderr, format);
	fprintf(stderr, "\n");
}

/************************************************************************/

/* Indentation to make the generated C code easier to read is 
   accomplished by using print_indent.  The amount of indentation
   is kept in the "indentation" variable, whose value is updated
   through use of the macros "indent" and "undent".
*/

int indentation = 0;

#define indent	indentation += 2
#define undent	indentation -= 2

/* SUPPORT ROUTINE
      ACTION - Prints the current amount of indentation to the 
	       specified file.
*/
void print_indent(void) {
	int i;
	for (i = 0; i < indentation; i++) fprintf(file_c, " ");
}

/************************************************************************/

/* SUPPORT ROUTINE
      ACTION - Checks the top n entries (0 checks NO entries) on 
	       top of the stack and returns true iff at least one 
	       of them is a StackError entry.
*/
boolean stack_error(int n) {
	int i;
	if (n-1 > top)
		error("COMPILER ERROR -- too few entries on semantic stack",
			(char*) NULL);
	for (i = 0; i < n; i++)
		if (stack[top-i].kind == StackError) return true;
	return false;
}

/* SUPPORT ROUTINE
      ACTION - Checks the top n entries (0 checks NO entries) on 
	       top of the stack and returns to the heap anything
	       that needs to be (general used for error processing).
	       The value of "top" is left unchanged.
*/
void stack_clear(int n) {
	int i;
	if (n-1 > top)
		error("COMPILER ERROR -- too few entries on semantic stack",
			(char*) NULL);
	for (i = 0; i < n; i++)
		if (stack[top-i].kind == StackExpr)
			myfree(stack[top-i].data.expr.addr);
		else if (stack[top-i].kind == StackToken)
			myfree(stack[top-i].data.token);
}

/************************************************************************/

/* A current count of the number of dimensions. */
int dimension_count = 0;

/* Used to count the number of fields declared for the cells of the universe. */
int cell_field_count = 0;

/* Used to determine if named fields were declared for the cells. */
boolean cell_fields_declared = false;

/* Used to count the number of fields declared for agents associated with
   each cell of the universe.
*/
int agent_field_count = 0;

/* Used to count the total number of fields (cell and agent). */
int field_count = 0;

/* Indicates the absence or presence of a delcared agent. */
boolean agent_declared = false;

/* Holds the field names given in the cell declaration, along with their
   upper and lower bounds, in the order they were declared.  Alias is true
   iff the field is part of an array.
*/
struct field_struct {
	char *name;
	range_entry range;
	boolean alias;
	boolean constant;
} field_list[MAX_SUPPORTED_FIELDS];

int longest_field_name = 0;

/* Constant_fields is set to true whenever a constant cell field is declared. */
boolean constant_fields = false;

/* Used to count the total number of cell array fields. */
int array_field_alias_count = 0;

/* Holds the array alais names given in the cell declaration, along with
   their upper and lower bounds, in the order they were declared.  The size
   indicates the number of elements in the array field.
*/
struct {
	char *name;
	range_entry range;
	unsigned int size;
	boolean constant;
} array_field_alias_list[MAX_SUPPORTED_FIELDS];

/************************************************************************/

/* Holds the guards used by do_assign in generating the if statements
   needed when compiling using microvectors.
*/
typedef struct mv_guard_fubar {
	char *true_guard;
	char *false_guard;
	struct mv_guard_fubar *next;
} mv_guard;

mv_guard *mv_guards = NULL;

/************************************************************************/

int nested_arrays = 0;

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Adds one to the count of nested array indices being computed.
*/
void start_array_index(void) {
	if (use_floats && nested_arrays)
		error("An array element cannot be used to index another array when using the -float option",
			(char*) NULL);
	nested_arrays++;
}

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Subtracts one from the count of nested array indices being
		computed.
*/
void end_array_index(void) {
	nested_arrays--;
}

/************************************************************************/

/* SUPPORT ROUTINE
      ACTION - Determines whether or note the cast type is float or not.
*/
char* cast_type(void) {
	return ((!use_floats || nested_arrays) ? "long int" : "float");
}

/* SUPPORT ROUTINE
      ACTION - Determines whether or note the cast type is float or not.
*/
char* format_type(void) {
	return (use_floats ? "%f" : "%ld");
}

/************************************************************************/

/* SUPPORT ROUTINE
      ACTION - Enter the predefined identifiers "time", "cell", and "random"
	       into the symbol table.
*/
void decl_time_and_cell_and_random(void) {
	attr_rec attribute;

	/* Enter the predefined symbol "cell". */
	attribute.assignable = true;
	attribute.is_float = false;
	attribute.field_name = VARIABLE;
	attribute.structured = cell_fields_declared;
	attribute.agent = false;
	attribute.array = false;
	attribute.index = false;
	attribute.address = mymalloc((unsigned int) (strlen("cell") + 1));
	strcpy(attribute.address, "cell");
	enterid("cell", attribute);

	/* Enter the predefined symbol "time". */
	attribute.assignable = false;
	attribute.is_float = use_floats;
	attribute.field_name = VARIABLE;
	attribute.structured = false;
	attribute.agent = false;
	attribute.array = false;
	attribute.index = false;
	attribute.address = mymalloc((unsigned int) (strlen("time_step") + 1));
	strcpy(attribute.address, "time_step");
	enterid("time", attribute);

	/* Enter the predefined symbol "random". */
	attribute.assignable = false;
	attribute.is_float = use_floats;
	attribute.field_name = VARIABLE;
	attribute.structured = false;
	attribute.agent = false;
	attribute.array = false;
	attribute.index = false;
	if (use_floats) {
		attribute.address = mymalloc((unsigned int) (strlen("myfrandom()") + 1));
		strcpy(attribute.address, "myfrandom()");
	} else {
		attribute.address = mymalloc((unsigned int) (strlen("myrandom()") + 1));
		strcpy(attribute.address, "myrandom()");
	}
	enterid("random", attribute);
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Set agent_declared to true.
*/
void set_agent(void) {
	agent_declared = true;
}

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Start the agent struct declaration.
*/
void begin_decl_agent(void) {
	/* Supporting declarations. */
	fprintf(file_c, "#include <malloc.h>\n\n");

	/* Begin the struct declaration. */
	fprintf(file_c, "struct agent_type {\n");
	indent;
}

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Writes code to declare the agent type associated with cells.
*/
void end_decl_agent(void) {
	int i;

	/* Finish the agent declaration. */
	print_indent();
	fprintf(file_c, "struct agent_type *agent;\n");
	undent;
	print_indent();
	fprintf(file_c, "};\n\n");

	agent_field_count = field_count - cell_field_count;

	/* Make sure not too many fields are used. */
	if (field_count > MAX_SUPPORTED_FIELDS) {
		error("Sorry, no more than %d total fields are supported",
		      (char*) MAX_SUPPORTED_FIELDS);
		return;
	}

	/* Write the macro definition for "agent_eq". */
	print_indent();
	fprintf(file_c, "#define agent_eq(x,y) (\\\n"); 
	indent;
	print_indent();
	for (i = cell_field_count; i < field_count; i++) {
		print_indent();
		fprintf(file_c, "((x).%s == (y).%s)", 
			field_list[i].name, field_list[i].name); 
		if (i < field_count - 1)
			fprintf(file_c, " && \\\n"); 
		else
			fprintf(file_c, ")\n\n"); 
	}
	undent;

	/* Write the macro definition for "assign_agent". */
	fprintf(file_c, "#define assign_agent(x,y) { \\\n"); 
	for (i = cell_field_count; i < field_count; i++) {
		fprintf(file_c, "\t(x).%s = (y).%s;", 
			field_list[i].name, field_list[i].name); 
		fprintf(file_c, " \\\n"); 
	}
	fprintf(file_c, "}\n\n"); 

	/* Write the macro definition for "create_agent". */
	fprintf(file_c, "#define create_agent("); 
	for (i = cell_field_count; i < field_count; i++) {
		fprintf(file_c, "a_%d", i - cell_field_count); 
		if (i < field_count - 1) fprintf(file_c, ","); 
	}
	fprintf(file_c, ") { \\\n"); 
	fprintf(file_c, "\ttmp_agent = (struct agent_type*) malloc(sizeof(struct agent_type)); \\\n"); 
	for (i = cell_field_count; i < field_count; i++) {
		fprintf(file_c, "\ttmp_agent->%s = a_%d", 
			field_list[i].name, i - cell_field_count); 
		fprintf(file_c, "; \\\n"); 
	}
	fprintf(file_c, "}\n\n"); 
}

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Since a range was not given, make sure that at least one
		field was declared.
*/
void non_empty(void) {
	if (field_count < 1)
		error("No field(s) were given in the cell declaration",
			(char*) NULL);
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr
      LEAVES  - n/a
      ACTION  - Set the number of dimensions in the cellular automata 
		universe to that indicated by the top of the stack.
*/
void set_dimensions(void) {
	if (stack_error(1))
		top--;
	else {
		dimension_count = atoi(stack[top--].data.expr.addr);

		/* Make sure not too many dimensions are used. */
		if (dimension_count > MAX_SUPPORTED_DIMENSIONS)
			error("Sorry, no more than %d dimensions are supported",
			      (char*) MAX_SUPPORTED_DIMENSIONS);

		/* Check the validity of the trace option, if given. */
		if (trace_count > 0 && trace_count != dimension_count)
			error("Trace option doesn't match the number of dimensions specified by the program",
			      (char*) NULL);

		/* Check the validity of the size option, if given. */
		if (dims_expected > 1 && dims_expected != dimension_count)
			error("Size option doesn't match the number of dimensions specified by the program",
			      (char*) NULL);
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Start the cell type declaration.
*/
void begin_decl_cells(void) {
	field_count = 0;
}

/* This variable is true iff a constant cell field is in the process of
   being declared.
*/
boolean const_field = false;

/* SUPPORT ROUTINE
      ACTION - Generates the definition for cell field copying.
*/
void gen_cell_copy(char *name, boolean copy_all, boolean copy_const) {
	int i;

	fprintf(file_c, "#define %s(X,Y) ", name); 
	if (cell_field_count > 0) {
		fprintf(file_c, "{ \\\n"); 
		for (i = 0; i < cell_field_count; i++) {
			if (!copy_all && field_list[i].constant != copy_const)
				continue;
			if (field_list[i].alias) {
				char tmp[MAX_STRING_LENGTH], *name, *index;
				strcpy(tmp, field_list[i].name);
				name = strtok(tmp, "[");
				index = strtok((char*) NULL, "]");
				fprintf(file_c,
					"\t%s_##X[%s] = %s_##Y[%s]",
					name, index, name, index);
			} else
				fprintf(file_c, "\t%s_##X = %s_##Y",
					field_list[i].name,
					field_list[i].name);
			fprintf(file_c, "; \\\n");
		}
		fprintf(file_c, "}\n\n");
	} else
		fprintf(file_c, ";\n\n");
}

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Writes code to declare the cell type of the cells which
		make up the cellular automata universe.
*/
void end_decl_cells(void) {
	int i;

	/* Check to insure that if a field was given on the command
	   line, that it is compatible with the program.
	*/
	if (field_given && (field_given > field_count && !agent_declared))
		error("Unknown field, '%d', given on command line",
			(char*) field_given);

	cell_fields_declared = field_count > 0;
	cell_field_count = field_count;

	/* Make sure not too many fields are used. */
	if (cell_field_count > MAX_SUPPORTED_FIELDS) {
		error("Sorry, no more than %d fields are supported",
		      (char*) MAX_SUPPORTED_FIELDS);
		return;
	}

	/* Write the macro definition for "neq". */
	fprintf(file_c, "#define neq(X,Y) ("); 
	if (cell_field_count > 0) {
		fprintf(file_c, "\\\n"); 
		for (i = 0; i < cell_field_count; i++) {
			if (field_list[i].alias) {
				char tmp[MAX_STRING_LENGTH], *name, *index;
				strcpy(tmp, field_list[i].name);
				name = strtok(tmp, "[");
				index = strtok((char*) NULL, "]");
				if (mv_bits > 0 && !mv_signed)
					fprintf(file_c,
						"\tUV_NEQUAL(%s_##X[%s],%s_##Y[%s],%d)",
						name, index, name, index, mv_bits);
				else if (mv_bits > 0 && mv_signed)
					fprintf(file_c,
						"\tUVO_NEQUAL(%s_##X[%s],%s_##Y[%s],%d)",
						name, index, name, index, mv_bits);
				else
					fprintf(file_c,
						"\t(%s_##X[%s] != %s_##Y[%s])",
						name, index, name, index);
			} else
				if (mv_bits > 0 && !mv_signed)
					fprintf(file_c,
						"\tUV_NEQUAL(%s_##X,%s_##Y,%d)",
						field_list[i].name,
						field_list[i].name, mv_bits);
				else if (mv_bits > 0 && mv_signed)
					fprintf(file_c,
						"\tUVO_NEQUAL(%s_##X,%s_##Y,%d)",
						field_list[i].name,
						field_list[i].name, mv_bits);
				else
					fprintf(file_c,
						"\t(%s_##X != %s_##Y)",
						field_list[i].name,
						field_list[i].name);
			if (i < cell_field_count - 1)
				if (mv_bits > 0)
					fprintf(file_c, "|");
				else
					fprintf(file_c, "||");
			fprintf(file_c, " \\\n");
		}
		fprintf(file_c, ")\n\n");
	} else
		fprintf(file_c, "0)\n\n"); 

	/* Write the macro definition for "changed".   This macro is just like
	   "neq" except that constant cell fields are not compared.  this
	   macro is only used in determining whether or not to output/display
	   the results of a cell.
	*/
	fprintf(file_c, "#define changed(X,Y) (0 "); 
	if (cell_field_count > 0) {
		fprintf(file_c, "\\\n"); 
		for (i = 0; i < cell_field_count; i++) {
			if (field_list[i].constant) continue;

			if (mv_bits > 0)
				fprintf(file_c, "\t| ");
			else
				fprintf(file_c, "\t|| ");

			if (field_list[i].alias) {
				char tmp[MAX_STRING_LENGTH], *name, *index;
				strcpy(tmp, field_list[i].name);
				name = strtok(tmp, "[");
				index = strtok((char*) NULL, "]");
				if (mv_bits > 0 && !mv_signed)
					fprintf(file_c,
						"UV_NEQUAL(%s_##X[%s],%s_##Y[%s],%d)",
						name, index, name, index, mv_bits);
				else if (mv_bits > 0 && mv_signed)
					fprintf(file_c,
						"UVO_NEQUAL(%s_##X[%s],%s_##Y[%s],%d)",
						name, index, name, index, mv_bits);
				else
					fprintf(file_c,
						"(%s_##X[%s] != %s_##Y[%s])",
						name, index, name, index);
			} else
				if (mv_bits > 0 && !mv_signed)
					fprintf(file_c,
						"UV_NEQUAL(%s_##X,%s_##Y,%d)",
						field_list[i].name,
						field_list[i].name, mv_bits);
				else if (mv_bits > 0 && mv_signed)
					fprintf(file_c,
						"UVO_NEQUAL(%s_##X,%s_##Y,%d)",
						field_list[i].name,
						field_list[i].name, mv_bits);
				else
					fprintf(file_c,
						"(%s_##X != %s_##Y)",
						field_list[i].name,
						field_list[i].name);
			fprintf(file_c, " \\\n");
		}
	}
	fprintf(file_c, ")\n\n");

	/* Write the macro definition for "eq". */
	fprintf(file_c, "#define eq(X,Y) ("); 
	if (cell_field_count > 0) {
		fprintf(file_c, "\\\n"); 
		for (i = 0; i < cell_field_count; i++) {
			if (field_list[i].alias) {
				char tmp[MAX_STRING_LENGTH], *name, *index;
				strcpy(tmp, field_list[i].name);
				name = strtok(tmp, "[");
				index = strtok((char*) NULL, "]");
				if (mv_bits > 0 && !mv_signed)
					fprintf(file_c,
						"\tUV_EQUAL(%s_##X[%s],%s_##Y[%s],%d)",
						name, index, name, index, mv_bits);
				else if (mv_bits > 0 && mv_signed)
					fprintf(file_c,
						"\tUVO_EQUAL(%s_##X[%s],%s_##Y[%s],%d)",
						name, index, name, index, mv_bits);
				else
					fprintf(file_c,
						"\t(%s_##X[%s] != %s_##Y[%s])",
						name, index, name, index);
			} else
				if (mv_bits > 0 && !mv_signed)
					fprintf(file_c,
						"\tUV_EQUAL(%s_##X,%s_##Y,%d)",
						field_list[i].name,
						field_list[i].name, mv_bits);
				else if (mv_bits > 0 && mv_signed)
					fprintf(file_c,
						"\tUVO_EQUAL(%s_##X,%s_##Y,%d)",
						field_list[i].name,
						field_list[i].name, mv_bits);
				else
					fprintf(file_c,
						"\t(%s_##X != %s_##Y)",
						field_list[i].name,
						field_list[i].name);
			if (i < cell_field_count - 1)
				if (mv_bits > 0)
					fprintf(file_c, "&");
				else
					fprintf(file_c, "&&");
			fprintf(file_c, " \\\n");
		}
		fprintf(file_c, ")\n\n");
	} else
		fprintf(file_c, "0)\n\n"); 

	gen_cell_copy("copy_cell", true, true);
	gen_cell_copy("assign_cell", false, false);
	gen_cell_copy("assign_const", false, true);

	/* Write the macro definition for "guarded_copy_cell". */
	if (mv_bits > 0) {
		fprintf(file_c, "#define guarded_copy_cell(X,Y) "); 
		if (cell_field_count > 0) {
			fprintf(file_c, "{ \\\n"); 
			for (i = 0; i < cell_field_count; i++) {
				if (field_list[i].alias) {
					char tmp[MAX_STRING_LENGTH];
					char *name, *index;
					strcpy(tmp, field_list[i].name);
					name = strtok(tmp, "[");
					index = strtok((char*) NULL, "]");
					fprintf(file_c,
						"\tguarded_copy(%s_##X[%s], %s_##Y[%s])",
						name, index, name, index);
				} else
					fprintf(file_c,
						"\tguarded_copy(%s_##X, %s_##Y)",
						field_list[i].name,
						field_list[i].name);
				fprintf(file_c, "; \\\n");
			}
			fprintf(file_c, "}\n\n");
		} else
			fprintf(file_c, ";\n\n"); 
	}

	fprintf(file_c, "\n\n");
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Sets a global variable (const_field) to indicate that
		the next cell field is a constant cell field.  Constant
		fields are only allowed for cell fields and NOT for agent
		fields. 
*/
void set_const_field() {
	if (agent_declared)
		error("Agent fields cannot be constant", (char*) NULL);
	else {
		const_field = true;
		constant_fields = true;
	}
}

/************************************************************************/

/* This is the data_type used for variables. */
#define VAR_TYPE (use_floats ? "float" : (mv_bits > 0 ? "uvec" : "long int"))

/* SUPPORT ROUTINE
      ACTION - Determine what C data type is the smallest that can 
	       accomodate the range of values in range.
*/
void get_type(char *data_type, range_entry range) {

	if (mv_bits > 0)
		strcpy(data_type, "uvec");
	else if (use_floats)
		strcpy(data_type, "float");

        /* Unsigned ranges. */
        else if (range.lb.lower_bound >= 0 && (unsigned) range.ub.upper_bound <= UCHAR_MAX)
		strcpy(data_type, "unsigned char");
        else if (range.lb.lower_bound >= 0 && (unsigned) range.ub.upper_bound <= USHRT_MAX)
		strcpy(data_type, "unsigned short int");
        else if (range.lb.lower_bound >= 0 && (unsigned) range.ub.upper_bound <= UINT_MAX)
		strcpy(data_type, "unsigned int");
        else if (range.lb.lower_bound >= 0 && (unsigned) range.ub.upper_bound <= ULONG_MAX)
		strcpy(data_type, "unsigned long int");

        /* Signed ranges. */
        else if (range.lb.lower_bound >= SCHAR_MIN && range.ub.upper_bound <= SCHAR_MAX)
		strcpy(data_type, "signed char");
        else if (range.lb.lower_bound >= SHRT_MIN && range.ub.upper_bound <= SHRT_MAX)
		strcpy(data_type, "short int");
        else if (range.lb.lower_bound >= INT_MIN && range.ub.upper_bound <= INT_MAX)
		strcpy(data_type, "int");
        else if (range.lb.lower_bound >= LONG_MIN && range.ub.upper_bound <= LONG_MAX)
		strcpy(data_type, "long int");
        else {
		char msg[MAX_STRING_LENGTH];
		sprintf(msg, "The range [%d..%d] is NOT supported",
			range.lb.lower_bound, range.ub.upper_bound);
		error(msg, (char*) NULL);

		strcpy(data_type, "long int");
	}
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackRange, StackToken, ... StackToken, StackMark
      LEAVES  - n/a
      ACTION  - Take the list of field identifiers from the stack and
		delcare them with the given range.  Declarations are
		generated.
*/
void declare_fields(void) {
	char data_type[MAX_STRING_LENGTH];
	int mark, element_count = 0;
	range_entry range;
	int field_type;
	boolean assignable;

	range = stack[top--].data.range;

	/* Determine the kind (cell or agent) of field to declare. */
	if (agent_declared) {
		field_type = AGENT_FIELD;
		assignable = false;
	} else {
		field_type = CELL_FIELD;
		assignable = !const_field;
	}

	/* Determine the data type to use for these fields. */
	get_type(data_type, range);

	/* Go down to the mark. */
        mark = top;
	while (stack[mark].kind != StackMark) mark--;

	/* Enter each of the fields into the symbol table. */ 
	while (++mark <= top) {
		attr_rec attribute;
		char *token = stack[mark].data.token;

		/* Enter into symbol table if necessary. */
		if (!intable(token, CELL_FIELD) && !intable(token, AGENT_FIELD)) {
			char address[MAX_STRING_LENGTH];
			attribute.assignable = assignable;
			attribute.is_float = use_floats;
			attribute.field_name = field_type;
			attribute.agent = false;
			attribute.array = false;
			attribute.index = false;
			attribute.structured = false;

			attribute.address = 
					mymalloc((unsigned int) 
						 (strlen(token) + 1));
			strcpy(attribute.address, token);

			field_list[field_count].alias = false;
			field_list[field_count].constant = const_field;

			enterid(token, attribute);

			/* Write the C code field declaration. */
			if (agent_declared) {
				print_indent();
				fprintf(file_c, "%s %s;\n", data_type, token);
			}
		} else
			error("Redeclaration of field, '%s'", token);

		field_list[field_count].name = 
			mymalloc((unsigned int) (strlen(attribute.address) + 1));
		strcpy(field_list[field_count].name, attribute.address);

		if (field_type == CELL_FIELD &&
		    longest_field_name <
			(int) strlen(field_list[field_count].name))
			longest_field_name =
				strlen(field_list[field_count].name);

		field_list[field_count].range = range;

		field_count++;
	}

	/* Clean stuff off the stack. */
	while (stack[top].kind != StackMark)
		myfree(stack[top--].data.token);
	top--;

	const_field = false;
}

/* SEMANTIC ROUTINE
      EXPECTS - StackRange, StackExpr, StackToken
      LEAVES  - n/a
      ACTION  - Declare the StackToken as an array of the given range.
		The size of the array is taken from the StackExpr.
		Declarations are generated.
*/
void declare_array_field(void) {
	char *array_name = NULL, data_type[MAX_STRING_LENGTH];
	range_entry range = stack[top--].data.range;
	int element_count = atoi(stack[top--].data.expr.addr);
	int field_type;
	boolean assignable;

	array_name = stack[top--].data.token;

	/* Determine the kind (cell or agent) of field to declare. */
	if (agent_declared) {
		field_type = AGENT_FIELD;
		assignable = false;
	} else {
		field_type = CELL_FIELD;
		assignable = !const_field;
	}

	/* Make sure that the array will contain at least 1 element. */
	if (element_count < 1) {
		error_prefix();
		fprintf(stderr, 
			"Warning; the array field, '%s', was declared with size %d, assuming 1\n", 
			array_name, element_count);
		element_count = 1;
	}

	/* Determine the data type to use for these fields. */
	get_type(data_type, range);

	/* Write the C code field declaration. */
	if (agent_declared) {
		print_indent();
		fprintf(file_c, "%s %s[%d];\n", data_type, array_name,
			element_count);
	}

	/* Enter the array_name into symbol table. */
	if (!intable(array_name, CELL_FIELD) && !intable(array_name, AGENT_FIELD)) {
		int i;
		attr_rec attribute;
		attribute.assignable = assignable;
		attribute.is_float = use_floats;
		attribute.field_name = field_type;
		attribute.agent = false;
		attribute.array = true;
		attribute.size = element_count;
		attribute.index = false;
		attribute.structured = false;
		attribute.address = mymalloc((unsigned int) 
						 (strlen(array_name) + 1));
		strcpy(attribute.address, array_name);
		enterid(array_name, attribute);

		/* Remember the field name(s). */
		for (i = 0; i < element_count; i++) {
			char address[MAX_STRING_LENGTH];

			sprintf(address, "%s[%d]", array_name, i);

			field_list[field_count].name = 
				mymalloc((unsigned int) (strlen(address) + 1));
			strcpy(field_list[field_count].name, address);

			if (field_type == CELL_FIELD &&
			    longest_field_name < 
				(int) strlen(field_list[field_count].name))
				longest_field_name =
					strlen(field_list[field_count].name);

			field_list[field_count].range = range;
			field_list[field_count].alias = true;
			field_list[field_count].constant = const_field;

			field_count++;
		}

		/* Save information about the array for later use
		   in declaring its universe array.
		*/
		array_field_alias_list[array_field_alias_count].name =
							array_name;
		array_field_alias_list[array_field_alias_count].range = range;
		array_field_alias_list[array_field_alias_count].size =
							element_count;
		array_field_alias_list[array_field_alias_count].constant =
							const_field;
		array_field_alias_count++;
	} else
		error("Redeclaration of field, '%s'", array_name);

	const_field = false;
}

/* SEMANTIC ROUTINE
      EXPECTS - StackRange
      LEAVES  - n/a
      ACTION  - Generate a declaration for cells to be of a single field
                with the given range.  No code is generated.
*/
void declare_default_field(void) {
	char data_type[MAX_STRING_LENGTH];
	range_entry range;

	/* Whether a field was specified on the command line or not,
	   treat it as if it were not (with the possible exception of
	   an error message).
	*/
	if (field_given && field_given != 1)
		error("Unknown field, '%d', given on command line",
			(char*) field_given);
	field_given = 0;

	range = stack[top--].data.range;

	get_type(data_type, range);

	print_indent();
	fprintf(file_c, "typedef %s cell_type;\n\n", data_type);

	field_list[field_count].range = range;

	cell_field_count = ++field_count;

	/* Build appropriate macros for "eq", "neq", "changed", and
	   "copy_cell".
	*/
	if (mv_bits > 0 && !mv_signed) {
		fprintf(file_c,
			"#define neq(X,Y) (UV_NEQUAL((X),(Y),%d))\n\n", mv_bits);
		fprintf(file_c,
			"#define eq(X,Y) (UV_EQUAL((X),(Y),%d))\n\n", mv_bits);
	} else if (mv_bits > 0 && mv_signed) {
		fprintf(file_c,
			"#define neq(X,Y) (UVO_NEQUAL((X),(Y),%d))\n\n", mv_bits);
		fprintf(file_c,
			"#define eq(X,Y) (UVO_EQUAL((X),(Y),%d))\n\n", mv_bits);
	} else {
		fprintf(file_c, "#define neq(X,Y) ((X)!=(Y))\n\n");
		fprintf(file_c, "#define eq(X,Y) ((X)==(Y))\n\n");
	}
	fprintf(file_c, "#define changed(X,Y) neq(X,Y)\n\n");
	fprintf(file_c, "#define copy_cell(X,Y) ((X)=(Y))\n\n");
	fprintf(file_c, "#define assign_cell(X,Y) ((X)=(Y))\n\n");
}

/************************************************************************/

/* SUPPORT ROUTINE
      ACTION - Writes C code for indexing [dim_1]...[dim_n].  The first
	       dimension is a special case and gets the variable via the
	       dim_1 parameter.
*/
void gen_current_index(char *dim_1) {
	int i, j;
	fprintf(file_c, "[");
	for (i = 1; i <= dimension_count; i++) {
		if (i == 1)
			fprintf(file_c, "%s", dim_1);
		else
			fprintf(file_c, "+dim_%d", i);
		for (j = dimension_count; j > i; j--)
			fprintf(file_c, "*(%d+MAX_OFFSET*2)", dim_size[j-1]);
	}
	fprintf(file_c, "]");
}

/************************************************************************/

/* Used to indicate whether or not semantic action processing is occuring
   within the context of a forall loop.  The begin_forall and the
   begin_forall_agent semantic actions add one to loop_count, whereas
   end_forall subtracts one.  The exit_forall semantic action can then
   determine whether or not it is being used within a loop (i.e.,
   loop_count > 0).
*/
int loop_count = 0;

/* Used to generate unique names for the macro names used to represent
   the lower and upper bounds of an index variable.
*/
int index_count = 0;

/* SEMANTIC ROUTINE
      EXPECTS - StackToken
      LEAVES  - StackIndexVar
      ACTION  - Declares the StackToken as an index variable and generates
		code to begin the forall loop.
*/
void begin_forall_agent(void) {
	attr_rec attribute;

	loop_count++;

	/* Make sure that agents were declared for this program. */
	if (!agent_declared)
		error("No agents declared for this program", (char*) NULL);

	attribute.assignable = false;
	attribute.is_float = use_floats;
	attribute.field_name = VARIABLE;
	attribute.structured = true;
	attribute.agent = true;
	attribute.array = false;
	attribute.index = true;
	attribute.cell_ref = false;
	attribute.index_count = ++index_count;
	attribute.address = stack[top].data.token;

	if (intable(attribute.address, VARIABLE))
		error("Redeclaration of variable, '%s'", attribute.address);
	else
		enterid(attribute.address, attribute);

	stack[top].kind = StackIndexVar;
	stack[top].data.index.agent = true;
	stack[top].data.index.attribute = 
				get_attribute(attribute.address, VARIABLE);
	stack[top].data.index.name = attribute.address;
	stack[top].data.index.attribute->range_set = true;
	stack[top].data.index.attribute->incomp_range = false;

	/* Declare the index variable. */
	print_indent();
	fprintf(file_c, "{\n");
	indent;
	print_indent();
	fprintf(file_c, "register struct agent_type *_%s;\n", attribute.address);

	/* Generate C code to start the loop. */
	print_indent();
	fprintf(file_c, "for (_%s = agent_now", attribute.address);
	gen_current_index("dim_1");
	fprintf(file_c, "; _%s != NULL; _%s = _%s->agent) {\n",
		attribute.address, attribute.address, attribute.address);
	indent;
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackToken
      LEAVES  - StackIndexVar
      ACTION  - Declares the StackToken as an index variable and generates
		code to begin the forall loop.
*/
void begin_forall(void) {
	attr_rec attribute;

	loop_count++;

	attribute.assignable = false;
	attribute.is_float = use_floats;
	attribute.field_name = VARIABLE;
	attribute.structured = false;
	attribute.agent = false;
	attribute.array = false;
	attribute.index = true;
	attribute.cell_ref = false;
	attribute.index_count = ++index_count;
	attribute.address = stack[top].data.token;

	if (intable(attribute.address, VARIABLE))
		error("Redeclaration of variable, '%s'", attribute.address);
	else
		enterid(attribute.address, attribute);

	stack[top].kind = StackIndexVar;
	stack[top].data.index.agent = false;
	stack[top].data.index.attribute = 
				get_attribute(attribute.address, VARIABLE);
	stack[top].data.index.name = attribute.address;
	stack[top].data.index.attribute->range_set = false;
	stack[top].data.index.attribute->incomp_range = false;

	/* Declare the index variable. */
	print_indent();
	fprintf(file_c, "{\n");
	indent;
	print_indent();
	fprintf(file_c, "register ");
	if (mv_bits > 0) fprintf(file_c, "uvec");
	else fprintf(file_c, "int");
	fprintf(file_c, " _%s;\n", attribute.address);

	/* Generate C code to start the loop. */
	print_indent();
	if (mv_bits > 0) {
		fprintf(file_c, 
			"for (_%s = %s(INDEX_%d_LB,%d); %s(_%s,%d) <= INDEX_%d_UB; _%s=%s(_%s,%s(1,%d),%d)) {\n",
			attribute.address,
			(mv_signed ? "UVO_S2VEC" : "UV_S2VEC"),
			attribute.index_count, mv_bits,
			(mv_signed ? "UVO_FSVAL" : "UV_FVAL"),
			attribute.address, mv_bits,
			attribute.index_count,
			attribute.address,
			(mv_signed ? "UVO_PLUS" : "UV_PLUS"),
			attribute.address,
			(mv_signed ? "UVO_S2VEC" : "UV_S2VEC"),
			mv_bits, mv_bits);
	} else
		fprintf(file_c, 
			"for (_%s = INDEX_%d_LB; _%s <= INDEX_%d_UB; _%s++) {\n",
			attribute.address, attribute.index_count,
			attribute.address, attribute.index_count,
			attribute.address);
	indent;
}

/* SEMANTIC ROUTINE
      EXPECTS - StackIndexVar
      LEAVES  - n/a
      ACTION  - Removes the index variable indicated by the StackIndexVar
		from the symbol table and generates code to end the forall
		loop.
*/
void end_forall(void) {
	attr_rec *attribute = stack[top].data.index.attribute;

	loop_count--;

	/* If the forall loop being closed had an error, just cleanup. */
	if (stack[top].kind == StackError) {
		stack_clear(1);
		top--;
		return;
	}

	/* Close off the body of the loop. */
	undent;
	print_indent();
	fprintf(file_c, "}\n");

	/* Close off the block containing the index variable declaration. */
	undent;
	print_indent();
	fprintf(file_c, "}\n");

	if (!attribute->agent) {
		/* Make sure a range was actually set for the index variable. */
		if (!attribute->range_set)
			error("Range of index variable '%s' could not be inferred", 
				stack[top].data.index.name);

		/* Make sure that all inferred ranges agreed. */
		if (attribute->incomp_range)
			error("Incompatible range(s) set/inferred for '%s'", 
				attribute->address);

		/* Define macros for bounds of index variable. */
		fprintf(file_h, "#define INDEX_%d_LB %d\n",
			attribute->index_count, attribute->lower);
		fprintf(file_h, "#define INDEX_%d_UB %d\n",
			attribute->index_count, 
			attribute->lower + attribute->size - 1);
		fprintf(file_h, "#define INDEX_%d_SIZE %d\n",
			attribute->index_count, attribute->size);

		/* Update MAX_OFFSET if necessary. */
		if (attribute->cell_ref) {
			if (abs(attribute->lower) > max_rel_index)
				max_rel_index = abs(attribute->lower);
			if (abs(attribute->lower+attribute->size) > max_rel_index)
				max_rel_index = abs(attribute->lower+attribute->size);
		}
	}

	removeid(stack[top].data.index.name, VARIABLE);
	top--;
}

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - If the value of loop_count > 0 then generate a C "break"
		statment to exit the deepest enclosing loop, otherwise
		issue an error message (exit not used within a loop).
		NOTE:  Exits are incompatible with microvectors since
		they would prevent further evaluation.
*/
void exit_forall(void) {
	if (mv_bits > 0)
		error("Exit statements are incompatible with microvectors",
		      (char*) NULL);
	else if (loop_count > 0) {
		print_indent();
		fprintf(file_c, "break;\n");
	} else
		error("An exit statement can only be used within a forall loop",
			(char*) NULL);
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr, StackExpr
      LEAVES  - StackRange
      ACTION  - Convert the top two entries on the semantic stack into
		a single range entry.
*/
void make_range(void) {
	if (stack_error(2)) {
		stack_clear(2);
		top--;
		stack[top].kind = StackError;
	} else {
		range_entry range;

		/* Build the range entry. */
		if (use_floats && loop_count == 0) {
			sscanf(stack[top].data.expr.addr, "%f",
			       &range.ub.float_upper_bound);
			myfree(stack[top--].data.expr.addr);
			sscanf(stack[top].data.expr.addr, "%f",
			       &range.lb.float_lower_bound);
			myfree(stack[top--].data.expr.addr);
		} else {
			range.ub.upper_bound = atoi(stack[top].data.expr.addr);
			myfree(stack[top--].data.expr.addr);
			range.lb.lower_bound = atoi(stack[top].data.expr.addr);
			myfree(stack[top--].data.expr.addr);
		}

		/* The range should be non-empty. */
		if (use_floats && range.ub.float_upper_bound <
						range.lb.float_lower_bound) {
			char string [MAX_STRING_LENGTH];
			sprintf(string,
				"Range [%d..%d] must be non-empty", 
				range.lb.lower_bound, range.ub.upper_bound);
			error(string, (char*) NULL);
		} else if (range.ub.upper_bound < range.lb.lower_bound) {
			char string [MAX_STRING_LENGTH];
			sprintf(string,
				"Range [%d..%d] must be non-empty", 
				range.lb.lower_bound, range.ub.upper_bound);
			error(string, (char*) NULL);
		}

		/* Push the range entry onto the stack. */
		new_entry();
		stack[top].kind = StackRange;
		stack[top].data.range = range;
	}
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackRange, StackIndexVar
      LEAVES  - StackIndexVar
      ACTION  - Set the bound on the Index variable using the range at
		the top of the stack.
*/
void set_index_bounds(void) {
	if (stack_error(2))
		top--;
	else {
		range_entry range;
		range = stack[top--].data.range;

		stack[top].data.index.attribute->range_set = true;

		/* Assume the bounds is NOT a floating point range. */
		stack[top].data.index.attribute->lower = range.lb.lower_bound;
		stack[top].data.index.attribute->size = 
			range.ub.upper_bound - range.lb.lower_bound + 1;
		stack[top].data.index.attribute->usr_def = true;
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackIndexVar
      LEAVES  - StackIndexVar
      ACTION  - Set the Index variable indicating that the range was
		NOT given by the programmer.
*/
void implicit_bounds(void) {
	stack[top].data.index.attribute->usr_def = false;
	stack[top].data.index.attribute->range_set = false;
}

/************************************************************************/

/* SUPPORT ROUTINE
      ACTION - Generates code to delcare the next/now variables for each
	       of the arrays (of fields) for the cell universe.
*/
void gen_now_next(void) {
	int i, j;

	if (!cell_fields_declared && cell_field_count == 1) {
		print_indent();
		fprintf(file_c, "register cell_type *now = &cells");
		fprintf(file_c, "[(MAX_OFFSET<<1)+1");
#if MAX_CPUS > 1
		if (cpus > 1) fprintf(file_c, "+id*(2*MAX_OFFSET+1)");
#endif
		fprintf(file_c, "]");
		for (i = 1; i < dimension_count; i++)
			fprintf(file_c, "[MAX_OFFSET]");
		fprintf(file_c, ";\n");

		print_indent();
		fprintf(file_c, "register cell_type *next = &cells");
		fprintf(file_c, "[MAX_OFFSET");
#if MAX_CPUS > 1
		if (cpus > 1) fprintf(file_c, "+id*(2*MAX_OFFSET+1)");
#endif
		fprintf(file_c, "]");
		for (i = 1; i < dimension_count; i++)
			fprintf(file_c, "[MAX_OFFSET]");
		fprintf(file_c, ";\n\n");
		return;
	}

	for (j = 0; j < cell_field_count; j++) {
		char data_type[MAX_STRING_LENGTH];
		range_entry range;

		if (field_list[j].alias) continue;

        	range = field_list[j].range;

		/* Determine the data type to use for this field. */
		get_type(data_type, range);

		print_indent();
		fprintf(file_c, "register %s *%s_now = &%s_cells",
			data_type, field_list[j].name,
			field_list[j].name);

		if (field_list[j].constant) {
			int i;
			for (i = 0; i < dimension_count; i++)
				fprintf(file_c, "[MAX_OFFSET]");
			fprintf(file_c, ";\n");
			continue;
		}

		fprintf(file_c, "[(MAX_OFFSET<<1)+1");
#if MAX_CPUS > 1
		if (cpus > 1) fprintf(file_c, "+id*(2*MAX_OFFSET+1)");
#endif
		fprintf(file_c, "]");
		for (i = 1; i < dimension_count; i++)
			fprintf(file_c, "[MAX_OFFSET]");
		fprintf(file_c, ";\n");

		print_indent();
		fprintf(file_c, "register %s *%s_next = &%s_cells",
			data_type, field_list[j].name,
			field_list[j].name);
		fprintf(file_c, "[MAX_OFFSET");
#if MAX_CPUS > 1
		if (cpus > 1) fprintf(file_c, "+id*(2*MAX_OFFSET+1)");
#endif
		fprintf(file_c, "]");
		for (i = 1; i < dimension_count; i++)
			fprintf(file_c, "[MAX_OFFSET]");
		fprintf(file_c, ";\n\n");
	}
	for (j = 0; j < array_field_alias_count; j++) {
		print_indent();
		fprintf(file_c, "register %s_type *%s_now = &%s_cells",
			array_field_alias_list[j].name,
			array_field_alias_list[j].name,
			array_field_alias_list[j].name);
		fprintf(file_c, "[(MAX_OFFSET<<1)+1");
#if MAX_CPUS > 1
		if (cpus > 1) fprintf(file_c, "+id*(2*MAX_OFFSET+1)");
#endif
		fprintf(file_c, "]");
		for (i = 1; i < dimension_count; i++)
			fprintf(file_c, "[MAX_OFFSET]");
		fprintf(file_c, ";\n");

		if (array_field_alias_list[j].constant) continue;

		print_indent();
		fprintf(file_c, "register %s_type *%s_next = &%s_cells",
			array_field_alias_list[j].name,
			array_field_alias_list[j].name,
			array_field_alias_list[j].name);
		fprintf(file_c, "[MAX_OFFSET");
#if MAX_CPUS > 1
		if (cpus > 1) fprintf(file_c, "+id*(2*MAX_OFFSET+1)");
#endif
		fprintf(file_c, "]");
		for (i = 1; i < dimension_count; i++)
			fprintf(file_c, "[MAX_OFFSET]");
		fprintf(file_c, ";\n\n");
	}
	if (agent_declared) {
		print_indent();
		fprintf(file_c, "register struct agent_type **agent_now = &agent_cells");
		fprintf(file_c, "[(MAX_OFFSET<<1)+1");
#if MAX_CPUS > 1
		if (cpus > 1) fprintf(file_c, "+id*(3*MAX_OFFSET+1)");
#endif
		fprintf(file_c, "]");
		for (i = 1; i < dimension_count; i++)
			fprintf(file_c, "[MAX_OFFSET]");
		fprintf(file_c, ";\n");

		print_indent();
		fprintf(file_c, "register struct agent_type **agent_next = &agent_cells");
		fprintf(file_c, "[MAX_OFFSET");
#if MAX_CPUS > 1
		if (cpus > 1) fprintf(file_c, "+id*(3*MAX_OFFSET+1)");
#endif
		fprintf(file_c, "]");
		for (i = 1; i < dimension_count; i++)
			fprintf(file_c, "[MAX_OFFSET]");
		fprintf(file_c, ";\n\n");
	}
}

/* SUPPORT ROUTINE
      ACTION - Change all occurences of the string "3+1" to "2" in the
	       address as the address is written to file.  This is needed
	       in order to make the size alotted to constant fields correct.
	       In addition, the string "+(CPUS-1)*(3*MAX_OFFSET+1)" should
	       be replaced with "+(0)".
*/
void change_offset(FILE *file, char *address) {
	unsigned int pos;
	for (pos = 0; pos < strlen(address); pos++)
		if (address[pos] == '3' && address[pos+1] == '+'
					&& address[pos+2] == '1') {
			fprintf(file, "2");
			pos+=2;
		} else if (address[pos] == 'C') {
			fprintf(file, "0)");
			pos+=23;
		} else
			fprintf(file, "%c", address[pos]);
}

/* SUPPORT ROUTINE
      ACTION - Generates code to delcare the arrays (of fields) for the
	       cell universe and/or variables.  If "all" is true then for
	       constant cell fields only generate enough extra space for
	       the boundary.
*/
void gen_field_arrays(FILE *file, char *store_class, char *name, char *dims, boolean all) {
	int j;

	if (!cell_fields_declared) {
		fprintf(file, "%s cell_type %s%s;\n", store_class, name, dims);
		return;
	}

	for (j = 0; j < cell_field_count; j++) {
		char data_type[MAX_STRING_LENGTH];
		range_entry range;

		if (field_list[j].alias) continue;

        	range = field_list[j].range;

		/* Determine the data type to use for this field. */
		get_type(data_type, range);

		if (field_list[j].constant) {
			fprintf(file, "%s %s %s_%s",
				store_class, data_type,
				field_list[j].name, name);
			if (all) change_offset(file, dims);
			else fprintf(file, "%s", dims);
		} else
			fprintf(file, "%s %s %s_%s%s",
				store_class, data_type,
				field_list[j].name, name, dims);

		fprintf(file, ";\n");
	}
	for (j = 0; j < array_field_alias_count; j++) {
		char data_type[MAX_STRING_LENGTH];

		if (strcmp(name, "cells") == 0) {
			range_entry range;

        		range = array_field_alias_list[j].range;

			/* Determine the data type to use for this field. */
			get_type(data_type, range);

			fprintf(file, "typedef %s %s_type[%d];\n",
				data_type, array_field_alias_list[j].name,
				array_field_alias_list[j].size);
		}

		if (array_field_alias_list[j].constant) {
			fprintf(file, "%s_type %s_%s",
				array_field_alias_list[j].name,
				array_field_alias_list[j].name, name);
			if (all) change_offset(file, dims);
			else fprintf(file, "%s", dims);
		} else
			fprintf(file, "%s_type %s_%s%s",
				array_field_alias_list[j].name,
				array_field_alias_list[j].name, name, dims);

		fprintf(file, ";\n");
	}
}

/* SUPPORT ROUTINE
      ACTION  - Declare var_name as a variable of type.
*/
void decl_variable(char *var_name, char *type) {
	attr_rec attribute;
	
	/* Enter into symbol table. */
	attribute.assignable = true;
	attribute.is_float = use_floats;
	attribute.field_name = VARIABLE;
	attribute.agent = (0 == strcmp(type, "struct agent_type"));
	attribute.structured = attribute.agent || 
					(0 == strcmp(type, "cell_type"));
	attribute.array = false;
	attribute.index = false;
	attribute.address = mymalloc((unsigned int) (strlen(var_name) + 1));
	strcpy(attribute.address, var_name);
	enterid(var_name, attribute);

	/* Write code to declare the variable. */
	if (!attribute.structured) fprintf(file_h, "register ");

	if (strcmp(type, "cell_type") == 0)
		gen_field_arrays(file_h, "", var_name, "", false);
	else
		fprintf(file_h, "%s _%s;\n", type, var_name);
}

/* SUPPORT ROUTINE
      ACTION  - Print the logical combinations of all of the true and false
		guards in the mv_guards stack as an assignment mask!
*/
void print_guard(void) {
	mv_guard *tmp = mv_guards;

	fprintf(file_c, "UV_SMUL((");
	tmp = mv_guards;
	while (tmp) {
		if (tmp->true_guard && tmp->false_guard)
			fprintf(file_c,
				"%s(%s(%s,%d),%s,%d)",
				(mv_signed ? "UVO_AND" : "UV_AND"),
				(mv_signed ? "UVO_NOT" : "UV_NOT"),
				tmp->false_guard, mv_bits,
				tmp->true_guard, mv_bits);
		else if (tmp->true_guard)
			fprintf(file_c, "%s", tmp->true_guard);
		else if (tmp->false_guard)
			fprintf(file_c, "%s(%s,%d)",
				(mv_signed ? "UVO_NOT" : "UV_NOT"),
				tmp->false_guard, mv_bits);
		else
			/* This should never happen, but... */
			fprintf(file_c, "%s(1,%d)",
				(mv_signed ? "UVO_S2VEC" : "UV_S2VEC"), mv_bits);

		tmp = tmp->next;
		if (tmp) fprintf(file_c, "&");
	}
	fprintf(file_c, ")&WORD_MASK,BIT_MASK,%d)", mv_bits);
}

/* SUPPORT ROUTINE
      ACTION - Declares the temporary variable "assign_index" inside the
	       main body of the generated code, by putting the declaration
	       in the file_h header file.  The declaration should only be
	       made once.  The declaration must be in the main body in order
	       to work properly for multiple processors since the value
	       should not be shared.
*/
boolean assign_index_declared = false;
void declare_assign_index() {
	if (!assign_index_declared) {
		fprintf(file_h, "unsigned long int assign_index;\n");
		assign_index_declared = true;
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr, ..., StackExpr, StackMark, StackExpr
      LEAVES  - StackExpr
      ACTION  - Writes code to perform an assignment.  Implicit declarations
	        of variables are made, with the type of the variable being
	        determined by the first right hand side given.  If micovectors
		are being used, then a mask is built using the guards in the
		mv_guards stack and used in performing the assignment.
*/
void do_assign(void) {
        int mark = top;
	boolean stack_error = false;
	boolean incompatible = false;
	boolean loop_assign = false;

        /* Go down to the mark, checking for legal stack entries. */
        while (stack[mark].kind != StackMark)
		if (stack[mark--].kind == StackError)
			stack_error = true;

	if (stack[mark-1].kind == StackError) stack_error = true;

	if (stack_error) {
		while (stack[top].kind != StackMark) {
			if (stack[top].kind != StackError)
				myfree(stack[top].data.expr.addr);
			top--;
		}
		top--;
	} else {
		expr_entry *assign_expr;
		expr_entry expr;
		boolean base_structured, multiple_expr = false;
		int index = 0;

		assign_expr = &stack[mark-1].data.expr;
		expr = stack[++mark].data.expr;

		/* Make sure this is not an index variable. */
		if (assign_expr->pre_declared && assign_expr->index) {
			error("The index variable, '%s', cannot be assigned",
				assign_expr->var_tname);
			while (stack[top].kind != StackMark)
				myfree(stack[top--].data.expr.addr);
			stack[--top].kind = StackError;
			return;
		}
		/* Make sure the left hand side is an assignable variable. */
		else if (assign_expr->pre_declared && !assign_expr->assignable) {
			error("The variable, '%s', cannot be assigned",
				assign_expr->var_tname);
			while (stack[top].kind != StackMark)
				myfree(stack[top--].data.expr.addr);
			stack[--top].kind = StackError;
			return;
		}

		/* Declare the variable if necessary. */
		if (!assign_expr->pre_declared) {
			boolean base_agent = expr.agent ||
						(assign_expr->field &&
						 intable(assign_expr->field_tname,
							 AGENT_FIELD));

			if (assign_expr->array && assign_expr->free_array
			    && assign_expr->field)
				error("First use of unindexed array '%s' must be without a field reference",
				      assign_expr->addr);

			base_structured = assign_expr->field ||
					    !expr.unstructured;

			assign_expr->unstructured = !base_structured
							||
						(base_structured
							&&
						 assign_expr->field != NULL);
			assign_expr->agent = base_agent;

			if ((assign_expr->array && !assign_expr->field) ||
			    assign_expr->attr_list) {
				attr_rec *attribute;

				/* Generate the C code declaration. */
				if (base_structured && base_agent)
					fprintf(file_h,
						"struct agent_type _%s[%d];\n", 
						assign_expr->addr, 
						assign_expr->size);
				else if (base_structured) {
					char dims[MAX_STRING_LENGTH];
					sprintf(dims, "[%d]", assign_expr->size);
					gen_field_arrays(file_h, "",
							 assign_expr->addr,
							 dims, false);
				} else
					fprintf(file_h, "%s _%s[%d];\n", 
						VAR_TYPE, assign_expr->addr, 
						assign_expr->size);
	
				/* Alter symbol table entry. */
				attribute = get_attribute(assign_expr->addr, 
								VARIABLE);
				attribute->agent = base_agent;
				attribute->structured = base_structured;
			} else
				/* Put it in the symbol table. */
				if (!base_structured)
					decl_variable(assign_expr->addr, 
							VAR_TYPE);
				else
					decl_variable(assign_expr->addr, 
						(!base_agent ? 
							"cell_type" : 
							"struct agent_type"));

			assign_expr->pre_declared = true;

			/* If the assignable expression was an array
			   declaration, then set the structuredness of the 
			   attributes in the list.  Also indicates whether
			   it is a cell or agent variable if structured.
			*/
			if (assign_expr->array) {
				attr_rec_list *next = assign_expr->attr_list;
				while (next) {
					attr_rec_list *tmp = next;

					next->attribute->structured =
							base_structured;

					if (base_structured)
						next->attribute->agent =
							assign_expr->agent;

					next = next->next;

					myfree(tmp);
				}
				assign_expr->attr_list = NULL;
			}
		}

		/* Mark the assign_expr as an agent if it is. */
		if (expr.agent || 
		    (assign_expr->field && 
				intable(assign_expr->field_tname, AGENT_FIELD)))
			assign_expr->agent = true;

		/* The list of expressions should be of length 1 except when 
		   assigning to an entire array.
		*/
		if ((top-mark+1 > 1) && !assign_expr->array)
			error("Cannot assign multiple expressions to non-array '%s'",
				assign_expr->addr);

		/* If an array is being assigned to, then the number of
		   expressions should either 1 or be the same as the size 
		   of the array.
		*/
		else if (assign_expr->array && (top-mark+1) != 1 &&
			 (top-mark+1) > assign_expr->size)
			error("Too many expressions to assign to array '%s'",
				assign_expr->addr);
		else if (assign_expr->array && (top-mark+1) != 1 &&
			 (top-mark+1) < assign_expr->size)
			error("Too few expressions to assign to array '%s'",
				assign_expr->addr);

		/* If microvectors are being used, then build the guard to
		   use for the assignment(s).
		*/
		if (mv_bits > 0 && mv_guards) {
			print_indent();
			fprintf(file_c, "guard = ");
			print_guard();
			fprintf(file_c, ";\n");
		}

		/* If an array is being assigned a single value, and if
		   the number of array elements is greater than MAX_N, assign
		   the same value to each of the array elements by means of
		   a for loop.
		*/
		if (assign_expr->array && (top-mark+1) == 1 &&
		    assign_expr->size > MAX_N) {
			print_indent();
			declare_assign_index();
			fprintf(file_c,
				"for (assign_index = 0; assign_index < %d; assign_index++) {\n",
				assign_expr->size);
			indent;
			loop_assign = true;
		}

		/* Assign all of the expressions. */
		do {
			char base_addr[MAX_STRING_LENGTH];
			char trace_addr[MAX_STRING_LENGTH];
			char value_addr[MAX_STRING_LENGTH];
			char tmp[MAX_STRING_LENGTH];
			expr = stack[mark].data.expr;

			/* The first time through the loop, determine if 
			   there are multiple expressions are on the stack.
			*/
			if (index == 0) multiple_expr = (top-mark+1) > 1;

			/* assign_expr must be an indexed array. */
			if (assign_expr->addr[0] == '_')
				sprintf(base_addr, "%s", assign_expr->addr);

			/* The variable "cell" is a special case. */
			else if (strcmp("cell", assign_expr->addr) == 0) {
				int i,j ;
				char dims[MAX_STRING_LENGTH];
				char tmp[MAX_STRING_LENGTH];

				strcpy(dims, "next[");
				for (i = 1; i < dimension_count; i++) {
					sprintf(tmp, "+dim_%d", i);
					strcat(dims, tmp);
					for (j = dimension_count; j > i; j--) {
						sprintf(tmp,
							"*(%d+MAX_OFFSET*2)",
							dim_size[j-1]);
						strcat(dims, tmp);
					}
				}
				sprintf(tmp, "+dim_%d]", dimension_count);
				strcat(dims, tmp);

				sprintf(base_addr, "%s", dims);
			} else
				sprintf(base_addr, "_%s", assign_expr->addr);

			/* Base value to use in printing the trace. */
			strcpy(value_addr, base_addr);

			/* Base name to use in printing the trace. */
			sprintf(trace_addr, "%s", assign_expr->var_tname);

			print_indent();

			if (assign_expr->unstructured && expr.unstructured) {
				char *base_name, *base_index;
				char *field_name, *field_index;

				/* If microvectors are being used, then used a
				   guarded assignment.
				*/
				if (mv_bits > 0 && mv_guards)
					fprintf(file_c, "guarded_copy(");

				base_name = strtok(base_addr, "[");
				base_index = strtok((char*) NULL, "]");

				if (assign_expr->field) {
					char tmp[MAX_STRING_LENGTH]; 
					strcpy(tmp, assign_expr->field);
					field_name = strtok(tmp, "[");
					field_index = strtok((char*) NULL, "]");
				}

				strcpy(value_addr, "");
				if (assign_expr->field && !assign_expr->agent) {
					fprintf(file_c, "%s", field_name);
					sprintf(tmp, "%s", field_name);
					strcat(value_addr, tmp);
					if (strcmp("cell", assign_expr->addr) == 0) {
						fprintf(file_c, "_");
						strcat(value_addr, "_");
					}
				}
				fprintf(file_c, "%s", base_name);
				strcat(value_addr, base_name);

				/* Determine whether or not the base variable
				   has been left free and thus the elements
				   of the array must be indexed by the compiler.
				*/
				if (assign_expr->array && assign_expr->free_array) {
					if (loop_assign) {
						fprintf(file_c, "[assign_index]");
						sprintf(tmp, "[%%d]");
					} else {
						fprintf(file_c, "[%d]", index++);
						sprintf(tmp, "[%d]", index-1);
					}

					strcat(trace_addr, tmp);

					if (loop_assign)
						sprintf(tmp, "[assign_index]");
					else
						sprintf(tmp, "[%d]", index-1);

					strcat(value_addr, tmp);
				} else if (assign_expr->array &&
					   assign_expr->free_field) {

					if (base_index) {
						fprintf(file_c, "[%s]",
							base_index);
						if (strcmp(trace_addr, "cell")) {
							sprintf(tmp, "[%s]",
								base_index);
							strcat(trace_addr, tmp);
						}
					}
					if (loop_assign) {
						fprintf(file_c,
							"[assign_index]");
					} else {
						fprintf(file_c, "[%d]", index++);
					}

					if (base_index) {
						sprintf(tmp, "[%s]", base_index);
						strcat(value_addr, tmp);
					}

					if (loop_assign)

						sprintf(tmp, "[assign_index]");
					else
						sprintf(tmp, "[%d]", index-1);
					strcat(value_addr, tmp);
				}
				if (assign_expr->field) {
					if (!assign_expr->field_index_tname) {
						sprintf(tmp, ".%s", 
							assign_expr->field);
						strcat(trace_addr, tmp);
					}
					if (assign_expr->agent)
						fprintf(file_c, ".%s", 
							assign_expr->field);
				}

				if (assign_expr->array && assign_expr->free_field) {
					if (loop_assign)
						sprintf(tmp, "[%%d]");
					else
						sprintf(tmp, "[%d]", index-1);
					strcat(trace_addr, tmp);
				} else if (base_index) {
					fprintf(file_c, "[%s]", base_index);
					sprintf(tmp, "[%s]", base_index);
					strcat(value_addr, tmp);
				}
				if (field_index && assign_expr->field) {
					fprintf(file_c, "[%s]", field_index);
					sprintf(tmp, "[%s]", field_index);
					strcat(value_addr, tmp);
				}

				if (mv_bits > 0 && mv_guards)
					fprintf(file_c, ", ");
				else
					fprintf(file_c, " = ");

				fprintf(file_c, "%s", expr.addr);

				if (mv_bits > 0 && mv_guards)
					fprintf(file_c, ")");
				fprintf(file_c, ";\n");
			} else if (!assign_expr->unstructured && !expr.unstructured) {
				char assign_call[20];

				strcpy(assign_call, (mv_bits > 0 && mv_guards)
							? "guarded_copy_cell"
							: "copy_cell");

				/* May have to index the base variable.  You
				   will never have to worry about a field being
				   indexed since that would mean an unstructured
				   value.
				*/
				if (assign_expr->array && !assign_expr->agent) {
					if (loop_assign) {
						fprintf(file_c,
							"%s(%s[assign_index], %s);\n", 
							assign_call,
							(base_addr[0] == '_' ?
								&base_addr[1] :
								base_addr),
							(expr.addr[0] == '_' ?
								&expr.addr[1] :
								expr.addr));

						sprintf(tmp, "[%%d]");
					} else {
						fprintf(file_c,
							"%s(%s[%d], %s);\n", 
							assign_call,
							(base_addr[0] == '_' ?
								&base_addr[1] :
								base_addr),
							index++,
							(expr.addr[0] == '_' ?
								&expr.addr[1] :
								expr.addr));

						sprintf(tmp, "[%d]", index-1);
					}

					strcat(trace_addr, tmp);

					if (loop_assign)
						sprintf(tmp, "%s[assign_index]",
							base_addr);
					else
						sprintf(tmp, "%s[%d]",
							base_addr, index-1);
					strcpy(value_addr, tmp);
				} else if (assign_expr->array && assign_expr->agent) {
					if (loop_assign) {
						fprintf(file_c, 
							"assign_agent(%s[%%d], %s);\n", 
							base_addr, expr.addr);

						sprintf(tmp, "[assign_index]");
					} else {
						fprintf(file_c, 
							"assign_agent(%s[%d], %s);\n", 
							base_addr, index++, expr.addr);

						sprintf(tmp, "[%d]", index-1);
					}

					strcat(trace_addr, tmp);

					if (loop_assign)
						sprintf(tmp, "%s[assign_index]", 
							base_addr);
					else
						sprintf(tmp, "%s[%d]", 
							base_addr, index-1);
					strcpy(value_addr, tmp);
				} else if (!assign_expr->agent) {
					fprintf(file_c, "%s(%s, %s);\n", 
						assign_call,
						(base_addr[0] == '_' ?
							&base_addr[1] :
							base_addr),
						(expr.addr[0] == '_' ?
							&expr.addr[1] :
							expr.addr));
				} else {
					fprintf(file_c, 
						"assign_agent(%s, %s);\n", 
						base_addr, expr.addr);
				}
			}
			/* Only print this once for each assignment(:=). */
			else if (!incompatible) {
				incompatible = true;
				error("Incompatible assignment", (char*) NULL);
			}

			/* If the trace option was specified, then generate 
			   the code to print the assignment result for the 
			   cell specified.
			*/
			if (trace_count > 0) {
				int i;

				print_indent();
				fprintf(file_c, 
					"if ((_time_step == next_report_time || _time_step == next_read_time)");
				for (i = 0; i < trace_count; i++)
					fprintf(file_c, " && dim_%d == %d",
						i + 1, trace_point[i]);
				fprintf(file_c, ") {\n");

				indent;
				print_indent();

				/* Generate code to write out a readable
				   left-hand-side for the assignment.
				*/
				fprintf(file_c, 
					"fprintf(stderr, \"   %s", trace_addr);

				if (assign_expr->var_index_tname)
					fprintf(file_c, "[%%d]");

				if (assign_expr->field_index_tname)
					fprintf(file_c, ".%s[%%d]",
						assign_expr->field_tname);

				fprintf(file_c, " := \"");

				if (assign_expr->var_index_tname)
					fprintf(file_c, ", %s",
						assign_expr->var_index_tname);
				else if (assign_expr->free_array &&
						loop_assign &&
						assign_expr->size > MAX_N)
					fprintf(file_c, ", assign_index");

				if (assign_expr->field_index_tname)
					fprintf(file_c, ", %s",
						assign_expr->field_index_tname);
				else if (assign_expr->free_field &&
						loop_assign &&
						assign_expr->size > MAX_N)
					fprintf(file_c, ", assign_index");

				fprintf(file_c, ");\n");

				/* Generate code to write out a readable
				   right-hand-side for the assignment.
				*/
				print_indent();
				if (expr.unstructured)
					fprintf(file_c, 
						"fprintf(stderr, \"%%d\", %s);\n",
						value_addr);
				else if (!assign_expr->agent) {
					/* Print out a structured cell value. */
					int i;

					fprintf(file_c, 
						"fprintf(stderr, \"cell(\");\n");

					for (i = 0; i < cell_field_count; i++) {
						char tmp[MAX_STRING_LENGTH];
						char *index;

						strcpy(tmp, field_list[i].name);

						print_indent();
						fprintf(file_c, 
							"fprintf(stderr, \"%%d");
						if (i == cell_field_count-1)
							fprintf(file_c, ")");
						else
							fprintf(file_c, ", ");

						fprintf(file_c, "\", %s%s",
							strtok(tmp, "["),
							value_addr);
						index = strtok(NULL, "]");
						if (index)
							fprintf(file_c,
								"[%s]", index);
						fprintf(file_c, ");\n");
					}
				} else {
					/* Print out a structured agent value. */
					int i;

					fprintf(file_c, 
						"fprintf(stderr, \"agent(\");\n");

					for (i = cell_field_count; i < field_count-1; i++) {
						print_indent();
						fprintf(file_c, 
							"fprintf(stderr, \"%%d, \", %s.%s);\n",
							value_addr,
							field_list[i].name);
					}
					print_indent();
					fprintf(file_c, 
						"fprintf(stderr, \"%%d)\", %s.%s);\n",
						value_addr,
						field_list[field_count-1].name);
				}

				print_indent();
				fprintf(file_c, 
					"fprintf(stderr, \" on/before line %d\\n\");\n",
					linenumber);
	
				undent;
				print_indent();
				fprintf(file_c, "}\n");
			}

			myfree(expr.addr);
			if (!assign_expr->array ||
			    (assign_expr->array && multiple_expr) ||
			    (assign_expr->array && !multiple_expr && 
				index == assign_expr->size))
				mark++;
		} while (mark <= top && !loop_assign);

		/* Close off the MAX_N for loop. */
		if (loop_assign) {
			undent;
			print_indent();
			fprintf(file_c, "}\n");
		}

		/* Clean up the semantic stack. */
		while (stack[top--].kind != StackMark);
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr
      LEAVES  - n/a
      ACTION  - Removes the assignable which is on top of the semantic
		stack, since it is no longer needed.  No code is written.
*/
void end_assign(void) {
	if (!stack_error(1)) {
		expr_entry entry = stack[top].data.expr;

		if (entry.var_index_tname) myfree(entry.var_index_tname);
		if (entry.field_tname) myfree(entry.field_tname);
		if (entry.field_index_tname) myfree(entry.field_index_tname);
		myfree(entry.addr);
	}
	top--;
}

/************************************************************************/

/* This variable is used to determine if the left-hand side of an assignment
   is "cell".  It is assigned the value of true in assignable_base, if the
   base is "cell".  The semantic actions field_reference, field_array_index,
   and field_array_reference will issue an error if the cell field is not
   assignable (i.e., a constant field) and the base of the left-hand side
   is "cell".  The semantic action end_lhs resets cell_lhs to false.
*/
boolean cell_lhs = false;

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Resets cell_lhs to false.
*/
void end_lhs(void) {
	cell_lhs = false;
}

/* SEMANTIC ROUTINE
      EXPECTS - StackToken
      LEAVES  - StackExpr
      ACTION  - Converts the entry on top of the stack into an assignable
		base entry.
*/
void assignable_base(void) {
	char *token = stack[top].data.token;

	/* Prepare to determine if the future assignment might try to assign
	   to a constant cell field.
	*/
	if (strcmp(token, "cell") == 0) cell_lhs = true;

	/* Convert StackToken entry into a StackExpr entry. */
	stack[top].kind = StackExpr;
	stack[top].data.expr.assignable = true;
	stack[top].data.expr.field = (char*) NULL;
	stack[top].data.expr.array = false;
	stack[top].data.expr.free_array = false;
	stack[top].data.expr.free_field = false;
	stack[top].data.expr.index = false;
	stack[top].data.expr.attr_list = NULL;
	stack[top].data.expr.var_tname = token;
	stack[top].data.expr.var_index_tname = NULL;
	stack[top].data.expr.field_tname = NULL;
	stack[top].data.expr.field_index_tname = NULL;

	/* Make sure the identifier hasn't previously been declared as a
	   constant.
	*/
	if (intable(token, CONSTANT)) {
		error("Cannot change value of constant '%s'", token);
		stack[top].data.expr.pre_declared = true;
	} else 
		stack[top].data.expr.pre_declared = intable(token, VARIABLE);

	if (stack[top].data.expr.pre_declared) {
		attr_rec *attr = get_attribute(token, VARIABLE);

		if (attr == NULL) attr = get_attribute(token, CONSTANT);

		stack[top].data.expr.unstructured = !attr->structured;
		stack[top].data.expr.assignable = attr->assignable;
		stack[top].data.expr.index = attr->index;
		stack[top].data.expr.agent = attr->agent;
		stack[top].data.expr.array = attr->array;
		stack[top].data.expr.size = attr->size;

		stack[top].data.expr.addr =
				mymalloc((unsigned int) 
						 (strlen(attr->address) + 1));
		strcpy(stack[top].data.expr.addr, attr->address);
	} else
		stack[top].data.expr.addr = token;
}

/************************************************************************/

/* This variable keeps track of the nesting of array references.  The front
   of the chain always points to the attribute records of the inner-most
   array reference.
*/
typedef struct array_nesting_fubar {
	attr_rec *attr;
	struct array_nesting_fubar *next;
} array_nesting_type;

array_nesting_type *array_nesting = NULL;

/* SUPPORT ROUTINE
      ACTION - Pushes an array attribute onto the array_nesting list, assuming
	       the attribute describes an array.
*/
void push_array(attr_rec *attr) {
	array_nesting_type *tmp = (array_nesting_type*)
			mymalloc(sizeof(array_nesting_type));
	tmp->next = array_nesting;
	tmp->attr = attr;
	array_nesting = tmp;
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr
      LEAVES  - StackExpr
      ACTION  - Confirms that the StackExpr describes a simple variable, i.e.
		NOT an array.
*/
void confirm_simple(void) {
	if (!stack_error(1)) {
		if (stack[top].data.expr.pre_declared &&
						stack[top].data.expr.array)
			error("Array variable '%s' improperly used",
				stack[top].data.expr.addr);
		stack[top].data.expr.array = false;
		stack[top].data.expr.free_array = false;
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr
      LEAVES  - StackExpr
      ACTION  - Confirms that the StackExpr describes an array.
*/
void confirm_array(boolean free_index) {
	if (!stack_error(1)) {
		if (!stack[top].data.expr.pre_declared && !free_index) {
			error("Array '%s' used before declared",
				stack[top].data.expr.addr);
			stack[top].kind = StackError;
			return;
		}

		if (stack[top].data.expr.pre_declared &&
						!stack[top].data.expr.array)
			error("Variable '%s' improperly used as an array",
				&stack[top].data.expr.addr[1]);

		stack[top].data.expr.array = true;
		stack[top].data.expr.free_array = free_index;

		if (!free_index) {
			if (intable(stack[top].data.expr.addr, VARIABLE))
				push_array(get_attribute(
						stack[top].data.expr.addr,
						VARIABLE));
			else
				push_array((attr_rec*) NULL);
		}
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr, StackExpr (array)
      LEAVES  - StackExpr (array)
      ACTION  - Declares the array by setting its size to the numeric that
		is on top of the stack and placing the necessary entry into
		the symbol table.
*/
void declare_array_variable(void) {
	char *array_name = NULL;
	attr_rec_list *new_list;
	int element_count = atoi(stack[top--].data.expr.addr);
	expr_entry *expr;

	/* This is the StackExpr entry built by assignable_base. */
	expr = &stack[top].data.expr;

	array_name = expr->addr;

	/* Make sure that the array will contain at least 1 element. */
	if (element_count < 1) {
		error_prefix();
		fprintf(stderr, 
			"Warning; the array, '%s', was declared with size %d, assuming 1\n", 
			array_name, element_count);
		element_count = 1;
	}

	/* Enter the array_name into symbol table. */
	if (!intable(array_name, VARIABLE)) {
		attr_rec attribute;
		attribute.assignable = true;
		attribute.is_float = use_floats;
		attribute.field_name = VARIABLE;
		attribute.array = true;
		attribute.size = element_count;
		attribute.index = false;
		attribute.address = 
			mymalloc((unsigned int) (strlen(array_name) + 1));
		strcpy(attribute.address, array_name);
		enterid(array_name, attribute);

		/* Remember attribute record. */
		new_list = (attr_rec_list*) 
				mymalloc((unsigned int) 
					 (sizeof(attr_rec_list)));
		new_list->next = NULL;
		new_list->attribute = get_attribute(array_name, VARIABLE);
		expr->attr_list = new_list;
	} else
		error("Redeclaration of variable, '%s'", array_name);

	/* Indicate that the assignable is an array. */
	stack[top].data.expr.array = true;
	stack[top].data.expr.size = element_count;
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr
      LEAVES  - n/a
      ACTION  - Writes code to conditionally perform a block of statements.
		If microvectors are used, no code is generated, but a guard
		is built and placed at the front of the mv_guard list.
*/
void jump_if(void) {
	if (stack_error(1))
		top--;
	else {
		if (!stack[top].data.expr.unstructured)
			error("Conditionals must be unstructured", (char*) NULL);
		else if (mv_bits > 0) {
			mv_guard *tmp = (mv_guard*) mymalloc(sizeof(mv_guard));
			unsigned int size = strlen(stack[top].data.expr.addr)+1;

			tmp->next = mv_guards;
			tmp->true_guard =
				mymalloc(strlen(stack[top].data.expr.addr) + 1);
			strcpy(tmp->true_guard, stack[top].data.expr.addr);
			tmp->false_guard = (char*) NULL;

			mv_guards = tmp;
		} else {
			print_indent();
			fprintf(file_c, "if (%s) {\n", stack[top].data.expr.addr);
			indent;
		}
		myfree(stack[top--].data.expr.addr);
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr
      LEAVES  - n/a
      ACTION  - Writes code to conditionally perform a block of statements 
		as an alternative to previous if's and elsif's.  If 
		microvectors are used, no code is generated, but the guards
		at the front of the mv_guard list must be updated.
*/
void jump_elsif(void) {
	if (stack_error(1))
		top--;
	else {
		if (!stack[top].data.expr.unstructured)
			error("Conditionals must be unstructured", (char*) NULL);
		else if (mv_bits > 0) {
			/* Update the false_guard by ORing the true_guard
			   with any existing false_guard.
			*/
			if (mv_guards->false_guard) {
				char string[MAX_STRING_LENGTH], *tmp;
				unsigned int size = 1;

				sprintf(string, "%s(,,%d)",
					(mv_signed ? "UVO_OR" : "UV_OR"),
					mv_bits);
				size += strlen(mv_guards->true_guard)
					+ strlen(mv_guards->false_guard)
					+ strlen(string);

				tmp = mymalloc(size);

				sprintf(tmp, "%s(%s,%s,%d)",
					(mv_signed ? "UVO_OR" : "UV_OR"),
					mv_guards->true_guard,
					mv_guards->false_guard, mv_bits);

				myfree(mv_guards->true_guard);
				myfree(mv_guards->false_guard);
				mv_guards->false_guard = tmp;
			} else {
				mv_guards->false_guard = mv_guards->true_guard;
			}

			/* Make the top of the stack the new true_guard. */
			mv_guards->true_guard =
				mymalloc(strlen(stack[top].data.expr.addr) + 1);
			strcpy(mv_guards->true_guard, stack[top].data.expr.addr);
		} else {
			undent;
			print_indent();
			fprintf(file_c, 
				"} else if (%s) {\n", stack[top].data.expr.addr);
			indent;
		}
		myfree(stack[top--].data.expr.addr);
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Writes code to unconditionally perform a block of 
		statements as an alternative to previous if's and elsif's.
		No code is generated if microvectors are being used, but
		the guards at the front of the mv_guard list must be updated. 
*/
void start_else(void) {
	if (mv_bits > 0) {
		if (mv_guards->false_guard) {
			char string[MAX_STRING_LENGTH];
			char *tmp;
			unsigned int size = 1;

			sprintf(string, "%s(,,%d)",
				(mv_signed ? "UVO_OR" : "UV_OR"), mv_bits);
			size += strlen(mv_guards->true_guard)
				+ strlen(mv_guards->false_guard)
				+ strlen(string);

			tmp = mymalloc(size);
			sprintf(tmp, "%s(%s,%s,%d)",
				(mv_signed ? "UVO_OR" : "UV_OR"),
				mv_guards->true_guard,
				mv_guards->false_guard,
				mv_bits);

			myfree(mv_guards->true_guard);
			myfree(mv_guards->false_guard);
			mv_guards->false_guard = tmp;
		} else
			mv_guards->false_guard = mv_guards->true_guard;

		mv_guards->true_guard = (char*) NULL;
	} else {
		undent;
		print_indent();
		fprintf(file_c, "} else {\n");
		indent;
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Writes code to close off an if statement, but only if
		microvectors are NOT being used.  If microvectors are
		being used, then the front of the mv_guards list should
		be removed.
*/
void end_if(void) {
	if (mv_bits) {
		mv_guard *tmp = mv_guards;
		mv_guards = mv_guards->next;
		if (tmp->true_guard) myfree(tmp->true_guard);
		if (tmp->false_guard) myfree(tmp->false_guard);
		myfree(tmp);
	} else {
		undent;
		print_indent();
		fprintf(file_c, "}\n");
	}
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - StackMark
      ACTION  - Pushes the StackMark onto the semantic stack.
*/
void push_mark(void) { 
	new_entry();
	stack[top].kind = StackMark; 
}

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - StackToken
      ACTION  - Pushes a string the stack.
*/
void push(char *string) {
	new_entry();
	stack[top].kind = StackToken;
	stack[top].data.token = mymalloc((unsigned int) (strlen(string) + 1));
	strcpy(stack[top].data.token, string);
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - StackExpr
      ACTION  - Pushes a StackExpr onto the stack that represents the
		value of the expression denoted by the literal.
*/
void process_literal(char *literal) {

	new_entry();

	if (nested_arrays > 0 && NumericIsFloat) {
		/* Floating point values may not be used in computing
		   array indices.
		*/
		error("Float values cannot be used within an array index.",
			(char*) NULL);
		stack[top].kind = StackError;
	} else {
		unsigned int size = strlen(literal) + 1;
		expr_entry expr;
		expr.constant = true;
		expr.assignable = false;
		expr.unstructured = true;
		expr.agent = false;
		expr.array = false;
		expr.index = false;
		expr.lower = 0;
		expr.vector = false;

		expr.addr = mymalloc(size);

		sprintf(expr.addr, "%s", literal);
		stack[top].kind = StackExpr;
		stack[top].data.expr = expr;
	}
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr
      LEAVES  - StackExpr
      ACTION  - Converts the StackExpr (a constant) on top of the stack
		into a micro-vector which represents the same value.
*/
void convert_to_uvector(void) {
	if (mv_bits > 0) {
		unsigned int size = 1;
		char string[MAX_STRING_LENGTH];
		char *old_value = stack[top].data.expr.addr;

		sprintf(string, "%d", mv_bits);
		size += strlen("UV_S2VEC(,)") + strlen(string);
		if (mv_signed) size += strlen("O");

		stack[top].data.expr.vector = true;

		stack[top].data.expr.addr = mymalloc(size);

		sprintf(stack[top].data.expr.addr, "%s(%s,%d)",
			(mv_signed ? "UVO_S2VEC" : "UV_S2VEC"),
			old_value, mv_bits);

		myfree(old_value);
	}
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - Stack1, Stack2
      LEAVES  - Stack2, Stack1
      ACTION  - Simply swaps the top two entries on the semantic stack.
		This allows some reorder of semantic entries which enables
		the parser to be written as an LL(1) grammar.
*/
void swap_top_2(void) { 
	stack_entry top_entry;
	top_entry = stack[top];
	stack[top] = stack[top-1];
	stack[top-1] = top_entry;
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr, StackToken
      LEAVES  - StackExpr
      ACTION  - Performs a simplified unary_op operation, building an
		integer value without () or spaces.  This allows the
		routine cell_reference to do an atoi call to get the
		value of the integer and to compare it against the size
		of the range for a dimension.
*/
void mk_signed_int(void) {
	stack_entry entry_1, entry_2;
	entry_2 = stack[top--];
	entry_1 = stack[top];
	stack[top].kind = StackExpr;
	stack[top].data.expr.assignable = false;
	stack[top].data.expr.constant = true;
	stack[top].data.expr.unstructured = false;
	stack[top].data.expr.agent = false;
	stack[top].data.expr.array = false;
	stack[top].data.expr.index = false;
	stack[top].data.expr.vector = false;
	if (strcmp(entry_1.data.token, "-") == 0) {
		if (entry_2.data.token[0] == '-') {
			unsigned int size = strlen(entry_2.data.token);

			stack[top].data.expr.addr = mymalloc(size);

			sprintf(stack[top].data.expr.addr, 
				"%s", &entry_2.data.expr.addr[1]);
		} else {
			unsigned int size = strlen(entry_2.data.token) + 2;

			stack[top].data.expr.addr = mymalloc(size);

			sprintf(stack[top].data.expr.addr, 
				"-%s", entry_2.data.expr.addr);
		}
		myfree(entry_2.data.expr.addr);
	} else
		stack[top].data.expr.addr = entry_2.data.expr.addr;
	myfree(entry_1.data.token);
}

/************************************************************************/

/* SUPPORT ROUTINE
      ACTION - Given a pointer to the attribute record for an index variable,
	       this routine checks to see if the parser is currently indexing
	       an array.  If so, the range of the index variable is inferred
	       and either set or checked against it's current value.
*/
void check_set_index_range(attr_rec *index_attr) {

	int size = -1;

	/* Agent index variables have no range to infer. */
	if (index_attr->agent) return;

	/* Try to infer size. */
	if (array_nesting) size = array_nesting->attr->size;

	/* If a size was inferred, then check and/or set the range of
	   the index variable.
	*/
	if (size >= 0)
		if (index_attr->range_set && index_attr->size != size)
			index_attr->incomp_range = true;
		else if (!index_attr->range_set) {
			index_attr->range_set = true;
			index_attr->size = size;
			index_attr->lower = 0;
		}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackToken
      LEAVES  - StackExpr
      ACTION  - Checks to insure that the token has been declared as
		a variable.  If not an error is generated, if so then
		the entry is converted into an expression.
*/
void get_variable(void) {
	char *string = stack[top].data.token;
	if (!intable(string, VARIABLE)) {
		error("Variable '%s' used before set", string); 
		stack_clear(1);
		stack[top].kind = StackError;  
	} else {
		attr_rec *attr = get_attribute(string, VARIABLE);
		unsigned int size = 2;

		/* Index variables used as identifiers should have a user
		   defined range specified to reduce the likelihood for an
		   error.  Since the range can be infered, print a warning
		   if the index is being used outside the context of an
		   array reference.
		*/
		if (attr->index && !attr->agent && !attr->usr_def && !array_nesting) {
			error_prefix();
			fprintf(stderr, 
				"Warning; the range of index variable '%s' starts at 0\n", 
				string);
		}

		/* Do the conversion into an expression. */
		stack[top].kind = StackExpr;  
		stack[top].data.expr.assignable = false; 
		stack[top].data.expr.constant = false; 
		stack[top].data.expr.pre_declared = true; 
		stack[top].data.expr.unstructured = !attr->structured;
		stack[top].data.expr.agent = attr->agent;
		stack[top].data.expr.field_index_tname = NULL;
		stack[top].data.expr.array = attr->array;
		stack[top].data.expr.index = attr->index;
		stack[top].data.expr.size = attr->size;
		stack[top].data.expr.vector = mv_bits > 0;

		if (attr->index) {
			check_set_index_range(attr);
			if (attr->agent) stack[top].data.expr.attribute = attr;
			stack[top].data.expr.lower = attr->lower;
		}
		if (attr->array || (attr->index && !attr->agent))
			stack[top].data.expr.size = attr->size;

		if (attr->array) push_array(attr);

		size += strlen(attr->address);

		if (attr->agent && attr->index) size += strlen("(*)");

		stack[top].data.expr.addr = mymalloc(size);

		/* The variable "cell" is a special case. */
		if (strcmp(string, "cell") == 0) {
			int i,j ;
			char dims[MAX_STRING_LENGTH];
			char tmp[MAX_STRING_LENGTH];

			strcpy(dims, "[");
			for (i = 1; i < dimension_count; i++) {
				sprintf(tmp, "+dim_%d", i);
				strcat(dims, tmp);
				for (j = dimension_count; j > i; j--) {
					sprintf(tmp, "*(%d+MAX_OFFSET*2)", dim_size[j-1]);
					strcat(dims, tmp);
				}
			}
			sprintf(tmp, "+dim_%d]", dimension_count);
			strcat(dims, tmp);

			myfree(stack[top].data.expr.addr);
			stack[top].data.expr.addr = mymalloc(strlen(dims)+4);

			sprintf(stack[top].data.expr.addr, "now%s", dims);

		/* The variable "random" is a special case.

		   NOTE:  Although random is suppose to produce a different
			  value each time it is used, it's use in microvectors
			  means that this is no longer done, but rather that
			  one random value is used by many different cells.
		*/
		} else if (strcmp(string, "random") == 0) {
			myfree(stack[top].data.expr.addr);
			if (mv_bits > 0) {
				char string[MAX_STRING_LENGTH];
				if (mv_signed)
					sprintf(string,
						"UVO_S2VEC(myrandom()&(BIT_MASK>>1),%d)",
						mv_bits);  
				else
					sprintf(string,
						"UV_S2VEC(myrandom()&BIT_MASK,%d)",
						mv_bits);  
				stack[top].data.expr.addr =
					mymalloc(strlen(string)+1);
				sprintf(stack[top].data.expr.addr, string);
			} else {
				stack[top].data.expr.addr =
					mymalloc(strlen("myrandom()")+1);
				sprintf(stack[top].data.expr.addr, "myrandom()");  
			}

		} else if (attr->agent && attr->index)
			sprintf(stack[top].data.expr.addr, 
				"(*_%s)", attr->address);  
		else
			sprintf(stack[top].data.expr.addr, "_%s", attr->address);  

		if (mv_bits > 0 && strcmp(string, "time") == 0)
			error("Variable `time' incompatible with microvectors",
			      (char*) NULL);

		myfree(string);
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackToken
      LEAVES  - StackExpr
      ACTION  - Checks to insure that the token has been declared as
		a constant.  If not an error is generated, if so then
		the entry is converted into an expression with the value
		of the simple constant.
*/
void get_simple_const(void) {
	char *string = stack[top].data.token;
	if (!intable(string, CONSTANT)) {
		error("Constant '%s' used before set", string); 
		stack_clear(1);
		stack[top].kind = StackError;  
	} else {
		attr_rec *attr = get_attribute(string, CONSTANT);
		myfree(string);

		if (attr->array)
			error("The constant array '%s' cannot be used here",
				string); 

		if (use_floats && nested_arrays > 0 && attr->is_float)
			error("The float constant '%s' cannot be used within an array index",
				string);

		stack[top].kind = StackExpr;
		stack[top].data.expr.constant = true;
		stack[top].data.expr.assignable = false;
		stack[top].data.expr.index = false;
		stack[top].data.expr.addr = mymalloc(strlen(attr->address)+1);
		sprintf(stack[top].data.expr.addr, "%s", attr->address);
	}
}

/* SUPPORT ROUTINE
      ACTION - Checks to insure that the StackToken on top of the stack
	       has been declared as an (index) variable.  If not an error
	       is generated, if so then the entry is converted into an
	       expression.
*/
void get_index(void) {
	char *string = stack[top].data.token;
	if (!intable(string, VARIABLE)) {
		error("Unknown index variable '%s'", string); 
		myfree(stack[top].data.token);
		stack[top].kind = StackError;  
	} else {
		attr_rec *attr = get_attribute(string, VARIABLE);
		unsigned int size = 2;

		/* Do the conversion into an expression. */
		stack[top].kind = StackExpr;  
		stack[top].data.expr.assignable = false; 
		stack[top].data.expr.constant = false;
		stack[top].data.expr.unstructured = !attr->structured;
		stack[top].data.expr.agent = !attr->agent;
		stack[top].data.expr.array = false;
		stack[top].data.expr.attribute = attr;
		stack[top].data.expr.index = attr->index;
		stack[top].data.expr.vector = mv_bits > 0;

		if (attr->index) {
			/* Shouldn't be an agent index variable. */
			if (attr->agent) {
				error("Improper use of agent index variable '%s'", 
					string); 
				myfree(stack[top].data.token);
				stack[top].kind = StackError;  
				return;
			}
			
			stack[top].data.expr.lower = attr->lower;
			stack[top].data.expr.size = attr->size;
		} else {
			error("Unknown index variable '%s'", string); 
			myfree(stack[top].data.token);
			stack[top].kind = StackError;  
			return;
		}

		size += strlen(attr->address);

		stack[top].data.expr.addr = mymalloc(size);

		sprintf(stack[top].data.expr.addr, "_%s", attr->address);  

		myfree(string);

		check_set_index_range(attr);
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackToken
      LEAVES  - StackExpr
      ACTION  - If the StackToken names a constant, call get_simple_const
		to do the constant processing.  If the token is a variable,
		then call get_index to process the token.  If the token is
		neither a constant nor a variable, generate an error.
*/
void get_simple_const_or_index() {
	char *string = stack[top].data.token;
	if (intable(string, CONSTANT)) {
		get_simple_const();
		if (mv_bits > 0) convert_to_uvector();
	} else if (intable(string, VARIABLE))
		get_index();
	else
		error("'%s' must be a constant or an index variable",
			string);
}

/* SEMANTIC ROUTINE
      EXPECTS - StackToken
      LEAVES  - StackExpr
      ACTION  - If the StackToken names a constant, get the constant (which
		may be either simple or an array) If the token is a variable,
		then call get_variable to process the token.  If the token is
		neither a constant nor a variable, generate an error.
*/
void get_const_or_variable() {
	char *string = stack[top].data.token;
	if (intable(string, CONSTANT)) {
		attr_rec *attr = get_attribute(string, CONSTANT);

		if (use_floats && nested_arrays > 0 && attr->is_float)
			error("The float constant '%s' cannot be used within an array index",
				string);

		/* Do the conversion into an expression. */
		stack[top].kind = StackExpr;  
		stack[top].data.expr.assignable = attr->assignable; 
		stack[top].data.expr.constant = true;
		stack[top].data.expr.unstructured = !attr->structured;
		stack[top].data.expr.agent = attr->agent;
		stack[top].data.expr.array = attr->array;
		stack[top].data.expr.size = attr->size;
		stack[top].data.expr.attribute = attr;
		stack[top].data.expr.index = attr->index;
		stack[top].data.expr.vector = mv_bits > 0;

		if (attr->array) push_array(attr);

		stack[top].data.expr.addr = mymalloc(strlen(attr->address)+1);
		strcpy(stack[top].data.expr.addr, attr->address);

		myfree(string);

		if (mv_bits > 0) convert_to_uvector();

	} else if (intable(string, VARIABLE))
		get_variable();
	else
		error("'%s' must be a constant or variable array",
			string);
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr, StackExpr
      LEAVES  - StackExpr
      ACTION  - Makes sure that the top of the stack is an unstructured
		value (i.e. array index) and that the next item down on
		the semantic stack describes an array.  The range of the
		index variable is inferred and set if it is currently
		unknown.  The array address entry is also updated.
		NOTE: Array must have been previously declared.
*/
void do_array_index(void) {
	array_nesting_type *free_this;
	if (stack_error(2)) {
		stack_clear(2);
                top--;
                stack[top].kind = StackError;
        } else {
		char const_value[MAX_STRING_LENGTH];
		expr_entry *index = &stack[top--].data.expr;
		expr_entry *array_expr = &stack[top].data.expr;

		if (mv_bits > 0)
			error("Microvectors cannot be used with indexed arrays",
				(char*) NULL);

		if (index->constant) {
			if (mv_bits > 0 && index->vector)
				if (mv_signed)
					sscanf(index->addr,
					       "UVO_S2VEC(%d,", &const_value);
				else
					sscanf(index->addr,
					       "UV_S2VEC(%d,", &const_value);
			else
				strcpy(const_value, index->addr);
		}

		/* Make sure an array is being indexed. */
		if (!array_expr->array) {
			error("Cannot index a non-array", (char*) NULL);
			stack[top].kind = StackError;
		}
		/* Make sure the range of the index is the same size as 
		   the array.
		*/
		else if (index->index && index->size != array_expr->size)
			stack[top].kind = StackError;
		else if (index->constant && 
			 (atoi(const_value) < 0 ||
			  atoi(const_value) >= array_expr->size)) {
			error("Integer index value outside the array range",
				(char*) NULL);
			stack[top].kind = StackError;
		} else {
			char *new_address;
			char string[MAX_STRING_LENGTH];
			unsigned int size = strlen("[]") + 1 +
						strlen(array_expr->addr) +
						strlen(index->addr);

			if (!index->constant) {
				sprintf(string, "RNG_CHK(,%d)",
					array_expr->size-1);
				size += strlen(string);
			}

			if (index->index && index->lower != 0) {
				sprintf(string, "-%d", index->lower);
				size += strlen(string);
			}

			if (mv_bits > 0 && index->vector) {
				sprintf(string, "%s(,%d)",
					(mv_signed ? "UVO_FSVAL" : "UV_FVAL"),
					mv_bits);
				size += strlen(string);
			}

			new_address = mymalloc(size);

			if (index->constant && index->lower != 0)
				sprintf(new_address, "%s[%s-%d]",
					array_expr->addr, const_value,
					index->lower);
			else if (index->constant)
				sprintf(new_address, "%s[%s]",
					array_expr->addr, const_value);
			else if (mv_bits > 0 && index->vector) {
				if (index->index && index->lower != 0)
					sprintf(new_address,
						"%s[RNG_CHK(%s(%s,%d)-%d,%d)]",
						array_expr->addr,
						(mv_signed ? "UVO_FSVAL" : "UV_FVAL"),
						index->addr, mv_bits,
						index->lower,
						array_expr->size-1);
				else
					sprintf(new_address,
						"%s[RNG_CHK(%s(%s,%d),%d)]",
						array_expr->addr,
						(mv_signed ? "UVO_FSVAL" : "UV_FVAL"),
						index->addr, mv_bits,
						array_expr->size-1);
			} else if (index->index && index->lower != 0)
				sprintf(new_address, "%s[RNG_CHK(%s-%d,%d)]",
					array_expr->addr, index->addr,
					index->lower, array_expr->size-1);
			else
				sprintf(new_address, "%s[RNG_CHK(%s,%d)]",
					array_expr->addr, index->addr,
					array_expr->size-1);

			myfree(array_expr->addr);

			array_expr->addr = new_address;
			if (index->index)
				array_expr->var_index_tname = index->addr;
		}

		/* Setup up description of the array element type. */
		array_expr->array = false;
	}
	if (array_nesting) {
		free_this = array_nesting;
		array_nesting = array_nesting->next;
		myfree((char*) free_this);
	}
}

/************************************************************************/

/* SUPPORT ROUTINE
      ACTION - Return the value of 2 raised to the nth power.  N is assumed
	       to be >= 0.
*/
int pow_2(int n) {
	int result = 1;
	while (n-- > 0)
		result *= 2;
	return result;
}

/* SUPPORT ROUTINE
      ACTION - Determine whether the number (> 1) is a power of 2.  The
	       value returned is the power used.
*/
int pow_of_2(int number) {
	int power = 0;
	while (number > 1)
		if (number & 1) return 0;
		else {
			number = number >> 1;
			power++;
		}
	return power;
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr
      LEAVES  - n/a
      ACTION  - Expects the StackExpr to describe an agent variable.  A
		copy of which is created and a reference to the newly
		created copy is in the generated C code variable "tmp_agent".
*/
void copy_agent(void) {
	if (stack_error(1)) {
		stack_clear(1);
		top--;
	} else {
		int i;
		char *agent = stack[top].data.expr.addr;

		/* Make sure this is an agent. */
		if (!stack[top].data.expr.agent || 
					stack[top].data.expr.unstructured)
			error("'%s' is not an agent", agent);
		else if (stack[top].data.expr.array)
			error("Unindexed array reference, '%s', should be an agent", 
				agent);
		else {

			/* Write code to create the new agent. */
			print_indent();
			fprintf(file_c, "create_agent(");
			for (i = cell_field_count; i < field_count; i++) {
				if (agent[0] == '_')
					/* Must be an indexed array. */
					fprintf(file_c, "%s[%s]->%s", agent, 
						field_list[i].name);
				else if (stack[top].data.expr.index)
					fprintf(file_c, "_%s->%s", 
						agent, field_list[i].name);
				else
					fprintf(file_c, "_%s.%s", 
						agent, field_list[i].name);
				if (i < field_count - 1) fprintf(file_c, ",");
			}
			fprintf(file_c, ");\n");
		}

		myfree(agent);
		top--;
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr, ..., StackExpr, StackMark
      LEAVES  - n/a
      ACTION  - Expects the StackExprs to describe the field values to
		assign to a newly created agent.  A reference to the newly
		created agent is in the generated C code variable "tmp_agent".
*/
void create_agent(void) {
	int index = top;
	boolean expr_error = false;

	/* Make sure that agents were declared for this program. */
	if (!agent_declared)
		error("No agents declared for this program", (char*) NULL);

	/* Look for bad stack entries while locating the StackMark. */
	while (stack[index].kind != StackMark)
		if (stack[index--].kind == StackError) expr_error = true;
	index++;

	/* Make sure the correct number of expressions was given. */
	if (top-index+1 < agent_field_count) {
		expr_error = true;
		error("Too few expressions given for agent creation", 
			(char*) NULL);
	} else if (top-index+1 > agent_field_count) {
		expr_error = true;
		error("Too many expressions given for agent creation", 
			(char*) NULL);
	}

	/* In the case of an error, just clean up the stack and return. */
	if (expr_error) {
		while (stack[top].kind != StackMark) {
			if (stack[top].kind != StackError)
				myfree(stack[top].data.expr.addr);
			top--;
		}
		top--;
		return;
	}

	/* Write code to create the new agent. */
	print_indent();
	fprintf(file_c, "create_agent(");
	while (index < top) {
		/* Make sure the expressions are unstructured. */
		if (!stack[index].data.expr.unstructured) {
			char msg[MAX_STRING_LENGTH];
			sprintf(msg, 
				"Expression number %d is not a simple value",
				index - top + agent_field_count);
			error(msg, (char*) NULL);
		} else
			fprintf(file_c, "%s,", stack[index].data.expr.addr);
		index++;
	}
	fprintf(file_c, "%s);\n", stack[index].data.expr.addr);

	/* Clean up the semantic stack. */
	while (stack[top].kind != StackMark)
		myfree(stack[top--].data.expr.addr);
	top--;
}

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Generates code to return the dynamically allocated memory
		of an agent.  This is necessary when a conditional agent
		movement didn't have a default action.  Thus this prevents
		memory leaks in the generated code.  The agent to delete is
		referenced by the variable "tmp_agent".
*/
void free_agent(void) {
	print_indent();
	fprintf(file_c, "free(tmp_agent);\n");
}

/* Indicates whether or not to build the components necessary to print a
   cell location as part of the trace of an agent statement involving a
   relative index.
*/
boolean relative_index_trace = false;

/* Remembers the generated code fragments used in determining the relative
   index of an associated cell for an agent statement.
*/
char *trace_rel_indx[MAX_SUPPORTED_DIMENSIONS];

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Expects the StackExpr at the very top of the stack to 
		describe a cell location at which to place the agent,
		which is second on the stack.  This action simply sets
		up to print a trace of the result at run-time, if desired.
*/
void trace_relative_index(void) {
	relative_index_trace = true;
}

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr(location)
      LEAVES  - n/a
      ACTION  - Expects the StackExpr at the very top of the stack to 
		describe a cell location at which to place the agent,
		which is expected to be pointed to by "tmp_agent".
*/
void put_agent(void) {
	if (stack_error(1)) {
		stack_clear(1);
		top--;
	} else {

		/* Make sure that agents were declared for this program. */
		if (!agent_declared)
			error("No agents declared for this program",
				(char*) NULL);

		/* Write code to put the agent at the desired cell. */
		print_indent();
		fprintf(file_c, "tmp_agent->agent = agent_%s;\n",
			stack[top].data.expr.addr);
		print_indent();
		fprintf(file_c, "agent_%s = tmp_agent;\n",
			stack[top].data.expr.addr);

		/* If the trace option was specified, then generate the code 
		   to print the agent statement result for this cell.
		*/
		if (trace_count > 0) {
			int i;

			print_indent();
			fprintf(file_c, 
				"if ((_time_step == next_report_time || _time_step == next_read_time)");
			for (i = 0; i < trace_count; i++)
				fprintf(file_c, " && dim_%d == %d",
					i + 1, trace_point[i]);
			fprintf(file_c, ") {\n");

			indent;

			/* Generate code to write out a readable agent value. */
			print_indent();
			fprintf(file_c, 
				"fprintf(stderr, \"   agent(\");\n");

			print_indent();
			fprintf(file_c, 
				"fprintf(stderr, \"%%d\", tmp_agent->%s);\n",
				field_list[cell_field_count].name);
			for (i = cell_field_count + 1; i < field_count; i++) {
				print_indent();
				fprintf(file_c, 
					"fprintf(stderr, \", %%d\", tmp_agent->%s);\n",
					field_list[i].name);
			}
			print_indent();
			fprintf(file_c, "fprintf(stderr, \") -> \");\n");

			print_indent();
			if (relative_index_trace) {
				fprintf(file_c, "fprintf(stderr, \"[\");\n");
				for (i = 0; i < dimension_count; i++) {
					print_indent();
					if (i > 0)
						fprintf(file_c, 
							"fprintf(stderr, \", %%d\", %s);\n",
							trace_rel_indx[i]);
					else
						fprintf(file_c, 
							"fprintf(stderr, \"%%d\", %s);\n",
							trace_rel_indx[i]);
					myfree(trace_rel_indx[i]);
				}
			} else
				fprintf(file_c, "fprintf(stderr, \"cell\");\n");

			if (relative_index_trace) {
				print_indent();
				fprintf(file_c, "fprintf(stderr, \"]\");\n");
			}

			print_indent();
			fprintf(file_c, 
				"fprintf(stderr, \" on/before line %d\\n\");\n",
				linenumber);

			undent;
			print_indent();
			fprintf(file_c, "}\n");

			relative_index_trace = false;
		}

		/* Clean up the semantic stack. */
		myfree(stack[top--].data.expr.addr);
	}
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr, ..., StackExpr, StackMark
      LEAVES  - StackExpr
      ACTION  - The number of StackExpr before the StackMark must be
		equal to the number of dimensions in the cellular
		universe.  The relative indices (StackExpr's) into the 
		universe are converted into the proper C array reference 
		code which uses these values as offsets from the current
		values of the dimension indices.  No arithmetic is done
		if the relative index is 0.
*/
void cell_reference(void) {
	char *new_addr;
	unsigned int size = strlen("now[") + 1;
	int mark = top;
	int tmp_top;
	int dim_count;

	/* Determine the size of the array reference. */ 
	dim_count = 0;
	while (stack[mark].kind != StackMark) {
		int j;
		char sub_expr[MAX_STRING_LENGTH];

		dim_count++;

		if (!stack[mark].data.expr.index) {
			/* The StackExpr is really an integer, thus the
			   relative cell reference can be optimized for 
			   better performance.
			*/
			int index_value;

			if (mv_bits > 0) {
				int found;
				if (mv_signed)
					found = sscanf(stack[mark].data.expr.addr,
						       "UVO_S2VEC(%d,", &index_value);
				else
					found = sscanf(stack[mark].data.expr.addr,
						       "UV_S2VEC(%d,", &index_value);
				if (found != 1) {
					fprintf(stderr,
						"INTERNAL ERROR: in cell_reference.\n");
					exit(1);
				}
			} else
				index_value = atoi(stack[mark].data.expr.addr);

			mark--;

			if (0 == index_value)
				sprintf(sub_expr, "+dim_%d", dim_count);
			else
				sprintf(sub_expr, "+dim_%d_%s_%d", dim_count, 
					(index_value < 0 ? "minus" : "plus"),
					(index_value < 0 ? -index_value : index_value));
		} else {
			/* The StackExpr is a, possibly shifted, index variable
			   value.  No optimization can be done.
			*/
			if (mv_bits > 0 && stack[mark].data.expr.vector)
				sprintf(sub_expr, "+((%s(%s,%d))+dim_%d)", 
					(mv_signed ? "UVO_FSVAL" : "UV_FVAL"),
					stack[mark].data.expr.addr, mv_bits,
					dim_count);
			else
				sprintf(sub_expr, "+((%s)+dim_%d)", 
					stack[mark].data.expr.addr, dim_count);
			mark--;
		}

		size += strlen(sub_expr);

		for (j = dimension_count; j > dim_count; j--) {
			sprintf(sub_expr, "*(%d+MAX_OFFSET*2)", dim_size[j-1]);
			size += strlen(sub_expr);
		}
	}
	size += strlen("]");

	/* The number of dimensions used should match those declared. */
	if (dimension_count != dim_count)
		error("Incorrect number of dimensions in cell reference",
		      (char*) NULL);
	
	/* Turn the StackMark entry into a StackExpr (the array reference). */
	stack[mark].kind = StackExpr;
	stack[mark].data.expr.assignable = false;
	stack[mark].data.expr.constant = false;
	stack[mark].data.expr.unstructured = 
				!(cell_fields_declared || agent_declared);
	stack[mark].data.expr.agent = false;
	stack[mark].data.expr.array = false;
	stack[mark].data.expr.index = false;
	new_addr = stack[mark].data.expr.addr = mymalloc(size);
	
	/* Build the array reference. */
	dim_count = 0;
	tmp_top = mark + 1;
	strcpy(new_addr, "now[");
	while (tmp_top <= top) {
		int j;
		char sub_expr[MAX_STRING_LENGTH];

		/* Build the array of relative cell indices to be used for
		   tracing an agent statement, if tracing is turned on.
		*/
		if (relative_index_trace) {
			trace_rel_indx[dim_count] = 
				mymalloc(strlen(stack[tmp_top].data.expr.addr)+1);
			strcpy(trace_rel_indx[dim_count], 
				stack[tmp_top].data.expr.addr);
		}

		dim_count++; 

		if (!stack[tmp_top].data.expr.index) {
			/* The StackExpr is really an integer, thus the
			   relative cell reference can be optimized for 
			   better performance.
			*/
			int index_value;

			if (mv_bits > 0) {
				if (mv_signed)
					sscanf(stack[tmp_top].data.expr.addr,
					       "UVO_S2VEC(%d,", &index_value);
				else
					sscanf(stack[tmp_top].data.expr.addr,
					       "UV_S2VEC(%d,", &index_value);
			} else
				index_value = atoi(stack[tmp_top].data.expr.addr);

			if (abs(index_value) > max_rel_index)
				max_rel_index = abs(index_value);

			if (0 == index_value)
				sprintf(sub_expr, "+dim_%d", dim_count);
			else
				sprintf(sub_expr, "+dim_%d_%s_%d", dim_count, 
					(index_value < 0 ? "minus" : "plus"),
					(index_value < 0 ? -index_value : index_value));

			/* If this relative cell referencing index offset has
			   not been previously used/declared, then declare it.
			   Only the name needs to be remembered since the rest
			   can be easily regenerated.
			*/
			if (0 != index_value && !intable(&sub_expr[1], VARIABLE)) {
				char value_expr[MAX_STRING_LENGTH];
				attr_rec attribute;

				attribute.assignable = false;
				attribute.is_float = use_floats;
				attribute.field_name = VARIABLE;
				attribute.structured = false;
				attribute.agent = false;
				attribute.array = false;
				attribute.index = false;
				attribute.address = (char*) NULL;
				enterid(&sub_expr[1], attribute);

				/* Build the "variable" value string. */
				if (index_value > 0)
					sprintf(value_expr, 
						"(dim_%d+%d)", dim_count, 
						index_value);
				else
					sprintf(value_expr, 
						"(dim_%d%d)", dim_count, 
						index_value);

				/* Write code to declare the "variable". */
				{
					char file_name[MAX_STRING_LENGTH];
					FILE *file_dim_h;

					sprintf(file_name, "%s%s%dd%d.h", 
						TMP_DIR, BASE_FILE_NAME, 
						pid, dim_count);
					file_dim_h = fopen(file_name, "a");

					fprintf(file_dim_h, 
						"register int %s = %s;\n", 
						&sub_expr[1], value_expr);

					fflush(file_dim_h);
					fclose(file_dim_h);
				}
			}
		} else {
			/* The StackExpr is a, possibly shifted, index variable
			   value.  No optimization can be done.
			*/
			if (mv_bits > 0 && stack[tmp_top].data.expr.vector)
				sprintf(sub_expr, "+((%s(%s,%d))+dim_%d)", 
					(mv_signed ? "UVO_FSVAL" : "UV_FVAL"),
					stack[tmp_top].data.expr.addr, mv_bits,
					dim_count);
			else
				sprintf(sub_expr, "+((%s)+dim_%d)", 
					stack[tmp_top].data.expr.addr, dim_count);
			stack[tmp_top].data.expr.attribute->cell_ref = true;
		}

		myfree(stack[tmp_top].data.expr.addr);
		strcat(new_addr, sub_expr);
		tmp_top++;

		for (j = dimension_count; j > dim_count; j--) {
			sprintf(sub_expr, "*(%d+MAX_OFFSET*2)", dim_size[j-1]);
			strcat(new_addr, sub_expr);
		}
	}
	strcat(new_addr, "]");

	/* Pop most of the semantic stack. */
	top = mark;
}

/* SUPPORT ROUTINE
      ACTION - Checks to make sure that the top of the stack represents
	       a legal field reference for the variable described below
	       it on the stack.  Returns 1 if okay, 0 if there is a problem.
*/
boolean bad_field(stack_entry base_entry, stack_entry field_entry) {

	/* Make sure this is really a field. */
	if (base_entry.data.expr.pre_declared &&
			base_entry.data.expr.agent && 
			!intable(field_entry.data.token, AGENT_FIELD)) {
		error("'%s' is not a known agent field", 
			field_entry.data.token);
		myfree(field_entry.data.token);
		return true;
	} else if (base_entry.data.expr.pre_declared &&
			!base_entry.data.expr.agent &&
			!intable(field_entry.data.token, CELL_FIELD)) {
		error("'%s' is not a known cell field", 
			field_entry.data.token);
		myfree(field_entry.data.token);
		return true;
	} else if (!base_entry.data.expr.pre_declared &&
			!intable(field_entry.data.token, AGENT_FIELD) &&
			!intable(field_entry.data.token, CELL_FIELD)) {
		error("'%s' is not a known cell field", 
			field_entry.data.token);
		myfree(field_entry.data.token);
		return true;
	} else
		return false;
}

/* SEMANTIC ROUTINE
      EXPECTS - StackToken StackExpr
      LEAVES  - StackExpr
      ACTION  - The StackToken is used to create a field reference of
		the StackExpr.
*/
void field_reference(void) {
	if (stack_error(2)) {
		stack_clear(2);
		top--;
		stack[top].kind = StackError;
	} else {
		stack_entry field_entry, base_entry;
		attr_rec *field_attr;
		
		field_entry = stack[top--];
		base_entry = stack[top];

		/* Make sure this is really a field. */
		if (bad_field(base_entry, field_entry)) {
			stack_clear(1);
			stack[top].kind = StackError;
			return;
		}

		stack[top].kind = StackExpr;
		stack[top].data.expr.assignable = base_entry.data.expr.assignable;
		stack[top].data.expr.unstructured = true;
		stack[top].data.expr.agent = false;

		if (intable(field_entry.data.token, AGENT_FIELD)) {
			base_entry.data.expr.agent = true;
			field_attr = get_attribute(field_entry.data.token,
							AGENT_FIELD);
		} else {
			base_entry.data.expr.agent = false;
			field_attr = get_attribute(field_entry.data.token,
							CELL_FIELD);
		}

		/* If this is a left-hand side with "cell" as the base,
		   make sure the cell field isn't constant.
		*/
		if (cell_lhs && !field_attr->assignable)
			error("Cannot change constant field '%s' of cell",
				field_entry.data.token);

		/* The field should NOT be an array. */
		if (field_attr->array)
			error("The array field '%s' needs []'s",
				field_entry.data.token);

		stack[top].data.expr.array = base_entry.data.expr.array;
		stack[top].data.expr.free_array = base_entry.data.expr.free_array;
		stack[top].data.expr.index = base_entry.data.expr.index;
		stack[top].data.expr.field_tname = field_entry.data.token;

		if (base_entry.data.expr.assignable)
			stack[top].data.expr.field = field_attr->address;
		else if (base_entry.data.expr.agent) {
			unsigned int size = strlen(base_entry.data.expr.addr) +
				          	strlen(field_attr->address) + 1;

			size += strlen("(.)");

			stack[top].data.expr.addr = mymalloc(size);

			sprintf(stack[top].data.expr.addr, "(%s.%s)",
				base_entry.data.expr.addr, field_attr->address);
			myfree(base_entry.data.expr.addr);
		} else {
			unsigned int size;

			size = strlen(base_entry.data.expr.addr) +
			       	strlen(field_attr->address) + strlen("_") + 1;

			if (base_entry.data.expr.addr[0] == '_') size--;

			stack[top].data.expr.addr = mymalloc(size);

			sprintf(stack[top].data.expr.addr, "%s_%s",
				field_attr->address,
				(base_entry.data.expr.addr[0] == '_' ?
					&base_entry.data.expr.addr[1] :
					base_entry.data.expr.addr
				));
			myfree(base_entry.data.expr.addr);
		}
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackToken StackExpr
      LEAVES  - StackExpr
      ACTION  - The StackToken is used to create a field reference of
		the StackExpr, where the field is an unindexed array.
*/
void field_array_reference(void) {
	array_nesting_type *free_this;
	if (stack_error(2)) {
		stack_clear(2);
		top--;
		stack[top].kind = StackError;
	} else {
		stack_entry field_entry, base_entry;
		attr_rec *field_attr;
		
		field_entry = stack[top--];
		base_entry = stack[top];

		/* Make sure this is really a field. */
		if (bad_field(base_entry, field_entry)) {
			stack_clear(1);
			stack[top].kind = StackError;
			return;
		}

		stack[top].kind = StackExpr;
		stack[top].data.expr.assignable = base_entry.data.expr.assignable;
		stack[top].data.expr.unstructured = true;
		stack[top].data.expr.agent = false;

		if (intable(field_entry.data.token, AGENT_FIELD)) {
			base_entry.data.expr.agent = true;
			field_attr = get_attribute(field_entry.data.token,
							AGENT_FIELD);
		} else {
			base_entry.data.expr.agent = false;
			field_attr = get_attribute(field_entry.data.token,
							CELL_FIELD);
		}

		/* If this is a left-hand side with "cell" as the base,
		   make sure the cell field isn't constant.
		*/
		if (cell_lhs && !field_attr->assignable)
			error("Cannot change constant array field '%s' of cell",
				field_entry.data.token);

		/* The field SHOULD be an array. */
		if (!field_attr->array)
			error("The field '%s' is not declared as an array",
				field_entry.data.token);

		/* The base and field canNOT both be arrays. */
		if (base_entry.data.expr.array) {
			char string[MAX_STRING_LENGTH];
			sprintf(string,
				"One of the arrays '%s'/'%s' must be indexed",
				base_entry.data.expr.addr, 
				field_entry.data.token);
			error(string, (char*) NULL);
		}

		stack[top].data.expr.array = true;
		stack[top].data.expr.free_field = true;
		stack[top].data.expr.index = base_entry.data.expr.index;
		stack[top].data.expr.free_array = base_entry.data.expr.free_array;
		stack[top].data.expr.size = field_attr->size;
		stack[top].data.expr.field_tname = field_entry.data.token;

		if (base_entry.data.expr.assignable)
			stack[top].data.expr.field = field_attr->address;
		else if (base_entry.data.expr.agent) {
			unsigned int size = strlen(base_entry.data.expr.addr) +
				          	strlen(field_attr->address) + 1;

			size += strlen("(.)");

			stack[top].data.expr.addr = mymalloc(size);

			sprintf(stack[top].data.expr.addr, "(%s.%s)",
				base_entry.data.expr.addr, field_attr->address);
			myfree(base_entry.data.expr.addr);
		} else {
			unsigned int size;

			size = strlen(base_entry.data.expr.addr) +
			       	strlen(field_attr->address) + strlen("_") + 1;

			if (base_entry.data.expr.addr[0] == '_') size--;

			stack[top].data.expr.addr = mymalloc(size);

			sprintf(stack[top].data.expr.addr, "%s_%s",
				field_attr->address,
				(base_entry.data.expr.addr[0] == '_' ?
					&base_entry.data.expr.addr[1] :
					base_entry.data.expr.addr
				));
			myfree(base_entry.data.expr.addr);
		}
	}
	free_this = array_nesting;
	array_nesting = array_nesting->next;
	myfree((char*) free_this);
}

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr(Index) StackToken StackExpr
      LEAVES  - StackExpr
      ACTION  - The StackExpr(Index) and StackToken are used to create 
		an array field reference of the StackExpr.
*/
void field_array_index(void) {
	array_nesting_type *free_this;
	if (stack_error(3)) {
		stack_clear(3);
		top-=2;
		stack[top].kind = StackError;
	} else {
		stack_entry field_entry, base_entry, index_entry;
		char field_address[MAX_STRING_LENGTH];
		char const_value[MAX_STRING_LENGTH];
		attr_rec *field_attr;

		if (mv_bits > 0)
			error("Microvectors cannot be used with indexed arrays",
				(char*) NULL);
		
		index_entry = stack[top--];
		field_entry = stack[top--];
		base_entry = stack[top];

		/* Make sure this is really a field. */
		if (bad_field(base_entry, field_entry)) {
			stack_clear(1);
			stack[top].kind = StackError;
			return;
		}

		stack[top].kind = StackExpr;
		stack[top].data.expr.assignable = base_entry.data.expr.assignable;
		stack[top].data.expr.unstructured = true;
		stack[top].data.expr.agent = false;

		if (intable(field_entry.data.token, AGENT_FIELD)) {
			base_entry.data.expr.agent = true;
			field_attr = get_attribute(field_entry.data.token,
							AGENT_FIELD);
		} else {
			base_entry.data.expr.agent = false;
			field_attr = get_attribute(field_entry.data.token,
							CELL_FIELD);
		}

		/* If this is a left-hand side with "cell" as the base,
		   make sure the cell field isn't constant.
		*/
		if (cell_lhs && !field_attr->assignable)
			error("Cannot change constant array field '%s' of cell",
				field_entry.data.token);

		/* Make sure this is really an array. */
		if (!field_attr->array)
			error("Field '%s' improperly used as an array",
			      field_entry.data.token);

		stack[top].data.expr.array = base_entry.data.expr.array;
		stack[top].data.expr.free_array = base_entry.data.expr.free_array;
		stack[top].data.expr.index = base_entry.data.expr.index;

		if (index_entry.data.expr.constant) {
			if (mv_bits > 0 && index_entry.data.expr.vector)
				if (mv_signed)
					sscanf(index_entry.data.expr.addr,
					       "UVO_S2VEC(%d,", &const_value);
				else
					sscanf(index_entry.data.expr.addr,
					       "UV_S2VEC(%d,", &const_value);
			else
				strcpy(const_value, index_entry.data.expr.addr);

			if (atoi(const_value) < 0
			    || atoi(const_value) >= field_attr->size)
				error("Integer index value outside the array range",
					(char*) NULL);
		}

		if (index_entry.data.expr.lower != 0) {
			if (index_entry.data.expr.constant)
				sprintf(field_address, "%s[%s-%d]",
					field_attr->address, const_value,
					index_entry.data.expr.lower);
			else if (mv_bits > 0 && index_entry.data.expr.vector)
				sprintf(field_address,
					"%s[RNG_CHK(%s(%s,%d)-%d,%d)]",
					field_attr->address, 
					(mv_signed ? "UVO_FSVAL" : "UV_FVAL"),
					index_entry.data.expr.addr, mv_bits,
					index_entry.data.expr.lower,
					field_attr->size-1);
			else
				sprintf(field_address,
					"%s[RNG_CHK(%s-%d,%d)]",
					field_attr->address, 
					index_entry.data.expr.addr, 
					index_entry.data.expr.lower,
					field_attr->size-1);
		} else {
			if (index_entry.data.expr.constant)
				sprintf(field_address, "%s[%s]",
					field_attr->address, 
					const_value);
			else if (mv_bits > 0 && index_entry.data.expr.vector)
				sprintf(field_address,
					"%s[RNG_CHK(%s(%s,%d),%d)]",
					field_attr->address, 
					(mv_signed ? "UVO_FSVAL" : "UV_FVAL"),
					index_entry.data.expr.addr, mv_bits,
					field_attr->size-1);
			else
				sprintf(field_address,
					"%s[RNG_CHK(%s,%d)]",
					field_attr->address, 
					index_entry.data.expr.addr,
					field_attr->size-1);
		}

		if (base_entry.data.expr.assignable) {
			stack[top].data.expr.field = 
				mymalloc((unsigned int) 
					 (strlen(field_address) + 1));
			strcpy(stack[top].data.expr.field, field_address);
		} else if (base_entry.data.expr.agent) {
			unsigned int size = strlen(base_entry.data.expr.addr) +
				          	strlen(field_attr->address) +
						strlen("(.)") + 1;

			stack[top].data.expr.addr = mymalloc(size);

			sprintf(stack[top].data.expr.addr, "(%s.%s)",
				base_entry.data.expr.addr, field_attr->address);
		} else {
			char *base_name, *base_index;
			char *field_name, *field_index;
			unsigned int size;

			size = strlen(base_entry.data.expr.addr) +
			          	strlen(field_address) + 1;

			if (base_entry.data.expr.addr[0] != '_') size++;

			stack[top].data.expr.addr = mymalloc(size);

			base_name = strtok(base_entry.data.expr.addr, "[");
			base_index = strtok((char*) NULL, "]");

			field_name = strtok(field_address, "[");
			field_index = strtok((char*) NULL, "]");

			if (base_index)
				sprintf(stack[top].data.expr.addr,
					"%s_%s[%s][%s]",
					field_name,
					(base_name[0] == '_' ?
						&base_name[1] : base_name),
					base_index, field_index);
			else
				sprintf(stack[top].data.expr.addr,
					"%s_%s[%s]",
					field_name,
					(base_name[0] == '_' ?
						&base_name[1] : base_name),
					field_index);
			myfree(base_entry.data.expr.addr);
		}

		stack[top].data.expr.field_tname = 
						field_entry.data.token;
		stack[top].data.expr.field_index_tname = 
						index_entry.data.expr.addr;
	}
	free_this = array_nesting;
	array_nesting = array_nesting->next;
	myfree((char*) free_this);
}

/************************************************************************/

/* SUPPORT ROUTINE
      ACTION - Returns the string which names the corresponding microvector
	       operation.
*/
char* which_mv_binary_op(char *token) {
	if (strcmp(token, "+") == 0)
		return (mv_signed ? "UVO_PLUS" : "UV_PLUS");
	else if (strcmp(token, "-") == 0)
		return (mv_signed ? "UVO_MINUS" : "UV_MINUS");
	else if (strcmp(token, "*") == 0)
		if (!mv_signed)
			error("'%s' is only available with signed microvectors",
				token);
		else
			return "UVO_MUL";
	else if (strcmp(token, "/") == 0)
		if (!mv_signed)
			error("'%s' is only available with signed microvectors",
				token);
		else
			return "UVO_DIV";
	else if (strcmp(token, "%") == 0) {
		if (!mv_signed)
			error("'%s' is only available with signed microvectors",
				token);
		else
			return "UVO_MOD";
	} else if (strcmp(token, "&&") == 0)
		return (mv_signed ? "UVO_AND" : "UV_AND");
	else if (strcmp(token, "||") == 0)
		return (mv_signed ? "UVO_OR" : "UV_OR");
	else if (strcmp(token, "<") == 0)
		if (!mv_signed)
			error("'%s' is only available with signed microvectors",
				token);
		else
			return "UVO_LT";
	else if (strcmp(token, "<=") == 0) {
		if (!mv_signed)
			error("'%s' is only available with signed microvectors",
				token);
		else
			return "UVO_LTE";
	} else if (strcmp(token, ">") == 0)
		if (!mv_signed)
			error("'%s' is only available with signed microvectors",
				token);
		else
			return "UVO_GT";
	else if (strcmp(token, ">=") == 0) {
		if (!mv_signed)
			error("'%s' is only available with signed microvectors",
				token);
		else
			return "UVO_GTE";
	} else if (strcmp(token, "==") == 0)
		return (mv_signed ? "UVO_EQUAL" : "UV_EQUAL");
	else if (strcmp(token, "!=") == 0)
		return (mv_signed ? "UVO_NEQUAL" : "UV_NEQUAL");
	else
		error("Unknown binary operation '%s' encountered", token);

	/* In case of an error, return something. */
	return "ERROR";
}

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr_1, StackToken, StackExpr_2
      LEAVES  - StackExpr
      ACTION  - Code for the indicated operation and operands is
		produced and replaces the 3 items on top of the
		stack.
*/
void binary_op(void) {
	if (stack_error(3)) {
		stack_clear(3);
		top -= 2;
		stack[top].kind = StackError;
	} else {
		stack_entry entry_1, entry_2, entry_3;
		entry_3 = stack[top--];
		entry_2 = stack[top--];
		entry_1 = stack[top];

		/* Neither expression can be an array. */
		if (entry_1.data.expr.array)
			error("Unindexed array ('%s') cannot be used in expressions",
				&entry_1.data.expr.addr[1]);
		if (entry_3.data.expr.array)
			error("Unindexed array ('%s') cannot be used in expressions",
				&entry_3.data.expr.addr[1]);

		/* Only unstructured types can use +, -, *, /, % */
		if ((!entry_1.data.expr.unstructured || 
		     !entry_3.data.expr.unstructured) &&
		    (strcmp(entry_2.data.token, "+") || 
		     strcmp(entry_2.data.token, "-") ||
		     strcmp(entry_2.data.token, "*") ||
		     strcmp(entry_2.data.token, "/") ||
		     strcmp(entry_2.data.token, "%")))
			error("Illegal value in arithmetic operation", (char*) NULL);
		/* Only fields can use (non)equality tests. */
		else if (!entry_1.data.expr.unstructured && 
		    !entry_3.data.expr.unstructured && 
		    !entry_1.data.expr.array && 
		    !entry_3.data.expr.array && 
		    !strcmp(entry_2.data.token, "=") && 
		    !strcmp(entry_2.data.token, "!="))
			error("Comparisons of structured cells and agents are limited to '=' and '!='",
				(char*) NULL);
		else if (!entry_1.data.expr.unstructured && 
					!entry_3.data.expr.unstructured &&
			 entry_1.data.expr.agent != entry_3.data.expr.agent)
			error("Value types of operation, '%s', are incompatable",
				entry_2.data.token);

		stack[top].kind = StackExpr;
		stack[top].data.expr.assignable = false;
		stack[top].data.expr.constant = false;
		stack[top].data.expr.unstructured = true;
		stack[top].data.expr.agent = false;
		stack[top].data.expr.array = false;
		stack[top].data.expr.index = false;
		stack[top].data.expr.vector = mv_bits > 0;

		if (entry_1.data.expr.unstructured || !cell_fields_declared) {
			unsigned int size = strlen(entry_1.data.expr.addr) +
					    strlen(entry_3.data.expr.addr) + 1;

			if (mv_bits > 0) {
				char string[MAX_STRING_LENGTH];
				sprintf(string, "(,,%d)", mv_bits);
				size += strlen(which_mv_binary_op(
							entry_2.data.token)) +
					strlen(string);
			}
			else if (strcmp(entry_2.data.token, "/") == 0)
				if (use_floats)
					size += strlen("()/()");
				else
					size += strlen("mydiv((long int),(long int))");
			else if (strcmp(entry_2.data.token, "%") == 0)
				size += strlen("mymod((long int),(long int))");
			else
				size += strlen(entry_2.data.token) + 
					strlen("(,,)") + 2*strlen("()") +
					2*strlen(cast_type());

			stack[top].data.expr.addr = mymalloc(size);

			if (mv_bits > 0)
				sprintf(stack[top].data.expr.addr, 
					"%s(%s,%s,%d)",
					which_mv_binary_op(entry_2.data.token),
					entry_1.data.expr.addr,
					entry_3.data.expr.addr, mv_bits);
			else if (strcmp(entry_2.data.token, "/") == 0)
				if (use_floats)
					sprintf(stack[top].data.expr.addr, 
						"(%s)/(%s)",
						entry_1.data.expr.addr,
						entry_3.data.expr.addr);
				else
					sprintf(stack[top].data.expr.addr, 
						"mydiv((long int)%s,(long int)%s)",
						entry_1.data.expr.addr,
						entry_3.data.expr.addr);
			else if (strcmp(entry_2.data.token, "%") == 0) {
				if (use_floats && nested_arrays == 0) {
					error_prefix();
					fprintf(stderr,
						"Warning: \"%%\" may give unexpected results when using the -float option\n");
				}
				sprintf(stack[top].data.expr.addr, 
					"mymod((long int)%s,(long int)%s)",
					entry_1.data.expr.addr,
					entry_3.data.expr.addr);
			} else
				sprintf(stack[top].data.expr.addr, 
					"((%s)%s %s (%s)%s)",
					cast_type(),
					entry_1.data.expr.addr,
					entry_2.data.token,
					cast_type(),
					entry_3.data.expr.addr);
		} else if (agent_declared) {
			unsigned int size = strlen(entry_1.data.expr.addr) +
					    strlen(entry_3.data.expr.addr) +
					    strlen("agent_eq(,)") + 1;

			if (strcmp(entry_2.data.token, "!=")) size++;

			stack[top].data.expr.addr = mymalloc(size);

			if (strcmp(entry_2.data.token, "!="))
				sprintf(stack[top].data.expr.addr, 
					"!agent_eq(%s,%s)",
					entry_1.data.expr.addr,
					entry_3.data.expr.addr);
			else
				sprintf(stack[top].data.expr.addr, 
					"agent_eq(%s,%s)",
					entry_1.data.expr.addr,
					entry_3.data.expr.addr);
		} else {
			unsigned int size = strlen(entry_1.data.expr.addr) +
					    strlen(entry_3.data.expr.addr) +
					    strlen("eq(,)") + 1;

			if (strcmp(entry_2.data.token, "!=")) size++;

			stack[top].data.expr.addr = mymalloc(size);

			if (strcmp(entry_2.data.token, "!="))
				sprintf(stack[top].data.expr.addr, 
					"neq(%s,%s)",
					entry_1.data.expr.addr,
					entry_3.data.expr.addr);
			else
				sprintf(stack[top].data.expr.addr, 
					"eq(%s,%s)",
					entry_1.data.expr.addr,
					entry_3.data.expr.addr);
		}

		myfree(entry_1.data.expr.addr);
		myfree(entry_2.data.token);
		myfree(entry_3.data.expr.addr);
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr, StackToken
      LEAVES  - StackExpr
      ACTION  - Code for the indicated operation and operand is
		produced and replaces the 2 items on top of the
		stack.
*/
void unary_op(void) {
	if (stack_error(2)) {
		stack_clear(2);
		top--;
		stack[top].kind = StackError;
	} else {
		unsigned int size;
		stack_entry entry_1, entry_2;
		entry_2 = stack[top--];
		entry_1 = stack[top];

		/* Only fields can use unary operations. */
		if (!entry_2.data.expr.unstructured && !entry_2.data.expr.agent)
			error("Unary operations must operate on integer values",
				(char*) NULL);
			
		stack[top].kind = StackExpr;
		stack[top].data.expr.assignable = false;
		stack[top].data.expr.constant = false;
		stack[top].data.expr.unstructured = 
						entry_2.data.expr.unstructured;
		stack[top].data.expr.agent = false;
		stack[top].data.expr.array = false;
		stack[top].data.expr.index = false;
		stack[top].data.expr.vector = mv_bits > 0;

		size = strlen(entry_1.data.token) +
			 strlen(entry_2.data.expr.addr) + strlen("()") + 1;

		if (mv_bits > 0)
			if (strcmp(entry_1.data.token, "-") == 0) {
				if (mv_signed)
					size += strlen("UVO_NEG");
				else
					error("Cannot do negatives with unsigned microvectors",
					      (char *) NULL);
			} else if (strcmp(entry_1.data.token, "!") == 0) {
				if (mv_signed)
					size += strlen("UVO_NOT");
				else
					size += strlen("UV_NOT");
			}

		stack[top].data.expr.addr = mymalloc(size);

		if (mv_bits > 0) {
			if (strcmp(entry_1.data.token, "-") == 0) {
				if (mv_signed)
					sprintf(stack[top].data.expr.addr,
						"UVO_NEG(%s,%d)",
						entry_2.data.expr.addr, mv_bits);
				else
					error("Cannot do negatives with unsigned microvectors",
					      (char *) NULL);
			} else if (strcmp(entry_1.data.token, "!") == 0) {
				if (mv_signed)
					sprintf(stack[top].data.expr.addr,
						"UVO_NOT(%s,%d)",
						entry_2.data.expr.addr, mv_bits);
				else
					sprintf(stack[top].data.expr.addr,
						"UV_NOT(%s,%d)",
						entry_2.data.expr.addr, mv_bits);
			} else
				strcpy(stack[top].data.expr.addr,
					entry_2.data.expr.addr);
		} else
			sprintf(stack[top].data.expr.addr, "(%s%s)",
				entry_1.data.token,
				entry_2.data.expr.addr);
		
		myfree(entry_1.data.token);
		myfree(entry_2.data.expr.addr);
	}
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr, StackExpr(Index)
      LEAVES  - StackExpr
      ACTION  - Expects the StackExpr(Index) to describe an index variable.  
		A new StackExpr entry is built that represents an addition/
		subtraction modulo the size of the range associated with 
		the index variable.  If possible, the range of the index 
		variable is inferred and either set or checked against it's 
		current value.
*/
void do_mod_index(void) {
	if (stack_error(2)) {
		stack_clear(2);
		stack[top].kind = StackError;
	} else {
		char *offset = stack[top--].data.expr.addr;
		char *index = &stack[top].data.expr.addr[1];
		attr_rec *attr = get_attribute(index, VARIABLE);
		char *new_address;
		char string1[MAX_STRING_LENGTH];
		char string2[MAX_STRING_LENGTH];
		char string3[MAX_STRING_LENGTH];
		unsigned int size;

		/* Make sure the variable is an index variable. */
		if (attr && !attr->index)
			error("'%s' is not an index variable", index);

		/* Make sure the variable is an index variable. */
		if (mv_bits > 0 && !mv_signed)
			error("'+%%' and '-%%' are only available with signed microvectors",
			      (char*) NULL);

		sprintf(string1, "(%s)", offset);

		sprintf(string2, "%d", attr->index_count);

		sprintf(string3, "%d", mv_bits);

		size = strlen("INDEX__LBINDEX__SIZEINDEX__SIZEINDEX__LB") + 
			strlen(index) + strlen(string1) + 4*strlen(string2) + 1;
		if (mv_bits > 0)
			size += 3*strlen((mv_signed ? "UVO_PLUS" : "UV_PLUS"))
				+ strlen((mv_signed ? "UVO_MINUS" : "UV_MINUS"))
				+ strlen((mv_signed ? "UVO_MOD" : "UV_MOD")) +
				+ 4*strlen((mv_signed ? "UVO_S2VEC" : "UV_S2VEC")) +
				+ 9*strlen(string3) + 5*strlen("(,,)")
				+ 4*strlen("(,)") + 1;
		else
			size += strlen("(mymod(_+-+,)+)");

		new_address = mymalloc(size);

		if (mv_bits > 0)
			sprintf(new_address, 
				"%s(%s(%s(%s(%s(_%s,%s(INDEX_%d_SIZE,%d),%d),%s,%d),%s(INDEX_%d_LB,%d),%d),%s(INDEX_%d_SIZE,%d),%d),%s(INDEX_%d_LB,%d),%d)",
				(mv_signed ? "UVO_PLUS" : "UV_PLUS"),
				(mv_signed ? "UVO_MOD" : "UV_MOD"),
				(mv_signed ? "UVO_MINUS" : "UV_MINUS"),
				(mv_signed ? "UVO_PLUS" : "UV_PLUS"),
				(mv_signed ? "UVO_PLUS" : "UV_PLUS"), index,
				(mv_signed ? "UVO_S2VEC" : "UV_S2VEC"),
				attr->index_count, mv_bits, mv_bits,
				string1, mv_bits,
				(mv_signed ? "UVO_S2VEC" : "UV_S2VEC"),
				attr->index_count, mv_bits, mv_bits,
				(mv_signed ? "UVO_S2VEC" : "UV_S2VEC"),
				attr->index_count, mv_bits, mv_bits,
				(mv_signed ? "UVO_S2VEC" : "UV_S2VEC"),
				attr->index_count, mv_bits, mv_bits);
		else
			sprintf(new_address, 
				"(mymod(_%s+INDEX_%d_SIZE-INDEX_%d_LB+%s,INDEX_%d_SIZE)+INDEX_%d_LB)",
				index, attr->index_count, attr->index_count,
				string1, attr->index_count, attr->index_count);

		stack[top].kind = StackExpr;
		stack[top].data.expr.addr = new_address;
		stack[top].data.expr.assignable = false;
		stack[top].data.expr.constant = false;
		stack[top].data.expr.agent = false;
		stack[top].data.expr.unstructured = true;
		stack[top].data.expr.index = true;
		stack[top].data.expr.attribute = attr;
		stack[top].data.expr.vector = mv_bits > 0;

		stack[top].data.expr.size = attr->size;
		stack[top].data.expr.lower = attr->lower;

		myfree(stack[top+1].data.expr.addr);

		check_set_index_range(attr);
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackToken
      LEAVES  - StackToken
      ACTION  - Expects the StackToken to be the name of an array cell field.
		The attribute record for the array is pushed onto the front
		of the array_nesting chain.
*/
void push_array_attr(void) {
	char *token = stack[top].data.token;

	if (intable(token, CELL_FIELD))
		push_array(get_attribute(token, CELL_FIELD));
	else if (intable(token, AGENT_FIELD))
		push_array(get_attribute(token, AGENT_FIELD));
	else
		push_array((attr_rec*) NULL);
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr
      LEAVES  - StackExpr
      ACTION  - Converts the StackExpr whose address references a "now"
		cell, to one that references a "next" cell.  The StackExpr 
		should be suitable for assignment of the cell for the next 
		time step.
*/
void  now2next(void) {
	if (stack_error(1)) {
		stack_clear(1);
		stack[top].kind = StackError;
	} else {
		char *old_addr = stack[top].data.expr.addr;
		char *new_addr = mymalloc(strlen(old_addr)+2);
		sprintf(new_addr, "next%s", &old_addr[3]);
		stack[top].data.expr.addr = new_addr;
	}
}

/* SEMANTIC ROUTINE
      EXPECTS - StackToken
      LEAVES  - StackExpr
      ACTION  - Expects the StackToken to describe a the "cell" variable.  
		The StackExpr should be suitable for assignment of the
		cell for the next time step.
*/
void get_cell_variable(void) {
	char *token = stack[top].data.token;

	if (strcmp(token, "cell") != 0) {
		error("Expected the identifier 'cell'\n", (char*) NULL);
		stack[top].kind = StackError;
	} else {
		int i,j ;
		char dims[MAX_STRING_LENGTH];
		char tmp[MAX_STRING_LENGTH];

		strcpy(dims, "next[");
		for (i = 1; i < dimension_count; i++) {
			sprintf(tmp, "+dim_%d", i);
			strcat(dims, tmp);
			for (j = dimension_count; j > i; j--) {
				sprintf(tmp, "*(%d+MAX_OFFSET*2)", dim_size[j-1]);
				strcat(dims, tmp);
			}
		}
		sprintf(tmp, "+dim_%d]", dimension_count);
		strcat(dims, tmp);

		stack[top].kind = StackExpr;
		stack[top].data.expr.assignable = true;
		stack[top].data.expr.constant = false;

		stack[top].data.expr.addr = mymalloc(strlen(dims) + 1);
		strcpy(stack[top].data.expr.addr, dims);
	}
	myfree(token);
}

/************************************************************************/

/* SUPPORT ROUTINE
      ACTION - Generates code to synchronize processes.
*/
void gen_barrier(void) {
#if MAX_CPUS > 1
	/* Generate code to keep multiple processes away from one another. */
	if (cpus > 1) {
		print_indent();
#if SUN
		fprintf(file_c, "barrier();\n");
#elif SGI
		fprintf(file_c, "m_sync();\n");
#endif
	}
#endif
}

/* SUPPORT ROUTINE
      ACTION - Generates code to share the display among processes.
*/
void gen_lock_display(void) {
#if MAX_CPUS > 1
	if (cpus > 1) {
		print_indent();
#if SUN
		fprintf(file_c, 
			"while (0 != mutex_trylock(&display_mutex));");
#elif SGI
		fprintf(file_c, "m_lock();");
#endif
		fprintf(file_c, "\n");
	}
#endif
}

/* SUPPORT ROUTINE
      ACTION - Generates code to share the display among processes.
*/
void gen_unlock_display(void) {
#if MAX_CPUS > 1
	if (cpus > 1) {
		print_indent();
#if SUN
		fprintf(file_c, 
			"while (0 != mutex_unlock(&display_mutex));");
#elif SGI
		fprintf(file_c, "m_unlock();");
#endif
		fprintf(file_c, "\n");
	}
#endif
}

/************************************************************************/

/* SUPPORT ROUTINE
      ACTION - Generates code for the initialization in the main 
	       body of the program.
*/
void gen_swap_now_next(void) {
	int i;

	/* Generate code to exchange the values for "next" and "now". */
	print_indent();
	fprintf(file_c, "{\n");
	indent;

	print_indent();
	fprintf(file_c, "register void *tmp;\n");
	if (!cell_fields_declared && !agent_declared) {
		print_indent();
		fprintf(file_c, "tmp = next;\n");
		print_indent();
		fprintf(file_c, "next = now;\n");
		print_indent();
		fprintf(file_c, "now = tmp;\n");
	} else {
		for (i = 0; i < cell_field_count; i++) {
			if (field_list[i].alias) continue;
			if (field_list[i].constant) continue;

			print_indent();
			fprintf(file_c, "tmp = %s_next;\n", field_list[i].name);
			print_indent();
			fprintf(file_c, "%s_next = %s_now;\n",
				field_list[i].name, field_list[i].name);
			print_indent();
			fprintf(file_c, "%s_now = tmp;\n", field_list[i].name);
		}
		for (i = 0; i < array_field_alias_count; i++) {
			if (array_field_alias_list[i].constant) continue;
			print_indent();
			fprintf(file_c, "tmp = %s_next;\n",
				array_field_alias_list[i].name);
			print_indent();
			fprintf(file_c, "%s_next = %s_now;\n",
				array_field_alias_list[i].name,
				array_field_alias_list[i].name);
			print_indent();
			fprintf(file_c, "%s_now = tmp;\n",
				array_field_alias_list[i].name);
		}
	}
	if (agent_declared) {
		print_indent();
		fprintf(file_c, "tmp = agent_next;\n");
		print_indent();
		fprintf(file_c, "agent_next = agent_now;\n");
		print_indent();
		fprintf(file_c, "agent_now = tmp;\n");
	}

	undent;
	print_indent();
	fprintf(file_c, "}\n");
}

/************************************************************************/

/* SUPPORT ROUTINE
      ACTION - Generates code to check that the read times are legal,
	       (i.e. greater than the current time).
*/
void gen_check_time(void) {
	print_indent();
	fprintf(file_c, "if (next_time <= _time_step) {\n");
	indent;
	print_indent();
	fprintf(file_c, 
		"fprintf(stderr, \"%%s; time %%lu, entry %%lu: Must go forwards in time.\\n\", command, _time_step, entry);\n");
	print_indent();
	fprintf(file_c, "exit_signaled(1);\n");
	undent;
	print_indent();
	fprintf(file_c, "}\n");
}

/* SUPPORT ROUTINE
      ACTION - Generates code to give a read error message and quit, when
	       a index/value should have been read but wasn't.
*/
void gen_read_error(void) {
	indent;
	print_indent();
	fprintf(file_c, "fprintf(stderr, \"%%s; time %%lu, entry %%lu: read error\\n\", command, _time_step, entry);\n");
	print_indent();
	fprintf(file_c, "exit_signaled(1);\n");
	undent;
}

/* SUPPORT ROUTINE
      ACTION - Generates code to give a read error message and quit, when
	       the value read in was incorrect.
*/
void gen_read_value_error(int position) {
	indent;
	print_indent();
	if (position <= dimension_count)
		fprintf(file_c, "fprintf(stderr, \"%%s; time %%lu, entry %%lu: value of index %d is out of range\\n\", command, _time_step, entry);\n",
			position);
	else
		fprintf(file_c, "fprintf(stderr, \"%%s; time %%lu, entry %%lu: value of field %d is out of range\\n\", command, _time_step, entry);\n",
			position-dimension_count);
	print_indent();
	fprintf(file_c, "exit_signaled(1);\n");
	undent;
}

/* SUPPORT ROUTINE
      ACTION - Generates the code to call set_cell_entry with the first
	       two arguments which are dimension indices.
*/
void gen_set_cell_prefix(char *dim_1) {

/* DANA -- problems with this code!  This won't allow arbitrary slices
           of higher dimensional automata to be viewed.
*/
	fprintf(file_c, "set");
	if (use_floats) fprintf(file_c, "_float");
	fprintf(file_c, "_cell_entry(%s, ", dim_1);
	if (dimension_count > 1)
		fprintf(file_c, " dim_2, ");
	else
		fprintf(file_c, " 0, "); /* Could be one dimensional. */
	if (dimension_count > 2)
		fprintf(file_c, " dim_3, ");
	else
		fprintf(file_c, " 0, "); /* Could be two dimensional. */
}

/* SUPPORT ROUTINE
      ACTION - Generates code to for either a simple (non-array) cell field
	       name or a cell field that is an indexed array.
*/
void gen_cell_field_name(struct field_struct field, char *dim_1) {
	int i;
	if (cell_fields_declared && !field.alias) {
		fprintf(file_c, "%s_cells", field.name);
		fprintf(file_c, "[%s]", dim_1);
		for (i = 2; i <= dimension_count; i++)
			fprintf(file_c, "[dim_%d+MAX_OFFSET]", i);
	} else if (cell_fields_declared && field.alias) {
		char name[MAX_STRING_LENGTH];
		strcpy(name, field.name);
		fprintf(file_c, "%s_cells", strtok(name, "["));
		fprintf(file_c, "[%s]", dim_1);
		for (i = 2; i <= dimension_count; i++)
			fprintf(file_c, "[dim_%d+MAX_OFFSET]", i);
		fprintf(file_c, "[%s]", strtok((char*) NULL, "]"));
	}
}

/* SUPPORT ROUTINE
      ACTION - Generates code for either a simple (non-array) cell field
	       name or a cell field that is an indexed array of "now" cell.
*/
void gen_now_field_name(struct field_struct field, char *dim_1) {
	if (cell_fields_declared && !field.alias) {
		fprintf(file_c, "%s_now", field.name);
		gen_current_index(dim_1);
	} else if (cell_fields_declared && field.alias) {
		char name[MAX_STRING_LENGTH];
		strcpy(name, field.name);
		fprintf(file_c, "%s_now", strtok(name, "["));
		gen_current_index(dim_1);
		fprintf(file_c, "[%s]", strtok((char*) NULL, "]"));
	}
}

/* SUPPORT ROUTINE
      ACTION - Generates code to read the data from stdandard input
	       when the time has reached the time associated with
	       the next set of cell-value combinations to read.
*/
void gen_read_loop(void) {
	int i;

	/* As long as the time of the automata is the time of the
	   data entries, keep looping to read all of the cell-value
	   data.
	*/
	print_indent();
	fprintf(file_c, "entry = 0;\n");
	print_indent();
	fprintf(file_c, "while (_time_step == next_time) {\n");
	indent;

	/* Generate code to read in the dimension values.  Because the
	   dimension variables are held in registers, the actually values
	   are read into temporary locations and the values then assigned
	   to the dimension variables.
	*/

	/* Declare temporaries. */
#if MAX_CPUS > 1 
	if (cpus > 1) {
		print_indent();
		fprintf(file_c, "int shift_1;\n");
	}
#endif
	for (i = 1; i <= dimension_count; i++) {
		print_indent();
		fprintf(file_c, "int temp_dim_%d;\n", i);
	}

	/* Read dimension values into temporaries. */
	print_indent();
	fprintf(file_c, "int scan_count = scanf(\" [ %%d");
	for (i = 1; i < dimension_count; i++)
		fprintf(file_c, " , %%d");
	fprintf(file_c, " \"");
	for (i = 1; i <= dimension_count; i++)
		fprintf(file_c, ", &temp_dim_%d", i);
	fprintf(file_c, ");\n");

	if (agent_declared) {
		/* Assign the default field_value. */
		print_indent();
		fprintf(file_c, "field_value = 0;\n");
	}

	/* Assign temporaries to the dimension variables. */
	for (i = 1; i <= dimension_count; i++) {
		print_indent();
		fprintf(file_c, "dim_%d = temp_dim_%d;\n", i, i);
	}

	/* Update the input entry count, for the printing of error messgaes. */
	print_indent();
	fprintf(file_c, "entry++;\n");

	/* If no dimensions values read, then maybe this is a time. */
	print_indent();
	fprintf(file_c, "if (0 == scan_count || EOF == scan_count) {\n");
	indent;
	print_indent();
	fprintf(file_c, "scan_count = scanf(\" %%ld \", &next_time);\n");

	/* If a time was read, make sure it was a legal time. */
	print_indent();
	fprintf(file_c, "if (1 == scan_count) {\n");
	indent;
	gen_check_time();
	undent;
	print_indent();
	fprintf(file_c, "}\n");

	/* If this is the end of the input, then set next_time so that
	   no attempt is made to read from the input again. 
	*/
	print_indent();
	fprintf(file_c, "else if (0 == scan_count || EOF == scan_count) {\n");
	indent;
	/* Make sure that only spaces, tabs and newlines are left; otherwise
	   indicate an error.
	*/
	print_indent();
	fprintf(file_c, "int c = getchar();\n");
	print_indent();
	fprintf(file_c, "while (c == '\\n' || c == '\\t' || c == ' ') c = getchar();\n");
	print_indent();
	fprintf(file_c, "if (c != EOF) {\n");
	gen_read_error();
	print_indent();
	fprintf(file_c, "}\n");

	print_indent();
	fprintf(file_c, "next_time = -1;\n");

	undent;
	print_indent();
	fprintf(file_c, "}\n");
	print_indent();
	fprintf(file_c, "break;\n");

	undent;
	print_indent();
	fprintf(file_c, "}\n");

	/* Generate code to check and see if only some of the cell
	   indices were read in.
	*/
	print_indent();
	fprintf(file_c, "else if (scan_count != %d) {\n", dimension_count);
	indent;
	print_indent();
	fprintf(file_c, "fprintf(stderr, \"%%s; time %%lu, entry %%lu: Incorrect number of dimensions referenced\\n\", command, _time_step, entry);\n");
	print_indent();
	fprintf(file_c, "exit_signaled(1);\n");
	undent;
	print_indent();
	fprintf(file_c, "}\n");

	/* Generate code to check the legality of the cell indices. */
	for (i = 1; i <= dimension_count; i++) {
		print_indent();
		fprintf(file_c, "if (%d < temp_dim_%d || temp_dim_%d < 0) {\n",
			dim_size[i-1]-1, i, i);
		gen_read_value_error(i);
		print_indent();
		fprintf(file_c, "}\n");
	}

	print_indent();
	fprintf(file_c, "scanf(\" ] =\");\n");

	print_indent();
	fprintf(file_c, "if (output) inrange = in_range(dim_1");
	for (i = 2; i <= dimension_count; i++)
		fprintf(file_c, ",dim_%d", i);
	fprintf(file_c, ");\n");

	if (mv_bits > 0) {
		print_indent();
		fprintf(file_c, "word = dim_1%%NUM_OF_WORDS;\n");
		print_indent();
		fprintf(file_c, "cell_in_word = dim_1/NUM_OF_WORDS;\n");
	}

#if MAX_CPUS > 1 
	if (cpus > 1 && cell_field_count > 0) {
		print_indent();
		if (mv_bits > 0)
			fprintf(file_c,
				"shift_1 = word+(word/part_size)*(2*MAX_OFFSET+1);\n");
		else
			fprintf(file_c,
				"shift_1 = dim_1+(dim_1/part_size)*(2*MAX_OFFSET+1);\n");
	}
#endif

	/* Write code to print the cell location just read in. */
	print_indent();
	fprintf(file_c, 
		"if (_time_step == next_report_time || _time_step == next_read_time)\n");
	indent;
	print_indent();
	fprintf(file_c, "if (output && inrange) ");
	fprintf(file_c, "printf(\"[%%d");
	for (i = 2; i <= dimension_count; i++)
		fprintf(file_c, ", %%d");
	fprintf(file_c, "] =\"");
	for (i = 1; i <= dimension_count; i++)
		fprintf(file_c, ", dim_%d", i);
	fprintf(file_c, ");\n");
	undent;

	/* Read, check and write out all of the cell fields. */
	for (i = 0; i < cell_field_count; i++) {

		/* Only do stuff with this value, if a value was read in. */
		if (i > 0) {
			print_indent();
			fprintf(file_c, "if (scanf(\" , %s\", &read_value) > 0) {\n",
				format_type());
		} else {
			print_indent();
			fprintf(file_c, "if (scanf(\" %s\", &read_value) > 0) {\n",
				format_type());
		}

		indent;

		/* Make sure read values fall within the supported range. */
		print_indent();
		if (use_floats)
			fprintf(file_c, "if (%f < read_value || read_value < %f) {\n",
				field_list[i].range.ub.float_upper_bound,
				field_list[i].range.lb.float_lower_bound);
		else
			fprintf(file_c, "if (%ld < read_value || read_value < %ld) {\n",
				field_list[i].range.ub.upper_bound,
				field_list[i].range.lb.lower_bound);
		gen_read_value_error(i + 1 + dimension_count);
		print_indent();
		fprintf(file_c, "}\n");

		/* Zero out the appropriate portion of the desired word
		   if microvectors are being used.
		*/
		if (mv_bits > 0) {
			print_indent();
			if (cell_fields_declared)
#if MAX_CPUS > 1 
				if (cpus > 1)
					if (field_list[i].constant)
						gen_now_field_name(field_list[i], "word"); 
					else
						gen_now_field_name(field_list[i], "shift_1"); 
				else
#endif
				gen_now_field_name(field_list[i], "word"); 
			else {
				fprintf(file_c, "now");
#if MAX_CPUS > 1 
				if (cpus > 1)
					gen_current_index("shift_1");
				else
#endif
				gen_current_index("word");
			}
			fprintf(file_c,
				" &= (%#x ^ (BIT_MASK << ((CELLS_PER_WORD-cell_in_word-1)*BITS_PER_CELL)));\n",
				UINT_MAX);
		}

		/* Convert the read value into a signed micro vector value
		   if necessary.
		*/
		if (mv_bits > 0) {
			print_indent();
			fprintf(file_c, "if (read_value < 0) new_value = ");
			fprintf(file_c,
				"BIT_MASK&(UVO_COMPL(-read_value,%d)+1);\n",
				mv_bits);
			print_indent();
			fprintf(file_c, "else new_value = read_value;\n");
		}

		/* Assign read cell value to the cell (field). */
		print_indent();
		if (cell_fields_declared)
#if MAX_CPUS > 1 
			if (cpus > 1)
				if (field_list[i].constant)
					if (mv_bits > 0)
						gen_now_field_name(field_list[i], "word"); 
					else
						gen_now_field_name(field_list[i], "dim_1"); 
				else
					gen_now_field_name(field_list[i], "shift_1"); 
			else
#endif
			if (mv_bits > 0)
				gen_now_field_name(field_list[i], "word"); 
			else
				gen_now_field_name(field_list[i], "dim_1"); 
		else {
			fprintf(file_c, "now");
#if MAX_CPUS > 1 
			if (cpus > 1)
				gen_current_index("shift_1");
			else
#endif
			if (mv_bits > 0)
				gen_current_index("word");
			else
				gen_current_index("dim_1");
		}
		fprintf(file_c, " ");
		if (mv_bits > 0) fprintf(file_c, "|");
		fprintf(file_c, "= ");
		if (mv_bits > 0) fprintf(file_c, "(new_value");
		else fprintf(file_c, "read_value");
		if (mv_bits > 0)
			fprintf(file_c,
				" << ((CELLS_PER_WORD-cell_in_word-1)*BITS_PER_CELL))");
		fprintf(file_c, ";\n");

		undent;
		print_indent();
		fprintf(file_c, "} else read_value = 0;\n");

		/* Write out the value of this cell's field, since it
		   might have been changed.
		*/
		print_indent();
		fprintf(file_c, 
			"if (_time_step == next_report_time || _time_step == next_read_time) {\n");
		indent;
		print_indent();
		fprintf(file_c, "if (output && inrange) ");
		if (i > 0)
			if (!agent_declared && i+1 == cell_field_count)
				fprintf(file_c, "printf(\", %s\\n\"",
					format_type());
			else
				fprintf(file_c, "printf(\", %s\"",
					format_type());
		else if (cell_fields_declared)
			if (!agent_declared && i+1 == cell_field_count)
				fprintf(file_c, "printf(\" %s\\n\"",
					format_type());
			else
				fprintf(file_c, "printf(\" %s\"",
					format_type());
		else
			fprintf(file_c, "printf(\" %s\\n\"", format_type());
		fprintf(file_c, ", (%s) read_value);\n", cast_type());

		if (field_given && field_given-1 == i) {
			print_indent();
			fprintf(file_c, "if (view) ");
			gen_set_cell_prefix("dim_1");
			fprintf(file_c, "(%s) read_value);\n", cast_type());
		} else if (!field_given && cell_fields_declared) {
			print_indent();
			fprintf(file_c,
				"if (field+1 == %d) field_value = read_value;\n", i+1);
		} else if (!cell_fields_declared) {
			print_indent();
			fprintf(file_c, "if (view) ");
			gen_set_cell_prefix("dim_1");
			fprintf(file_c, "(%s) read_value);\n", cast_type());
		}
		undent;
		print_indent();
		fprintf(file_c, "}\n");
	}

	/* Read as many agent field values as given. */
	if (agent_declared) {

#if MAX_CPUS > 1 
		if (cpus > 1) {
			print_indent();
			fprintf(file_c,
				"shift_1 = dim_1+(dim_1/part_size)*(3*MAX_OFFSET+1);\n");
		}
#endif

		/* Skip a comma in the input if necessary. */
		if (!cell_fields_declared) {
			print_indent();
			fprintf(file_c, "ungetc(',', stdin);\n");
		}

		/* The loop to read all agents. */
		print_indent();
		fprintf(file_c, "count = %d;\n", cell_field_count);

		print_indent();
		fprintf(file_c, "while (1) {\n");
		indent;

		print_indent();
		fprintf(file_c, "int c;\n");

		for (i = cell_field_count; i < field_count; i++) {
			/* See if there is a next field. */
			print_indent();
			fprintf(file_c, "scanf(\" \");\n");
			print_indent();
			fprintf(file_c, "if ((c = getchar()) != ',') {\n");
			indent;
			print_indent();
			fprintf(file_c, "ungetc(c, stdin);\n");
			print_indent();
			fprintf(file_c, "break;\n");
			undent;
			print_indent();
			fprintf(file_c, "}\n");

			/* Create an agent, and add it to the _cell. */
			if (i == cell_field_count) {
				int j;
				print_indent();
				fprintf(file_c, "create_agent(");
				for (j = cell_field_count; j < field_count; j++)
					if (j == field_count - 1)
						fprintf(file_c, "0);\n");
					else
						fprintf(file_c, "0,");
				print_indent();
				fprintf(file_c, "tmp_agent->agent = agent_now");
#if MAX_CPUS > 1 
				if (cpus > 1)
					gen_current_index("shift_1");
				else
#endif
				gen_current_index("dim_1");
				fprintf(file_c, ";\n", dimension_count);
				print_indent();
				fprintf(file_c, "agent_now");
#if MAX_CPUS > 1 
				if (cpus > 1)
					gen_current_index("shift_1");
				else
#endif
				gen_current_index("dim_1");
				fprintf(file_c, " = tmp_agent;\n");
			}

			/* Count the fields. */
			print_indent();
			fprintf(file_c, "count++;\n");

			/* Get the value of the next field.  Only do stuff
			   with this value, if a value was read in.
			*/
			print_indent();
			fprintf(file_c, 
				"if (scanf(\" %s\", &read_value) > 0) {\n",
				format_type());
			indent;

			/* Make sure read values fall within the supported 
			   range.
			*/
			print_indent();
			if (use_floats)
				fprintf(file_c, 
					"if (%f < read_value || read_value < %f) {\n",
					field_list[i].range.ub.float_upper_bound,
					field_list[i].range.lb.float_lower_bound);
			else
				fprintf(file_c, 
					"if (%d < read_value || read_value < %d) {\n",
					field_list[i].range.ub.upper_bound,
					field_list[i].range.lb.lower_bound);
			gen_read_value_error(i + 1 + dimension_count);
			print_indent();
			fprintf(file_c, "}\n");

			/* Assign the read value to the agent field. */
			print_indent();
			fprintf(file_c, "tmp_agent->%s = read_value;\n",
				field_list[i].name);

			undent;
			print_indent();
			fprintf(file_c, "}\n");

			/* Write out the value for this field. */
			print_indent();
			fprintf(file_c, 
				"if (_time_step == next_report_time || _time_step == next_read_time)\n");
			indent;
			print_indent();
			fprintf(file_c, "if (output && inrange) {\n");
			indent;
			if (!cell_fields_declared && i == cell_field_count) {
				print_indent();
				fprintf(file_c, "if (scan_count > 0) {\n");
				indent;
				print_indent();
				fprintf(file_c, "scan_count = 0;\n");
				print_indent();
				fprintf(file_c, 
					"printf(\" %s\", tmp_agent->%s);\n",
					format_type(), field_list[i].name);
				undent;
				print_indent();
				fprintf(file_c, "} else ");
			} else
				print_indent();

			fprintf(file_c, "printf(\", %s\", tmp_agent->%s);\n",
				format_type(), field_list[i].name);

			undent;

			/* Set "field_value" when appropriate. */
			print_indent();
			if (field_given && (field_given-cell_field_count)%agent_field_count == i+1)
				fprintf(file_c, 
					"} else if (%d == count) field_value = tmp_agent->%s;\n", 
					field_given, field_list[i].name);
			else if (!field_given)
				fprintf(file_c, 
					"} else if (field+1 == count) field_value = tmp_agent->%s;\n", 
					field_list[i].name);
			else 
				fprintf(file_c, "}\n");

			undent;
		}

		undent;
		print_indent();
		fprintf(file_c, "}\n");
	}

	if ((!field_given && cell_fields_declared) || agent_declared ||
			(field_given && field_given > cell_field_count)
	   ) {

		/* Finish up output line. */
		print_indent();
		fprintf(file_c, 
			"if (_time_step == next_report_time || _time_step == next_read_time)\n");
		indent;

		/* Only generate code to finish a line of output if agent
		   values might have been printed.
		*/
		if (agent_declared) {
			print_indent();
			fprintf(file_c, "if (output && inrange) ");
			fprintf(file_c, "printf(\"\\n\");\n");
		}

		if ((!field_given && cell_fields_declared) || 
			(field_given && field_given > cell_field_count)) {
			print_indent();
			if (agent_declared) fprintf(file_c, "else ");
			fprintf(file_c, "{\n");
			indent;

			if (cell_fields_declared && !agent_declared) {
				print_indent();
				fprintf(file_c, "if (field > %d) {\n",
					cell_field_count);
				indent;

				print_indent();
				fprintf(file_c, 
					"fprintf(stderr, \"There is no field %%d\\n\", field+1);\n");
				print_indent();
				fprintf(file_c, "exit_signaled(1);\n");

				undent;
				print_indent();
				fprintf(file_c, "}\n");
			}

			print_indent();
			fprintf(file_c, "if (view) ");
			gen_set_cell_prefix("dim_1");
			if (cell_fields_declared || agent_declared)
				fprintf(file_c, "(%s) field_value", cast_type());
			else {
				fprintf(file_c, "(%s) now", cast_type());
				gen_current_index("dim_1");
			}
			fprintf(file_c, ");\n");

			undent;
			print_indent();
			fprintf(file_c, "}\n");
		}
		undent;
	}
	
	/* Close of the loop for reading data. */
	undent;
	print_indent();
	fprintf(file_c, "}\n");
}

/* SUPPORT ROUTINE
      ACTION - generates the C code, which copies the boundary values for
	       the constant cell fields only.
*/
void gen_copy_const_boundary(void) {
	int i, j, k, l;

	/* A surface for each dimension must be copied. */
	for (i = 1; i <= dimension_count; i++) {
		for (j = 1; j <= dimension_count; j++) {
			print_indent();
			if (i == j)
				fprintf(file_c,
					"for (dim_%d=0;dim_%d<MAX_OFFSET;dim_%d++)\n",
					j, j, j);
			else if (j > i)
				fprintf(file_c,
					"for (dim_%d=0;dim_%d<%d;dim_%d++)\n",
					j, j, dim_size[j-1], j);
			else if (mv_bits > 0 && j == 1)
				fprintf(file_c,
					"for (dim_%d=-MAX_OFFSET;dim_%d<NUM_OF_WORDS+MAX_OFFSET;dim_%d++)\n",
					j, j, j);
			else
				fprintf(file_c,
					"for (dim_%d=-MAX_OFFSET;dim_%d<%d+MAX_OFFSET;dim_%d++)\n",
					j, j, dim_size[j-1], j);
		}
		print_indent();
		fprintf(file_c, "{\n");
		indent;

		/* Generate the first assignment statement for copying. */
		print_indent();
		if (mv_bits > 0 && i == 1)
			fprintf(file_c, "left_shift_assign_const(now[");
		else
			fprintf(file_c, "assign_const(now[");
		for(k = 1; k <= dimension_count; k++) {
			if (mv_bits > 0 && k == 1 && i == k)
				fprintf(file_c, "+(NUM_OF_WORDS+dim_%d)", k);
			else if (i == k)
				fprintf(file_c, "+(%d+dim_%d)", dim_size[k-1], k);
			else
				fprintf(file_c, "+dim_%d", k);

			for (l = dimension_count; l > k; l--)
				fprintf(file_c,
					"*(%d+MAX_OFFSET*2)", dim_size[l-1]);
		}
		fprintf(file_c, "], now");
		gen_current_index("dim_1");
		fprintf(file_c, ");\n");

		/* Generate the second assignment statement for copying. */
		print_indent();
		if (mv_bits > 0 && i == 1)
			fprintf(file_c, "right_shift_assign_const(now[");
		else
			fprintf(file_c, "assign_const(now[");
		for(k = 1; k <= dimension_count; k++) {
			if (i == k)
				fprintf(file_c, "+(dim_%d-MAX_OFFSET)", k);
			else
				fprintf(file_c, "+dim_%d", k);

			for (l = dimension_count; l > k; l--)
				fprintf(file_c,
					"*(%d+MAX_OFFSET*2)", dim_size[l-1]);
		}
		fprintf(file_c, "], now[");
		for(k = 1; k <= dimension_count; k++) {
			if (mv_bits > 0 && k == 1 && i == k)
				fprintf(file_c, "+(NUM_OF_WORDS-MAX_OFFSET+dim_%d)",
					k);
			else if (i == k)
				fprintf(file_c, "+(%d-MAX_OFFSET+dim_%d)",
					dim_size[k-1], k);
			else
				fprintf(file_c, "+dim_%d", k);

			for (l = dimension_count; l > k; l--)
				fprintf(file_c,
					"*(%d+MAX_OFFSET*2)", dim_size[l-1]);
		}
		fprintf(file_c, "]);\n");

		undent;
		print_indent();
		fprintf(file_c, "}\n");
	}
}

/* SUPPORT ROUTINE
      ACTION - generates the C code, which copies the boundary values.
*/
void gen_copy_boundary(void) {
	int i, j, k, l;

#if MAX_CPUS > 1
	if (cpus > 1) {
		/* Only the first (0th) process does the copying. */
		print_indent();
		fprintf(file_c, "if (!id) {\n");
		indent;
	}
#endif

	/* A surface for each dimension must be copied. */
	for (i = 1; i <= dimension_count; i++) {
		for (j = 1; j <= dimension_count; j++) {
			print_indent();
			if (i == j)
				fprintf(file_c,
					"for (dim_%d=0;dim_%d<MAX_OFFSET;dim_%d++)\n",
					j, j, j);
			else if (j > i)
				fprintf(file_c,
					"for (dim_%d=0;dim_%d<%d;dim_%d++)\n",
					j, j, dim_size[j-1], j);
#if MAX_CPUS > 1
			else if (cpus > 1 && j == 1)
				if (mv_bits > 0)
					fprintf(file_c,
						"for (dim_%d=-MAX_OFFSET;dim_%d<NUM_OF_WORDS+(CPUS-1)*(2*MAX_OFFSET+1)+MAX_OFFSET;dim_%d++)\n",
						j, j, j);
				else
					fprintf(file_c,
						"for (dim_%d=-MAX_OFFSET;dim_%d<%d+(CPUS-1)*(2*MAX_OFFSET+1)+MAX_OFFSET;dim_%d++)\n",
						j, j, dim_size[j-1], j);
#endif
			else if (mv_bits > 0 && j == 1)
				fprintf(file_c,
					"for (dim_%d=-MAX_OFFSET;dim_%d<NUM_OF_WORDS+MAX_OFFSET;dim_%d++)\n",
					j, j, j);
			else
				fprintf(file_c,
					"for (dim_%d=-MAX_OFFSET;dim_%d<%d+MAX_OFFSET;dim_%d++)\n",
					j, j, dim_size[j-1], j);
		}
		print_indent();
		fprintf(file_c, "{\n");
		indent;

		/* Generate the first assignment statement for copying. */
		print_indent();
		if (mv_bits > 0 && i == 1)
			fprintf(file_c, "left_shift_assign_cell(now[");
		else
			fprintf(file_c, "assign_cell(now[");
		for(k = 1; k <= dimension_count; k++) {
#if MAX_CPUS > 1
			if (cpus > 1 && i == k && k == 1)
				if (mv_bits > 0)
					fprintf(file_c,
						"+(NUM_OF_WORDS+(CPUS-1)*(2*MAX_OFFSET+1)+dim_%d)",
						k);
				else
					fprintf(file_c,
						"+(%d+(CPUS-1)*(2*MAX_OFFSET+1)+dim_%d)",
						dim_size[k-1], k);
			else
#endif
			if (mv_bits > 0 && k == 1 && i == k)
				fprintf(file_c, "+(NUM_OF_WORDS+dim_%d)", k);
			else if (i == k)
				fprintf(file_c, "+(%d+dim_%d)", dim_size[k-1], k);
			else
				fprintf(file_c, "+dim_%d", k);

			for (l = dimension_count; l > k; l--)
				fprintf(file_c,
					"*(%d+MAX_OFFSET*2)", dim_size[l-1]);
		}
		fprintf(file_c, "], now");
		gen_current_index("dim_1");
		fprintf(file_c, ");\n");

		/* Generate the first assignment statement for NULL agent. */
		if (agent_declared) {
			print_indent();
			fprintf(file_c, "agent_now[");
			for(k = 1; k <= dimension_count; k++) {
#if MAX_CPUS > 1
				if (cpus > 1 && i == k && k == 1)
					fprintf(file_c,
						"+(%d+(CPUS-1)*(3*MAX_OFFSET+1)+dim_%d)",
						dim_size[k-1], k);
				else
#endif
				if (i == k)
					fprintf(file_c, "+(%d+dim_%d)",
						dim_size[k-1], k);
				else
					fprintf(file_c, "+dim_%d", k);

				for (l = dimension_count; l > k; l--)
					fprintf(file_c, "*(%d+MAX_OFFSET*2)",
						dim_size[l-1]);
			}
			fprintf(file_c, "] = NULL;\n");
		}

		/* Generate the second assignment statement for copying. */
		print_indent();
		if (mv_bits > 0 && i == 1)
			fprintf(file_c, "right_shift_assign_cell(now[");
		else
			fprintf(file_c, "assign_cell(now[");
		for(k = 1; k <= dimension_count; k++) {
			if (i == k)
				fprintf(file_c, "+(dim_%d-MAX_OFFSET)", k);
			else
				fprintf(file_c, "+dim_%d", k);

			for (l = dimension_count; l > k; l--)
				fprintf(file_c,
					"*(%d+MAX_OFFSET*2)", dim_size[l-1]);
		}
		fprintf(file_c, "], now[");
		for(k = 1; k <= dimension_count; k++) {
#if MAX_CPUS > 1
			if (cpus > 1 && i == k && k == 1)
				if (mv_bits > 0)
					fprintf(file_c,
						"+(NUM_OF_WORDS-MAX_OFFSET+(CPUS-1)*(2*MAX_OFFSET+1)+dim_%d)",
						k);
				else
					fprintf(file_c,
						"+(%d-MAX_OFFSET+(CPUS-1)*(2*MAX_OFFSET+1)+dim_%d)",
						dim_size[k-1], k);
			else
#endif
			if (mv_bits > 0 && k == 1 && i == k)
				fprintf(file_c, "+(NUM_OF_WORDS-MAX_OFFSET+dim_%d)",
					k);
			else if (i == k)
				fprintf(file_c, "+(%d-MAX_OFFSET+dim_%d)",
					dim_size[k-1], k);
			else
				fprintf(file_c, "+dim_%d", k);

			for (l = dimension_count; l > k; l--)
				fprintf(file_c,
					"*(%d+MAX_OFFSET*2)", dim_size[l-1]);
		}
		fprintf(file_c, "]);\n");

		/* Generate the second assignment statement for NULL agent. */
		if (agent_declared) {
			print_indent();
			fprintf(file_c, "agent_now[");
			for(k = 1; k <= dimension_count; k++) {
				if (i == k)
					fprintf(file_c, "+(dim_%d-MAX_OFFSET)", k);
				else
					fprintf(file_c, "+dim_%d", k);

				for (l = dimension_count; l > k; l--)
					fprintf(file_c, "*(%d+MAX_OFFSET*2)",
						dim_size[l-1]);
			}
			fprintf(file_c, "] = NULL;\n");
		}

		undent;
		print_indent();
		fprintf(file_c, "}\n");
	}

	if (constant_fields) gen_copy_const_boundary();

#if MAX_CPUS > 1
	if (cpus > 1) {
		undent;
		print_indent();
		fprintf(file_c, "}\n");
	}
#endif
}

/* SUPPORT ROUTINE
      ACTION - Generates code to copy the MAX_OFFSET sized gaps between
	       portions of the cellular universe that are being computed
	       by different processors when multi-processing is being used.
*/
void gen_copy_gap(void) {
	int i;

	print_indent();
	fprintf(file_c, "if ((inc>0 && id<CPUS-1) || (inc<0 && id>0)) {\n");
	indent;

	for (i = dimension_count; i >= 2; i--) {
		print_indent();
		fprintf(file_c, "for (dim_%d=0;dim_%d<%d;dim_%d++)",
			i, i, dim_size[i-1], i);
		if (i == 2) fprintf(file_c, " {");
		fprintf(file_c, "\n");
	}
	if (dimension_count > 1) indent;

	if (agent_declared) {
		print_indent();
		fprintf(file_c, "agent_now");
		gen_current_index("(dim_1_ub+inc*(2*MAX_OFFSET+1))");
		fprintf(file_c, " = NULL;\n");
	}

	print_indent();
	fprintf(file_c,
		"for (dim_1=dim_1_ub-inc*(MAX_OFFSET-1);dim_1!=dim_1_ub+inc;dim_1+=inc) {\n");
	indent;

	print_indent();
	fprintf(file_c, "assign_cell(now");
	gen_current_index("(dim_1+inc*(2*MAX_OFFSET+1))");
	fprintf(file_c, ", now");
	gen_current_index("dim_1");
	fprintf(file_c, ");\n");

	print_indent();
	fprintf(file_c, "assign_cell(now");
	gen_current_index("(dim_1+inc*MAX_OFFSET)");
	fprintf(file_c, ", now");
	gen_current_index("(dim_1+inc*(3*MAX_OFFSET+1))");
	fprintf(file_c, ");\n");

	if (agent_declared) {
		print_indent();
		fprintf(file_c, "agent_now");
		gen_current_index("(dim_1+inc*(3*MAX_OFFSET+1))");
		fprintf(file_c, " = NULL;\n");

		print_indent();
		fprintf(file_c, "agent_now");
		gen_current_index("(dim_1+inc*MAX_OFFSET)");
		fprintf(file_c, " = NULL;\n");

		print_indent();
		fprintf(file_c, "agent_now");
		gen_current_index("(dim_1+inc*2*MAX_OFFSET)");
		fprintf(file_c, " = NULL;\n");
	}

	undent;
	print_indent();
	fprintf(file_c, "}\n");
	if (dimension_count > 1) {
		undent;
		print_indent();
		fprintf(file_c, "}\n");
	}
	undent;
	print_indent();
	fprintf(file_c, "}\n");
}

/* SUPPORT ROUTINE
      ACTION - Generates code to perform the looping over the cellular
	       universe for each time step, starting at 0.
*/
void gen_begin_time_loop(void) {
	print_indent();
	/* Begin a possibly infinite "time" loop. */
	fprintf(file_c, "do {\n");
	indent;

	/* Declare variable used for both input and output. */
	if (agent_declared || !field_given) {
		print_indent();
		fprintf(file_c, "%s field_value;\n", cast_type());
	}
	if (agent_declared) {
		print_indent();
		fprintf(file_c, "long int count;\n");
	}

#if MAX_CPUS > 1
	if (cpus > 1) {
		gen_barrier();
		print_indent();
		fprintf(file_c, "if (!id) {\n");
		indent;
	}
#endif

	/* Generate code to write out the time. */
	print_indent();
	fprintf(file_c, "if (_time_step == next_report_time || _time_step == next_read_time)\n");
	indent;
	print_indent();
	fprintf(file_c, "if (output) ");
	fprintf(file_c, "printf(\"%%lu\\n\", _time_step);\n");
	undent;

	if (trace_count > 0) {
		int i;

		print_indent();
		fprintf(file_c, "fprintf(stderr, \"Trace of cell [");

		for (i = 0; i < trace_count-1; i++)
			fprintf(file_c, "%d, ", trace_point[i]);
		fprintf(file_c, "%d", trace_point[trace_count-1]);

		fprintf(file_c, "] at time %%lu\\n\", _time_step);\n");
	}

	gen_read_loop();

#if MAX_CPUS > 1
	/* Close the if statement for master only processing. */
	if (cpus > 1) {
		undent;
		print_indent();
		fprintf(file_c, "}\n");

		/* Must synchronize BEFORE copying the gaps! */
		gen_barrier();
		gen_copy_gap();
	}
#endif
	/* Must synchronize BEFORE copying the boundary! */
	gen_barrier();

	gen_copy_boundary();

	gen_barrier();
}

/* SUPPORT ROUTINE
      ACTION - Generates code to perform the looping over the cellular
	       universe for each time step, starting at 0.
*/
void gen_end_time_loop(void) {

	/* Close off the infinite "time" loop. */
	undent;
	print_indent();
#if MAX_CPUS > 1
	if (cpus > 1)
		fprintf(file_c, "} while (!view || id != CPUS-1);\n");
	else
#endif
		fprintf(file_c, "} while (!view);\n");
}

/* SUPPORT ROUTINE
      ACTION - Generates code to initialize the cells of the cellular
	       universe to 0.
*/
void gen_cell_init(void) {
	int i;

	print_indent();
	fprintf(file_c, "{\n");
	indent;

	/* Declare size variables. */
	print_indent();
	fprintf(file_c, "register unsigned int normal_size = ");

#if MAX_CPUS > 1 
	if (cpus > 1)
		if (mv_bits > 0)
			fprintf(file_c,
				"(NUM_OF_WORDS+MAX_OFFSET*3+1+(CPUS-1)*(2*MAX_OFFSET+1))");
		else
			fprintf(file_c,
				"(%d+MAX_OFFSET*3+1+(CPUS-1)*(2*MAX_OFFSET+1))",
				dim_size[0]);
	else
#endif
	if (mv_bits > 0)
		fprintf(file_c, "(NUM_OF_WORDS+MAX_OFFSET*3+1)");
	else
		fprintf(file_c, "(%d+MAX_OFFSET*3+1)", dim_size[0]);
	for (i = 1; i < dimension_count; i++)
		fprintf(file_c, "*(%d+MAX_OFFSET*2)", dim_size[i]);
	fprintf(file_c, ";\n");

	print_indent();
	fprintf(file_c, "register unsigned int const_size = ");
	if (mv_bits > 0)
		fprintf(file_c, "(NUM_OF_WORDS+MAX_OFFSET*2)");
	else
		fprintf(file_c, "(%d+MAX_OFFSET*2)", dim_size[0]);
	for (i = 1; i < dimension_count; i++)
		fprintf(file_c, "*(%d+MAX_OFFSET*2)", dim_size[i]);
	fprintf(file_c, ";\n");

	/* The following code initializes the field(s) of cells to 0/NULL. */
	for (i = 0; i < cell_field_count; i++) {
		range_entry range;
		char data_type[MAX_STRING_LENGTH];

		if (field_list[i].alias) continue;

		print_indent();
		fprintf(file_c, "memset(");
		if (cell_fields_declared)
			fprintf(file_c, "%s_", field_list[i].name);

		range = field_list[i].range;

		/* Determine the data type to use for this field. */
		get_type(data_type, range);

		fprintf(file_c, "cells, 0, sizeof(%s)*", data_type);

		if (field_list[i].constant)
			fprintf(file_c, "const");
		else
			fprintf(file_c, "normal");
		fprintf(file_c, "_size);\n");
	}
	for (i = 0; i < array_field_alias_count; i++) {
		range_entry range;
		char data_type[MAX_STRING_LENGTH];

		print_indent();
		fprintf(file_c, "memset(");
		fprintf(file_c, "%s_cells", array_field_alias_list[i].name);

        	range = array_field_alias_list[i].range;

		/* Determine the data type to use for this field. */
		get_type(data_type, range);

		fprintf(file_c, ", 0, sizeof(%s)*", data_type);

		if (array_field_alias_list[i].constant)
			fprintf(file_c, "const");
		else
			fprintf(file_c, "normal");
		fprintf(file_c, "_size*%d);\n", array_field_alias_list[i].size);
	}
	if (agent_declared) {
		print_indent();
		fprintf(file_c, "normal_size = ");
#if MAX_CPUS > 1 
		if (cpus > 1)
			fprintf(file_c,
				"(%d+MAX_OFFSET*3+1+(CPUS-1)*(3*MAX_OFFSET+1))",
				dim_size[0]);
		else
#endif
		fprintf(file_c, "(%d+MAX_OFFSET*3+1)", dim_size[0]);
		for (i = 1; i < dimension_count; i++)
			fprintf(file_c, "*(%d+MAX_OFFSET*2)", dim_size[i]);
		fprintf(file_c, ";\n");

		print_indent();
		fprintf(file_c,
			"memset(agent_cells, 0, sizeof(void*)*normal_size);\n");
	}

	undent;
	print_indent();
	fprintf(file_c, "}\n");
}

/* SUPPORT ROUTINE
      ACTION - Generates code for the remainder of the expression used to
	       determine what slice of the cell universe each cpu computes.
*/
void gen_rest_partition_expr(void) {
	if (pow_of_2(cpus))
		fprintf(file_c, ">>%d))", pow_of_2(cpus));
	else
		fprintf(file_c, "/CPUS))");
	fprintf(file_c, ";\n");
}

/* SUPPORT ROUTINE
      ACTION - Generates code for the variable declarations in the main 
	       body of the program.
*/
void gen_declarations(void) {
	int i;

	/* Declare (locally) the now/next offset variables. */
	gen_now_next();

	/* Declare (locally) the variables for placing values into 
	   microvectors.
	*/
	if (mv_bits > 0) {
		print_indent();
		fprintf(file_c, "register int word;\n");
		print_indent();
		fprintf(file_c, "register int cell_in_word;\n");
	}

	/* Declare (locally) the dimension variables. */
	for (i = 1; i <= dimension_count; i++) {
		print_indent();
		fprintf(file_c, "register int dim_%d;\n", i);
	}

	/* For counting lines of the input data to the cellular atuomata. */
	print_indent();
	fprintf(file_c, "unsigned long int entry = 1;\n");

	/* Temporary variable used for reading in cell field values. */
	print_indent();
	fprintf(file_c, "%s read_value;\n", cast_type());

	print_indent();
	fprintf(file_c, "int index_value;\n");

	print_indent();
	fprintf(file_c, "int inrange = 0;\n");
	if (mv_bits > 0) {
		print_indent();
		fprintf(file_c, "uvec new_value;\n");
	}

	/* Supporting declaration for agent. */
	if (agent_declared) {
		print_indent();
		fprintf(file_c, "struct agent_type *tmp_agent;\n");
	}

	/* Declare variables to hold the upper and lower bounds on the 
	   first dimension within the function so that different processes
	   don't clobber one another.
	*/
#if MAX_CPUS > 1 
	if (cpus > 1) {
		print_indent();
		fprintf(file_c, "unsigned int part_lb, part_ub;\n");
	}
#endif
#if MAX_CPUS > 1 
	if (cpus > 1) {
		print_indent();
		fprintf(file_c, "unsigned int dim_1_lb, dim_1_ub;\n");
	}
#endif

	/* (Re)Initialize the variables part_lb, part_ub, dim_1_lb and
	   dim_1_ub, which the bounds of indexing for the outermost for
	   loop, if necesary.
	*/
#if MAX_CPUS > 1 
	if (cpus > 1) {
		print_indent();
		fprintf(file_c, "dim_1_lb = part_lb = id*((");
		if (mv_bits > 0)
			fprintf(file_c, "NUM_OF_WORDS");
		else
			fprintf(file_c, "%d", dim_size[0]);
		gen_rest_partition_expr();

		print_indent();
		fprintf(file_c,
			"if (CPUS-1 != id) part_ub = (id+1)*((");
		if (mv_bits > 0)
			fprintf(file_c, "NUM_OF_WORDS");
		else
			fprintf(file_c, "%d", dim_size[0]);
		gen_rest_partition_expr();
		print_indent();
		fprintf(file_c, "else part_ub = ");
		if (mv_bits > 0)
			fprintf(file_c, "NUM_OF_WORDS");
		else
			fprintf(file_c, "%d", dim_size[0]);
		fprintf(file_c, ";\n");

		print_indent();
		fprintf(file_c, "dim_1_ub = part_ub-1;\n");

		/* Set and check the partition size. */
		print_indent();
		fprintf(file_c, "if (!id) {\n");
		indent;

		print_indent();
		fprintf(file_c, "part_size = dim_1_ub - dim_1_lb + 1;\n");
		print_indent();
		fprintf(file_c, "if (part_size < MAX_OFFSET) {\n");
		indent;

		print_indent();
		fprintf(file_c,
			"fprintf(stderr, \"ERROR: Recompile with fewer CPUS\\n\");\n");
		print_indent();
		fprintf(file_c, "exit_signaled(1);\n");

		undent;
		print_indent();
		fprintf(file_c, "}\n");

		undent;
		print_indent();
		fprintf(file_c, "}\n");
	}
#endif

	print_indent();
	fprintf(file_c, "if (inc < 0) {\n");
	indent;

#if MAX_CPUS > 1 
	if (cpus > 1) {
		/* Swap the values for dim_1_lb and dim_1_ub if necessary. */
		print_indent();
		fprintf(file_c, "register unsigned int tmp;\n");
		print_indent();
		fprintf(file_c, "tmp = dim_1_lb;\n");
		print_indent();
		fprintf(file_c, "dim_1_lb = dim_1_ub;\n");
		print_indent();
		fprintf(file_c, "dim_1_ub = tmp;\n");
	}
#endif

	gen_swap_now_next();

	undent;
	print_indent();
	fprintf(file_c, "}\n");
}

/* SUPPORT ROUTINE
      ACTION - Generates code for the initialization in the main 
	       body of the program.
*/
void gen_initialize(void) {

	/* If no data is given, don't try to read from it ever again. */
	print_indent();
	fprintf(file_c, "int scan_count = scanf(\" %%ld \", &next_time);\n");
	print_indent();
	fprintf(file_c, "if (0 == scan_count || EOF == scan_count) {\n");
	indent;
	/* Make sure that only spaces, tabs and newlines are left; otherwise
	   indicate an error.
	*/
	print_indent();
	fprintf(file_c, "int c = getchar();\n");
	print_indent();
	fprintf(file_c, "while (c == '\\n' || c == '\\t' || c == ' ') c = getchar();\n");

	print_indent();
	fprintf(file_c, "if (c != EOF) {\n");
	indent;
	print_indent();
	fprintf(file_c, "fprintf(stderr, \"%%s: read error at beginning of input file\\n\", command);\n");
	print_indent();
	fprintf(file_c, "exit_signaled(1);\n");
	undent;
	print_indent();
	fprintf(file_c, "}\n");

	print_indent();
	fprintf(file_c, "next_time = -1;\n");
	undent;

	/* Make sure that the time read in is okay. */
	print_indent();
	fprintf(file_c, "} else if (next_time < 0) {\n");
	indent;
	print_indent();
	fprintf(file_c, 
		"fprintf(stderr, \"%%s: First time in input file must be >= 0\\n\", command);\n");
	print_indent();
	fprintf(file_c, "exit_signaled(1);\n");
	undent;
	print_indent();
	fprintf(file_c, "}\n");

	gen_cell_init();
}

/* SUPPORT ROUTINE
      ACTION - Generates code to read the report time file entries.
*/
void gen_get_report_time(void) {
	print_indent();
	fprintf(file_c, "void get_report_time(FILE *file) {\n");
	indent;

	/* Get the next entry from the file. */
	print_indent();
	fprintf(file_c, "int c;\n");
	print_indent();
	fprintf(file_c, "if (file == NULL) return;\n");
	print_indent();
	fprintf(file_c, "report_time_step = 0;\n");
	print_indent();
	fprintf(file_c, "while (next_read_time >= 0)\n");
	indent;
	print_indent();
	fprintf(file_c, "if (1 == fscanf(file, \" +%%lu\", &report_time_step));\n");
	print_indent();
	fprintf(file_c, "else if (1 == fscanf(file, \" %%ld\", &next_read_time)) break;\n");
	print_indent();
	fprintf(file_c, "else if ((c = getc(file)) == '!') ");
	fprintf(file_c, "exit_signaled(0);\n");
	print_indent();
	fprintf(file_c, "else if (c == ' ' || c == '\\n');\n");
	print_indent();
	fprintf(file_c, "else next_read_time = -1;\n");
	undent;

	print_indent();
	fprintf(file_c, "if (report_time_step > 0) {\n");
	indent;

	print_indent();
	fprintf(file_c, "next_report_time = _time_step + report_time_step;\n");
	print_indent();
	fprintf(file_c, "if (next_read_time >= 0 && next_report_time > next_read_time)\n");
	indent;
	print_indent();
	fprintf(file_c, "next_report_time = next_read_time;\n");
	undent;

	undent;
	print_indent();
	fprintf(file_c, "} else\n");
	indent;
	print_indent();
	fprintf(file_c, "next_report_time = next_read_time;\n");
	undent;

	undent;
	print_indent();
	fprintf(file_c, "}\n\n");
}

/* SUPPORT ROUTINE
      ACTION - Generates code to read the report time file entries.
*/
void gen_init_cells(void) {
	print_indent();
	fprintf(file_c, "void init_cells(void) {\n");
	indent;

	gen_initialize();

	print_indent();
	fprintf(file_c, "signal(SIGFPE, bad_arith);\n");

	print_indent();
	fprintf(file_c, "get_report_time(time_file);\n");

	/* Call all of the constant array reader functions.  Because it is
	   possible to have constant arrays declared after the cell
	   declaration, which is when this function is generated, a function
	   generated at the end of the compilation must be called, which in
	   turn calls the individual array readers.
	*/
	print_indent();
	fprintf(file_c, "const_array_readers();\n");

	undent;
	print_indent();
	fprintf(file_c, "}\n\n");
}

/************************************************************************/

/* SUPPORT ROUTINE
      ACTION - Generates the definition for right_shifts for microvectors.
*/
void gen_right_shift(char *name, boolean copy_all, boolean copy_const) {
	int i;
	int bits_per_word = sizeof(int)*8;
	int cells_per_word = bits_per_word/mv_bits;
	unsigned int mask = pow_2(mv_bits)-1;
	unsigned int reverse_mask = 0;

	for (i = 1; i < cells_per_word; i++)
		reverse_mask = (reverse_mask << mv_bits) | mask;

	fprintf(file_c, "#define %s(X,Y) ", name);
	if (cell_fields_declared && cell_field_count > 0) {
		fprintf(file_c, "{ \\\n"); 
		for (i = 0; i < cell_field_count; i++) {
			if (!copy_all && field_list[i].constant != copy_const)
				continue;
			if (field_list[i].alias) {
				char tmp[MAX_STRING_LENGTH];
				char *name, *index;
				strcpy(tmp, field_list[i].name);
				name = strtok(tmp, "[");
				index = strtok((char*) NULL, "]");

				fprintf(file_c,
					"\t%s_##X[%s] = (((%s_##Y[%s] >> BITS_PER_CELL) & ",
					name, index, name, index);
				fprintf(file_c, "%#x) \\\n", reverse_mask);
				fprintf(file_c,
					"\t\t| ((%s_##Y[%s] & BIT_MASK) << %d))",
					name, index,
					mv_bits*(cells_per_word-1));
			} else {
				fprintf(file_c,
					"\t%s_##X = (((%s_##Y >> BITS_PER_CELL) & ",
					field_list[i].name, field_list[i].name);
				fprintf(file_c, "%#x) \\\n", reverse_mask);
				fprintf(file_c,
					"\t\t| ((%s_##Y & BIT_MASK) << %d))",
					field_list[i].name,
					mv_bits*(cells_per_word-1));
			}
			fprintf(file_c, "; \\\n");
		}
		fprintf(file_c, "}\n\n");
	} else if (cell_field_count == 0)
		fprintf(file_c, ";\n\n"); 
	else {
		fprintf(file_c, "\\\n"); 
		fprintf(file_c, "\tX = (((Y >> BITS_PER_CELL) & ");
		fprintf(file_c, "%#x) \\\n", reverse_mask);
		fprintf(file_c, "\t\t| ((Y & BIT_MASK) << %d))\n\n",
			mv_bits*(cells_per_word-1));
	}
}

/* SUPPORT ROUTINE
      ACTION - Generates the definition for left_shifts for microvectors.
*/
void gen_left_shift(char *name, boolean copy_all, boolean copy_const) {
	int i;
	int bits_per_word = sizeof(int)*8;
	int cells_per_word = bits_per_word/mv_bits;
	unsigned int mask = pow_2(mv_bits)-1;
	unsigned int reverse_mask = 0;

	for (i = 1; i < cells_per_word; i++)
		reverse_mask = (reverse_mask << mv_bits) | mask;

	fprintf(file_c, "#define %s(X,Y) ", name);
	if (cell_fields_declared && cell_field_count > 0) {
		fprintf(file_c, "{ \\\n"); 
		for (i = 0; i < cell_field_count; i++) {
			if (!copy_all && field_list[i].constant != copy_const)
				continue;
			if (field_list[i].alias) {
				char tmp[MAX_STRING_LENGTH];
				char *name, *index;
				strcpy(tmp, field_list[i].name);
				name = strtok(tmp, "[");
				index = strtok((char*) NULL, "]");
				fprintf(file_c,
					"\t%s_##X[%s] = ((%s_##Y[%s] << BITS_PER_CELL) | ((%s_##Y[%s] & ",
					name, index, name, index, name, index);
				fprintf(file_c, "((uvec)BIT_MASK << %d))",
					mv_bits*(cells_per_word-1));
				fprintf(file_c, " >> %d))",
					mv_bits*(cells_per_word-1));
			} else {
				fprintf(file_c,
					"\t%s_##X = ((%s_##Y << BITS_PER_CELL) | ((%s_##Y & ",
					field_list[i].name,
					field_list[i].name,
					field_list[i].name);
				fprintf(file_c,
					"((uvec)BIT_MASK << %d))",
					mv_bits*(cells_per_word-1));
				fprintf(file_c, " >> %d))",
					mv_bits*(cells_per_word-1));
			}
			fprintf(file_c, "; \\\n");
		}
		fprintf(file_c, "}\n\n");
	} else if (cell_field_count == 0)
		fprintf(file_c, ";\n\n"); 
	else {
		fprintf(file_c, "\\\n"); 
		fprintf(file_c, "\tX = ((Y << BITS_PER_CELL) | ((Y & ");
		fprintf(file_c, "((uvec)BIT_MASK << %d))",
			mv_bits*(cells_per_word-1));
		fprintf(file_c, " >> %d))\n\n", mv_bits*(cells_per_word-1));
	}
}

/* SUPPORT ROUTINE
      ACTION - Generates the definition for in_range, which is used to
	       determine whether or not a cell value should be printed
	       when the -o (output) option is given.
*/
void gen_in_range(void) {
	int i;

	fprintf(file_c, "#define in_range(INDEX_1");
	for (i = 1; i < dimension_count; i++)
		fprintf(file_c, ",INDEX_%d", i+1);
	fprintf(file_c, ") (!dim_given || (1 \\\n");

	for (i = 0; i < dimension_count; i++)
		fprintf(file_c,
			"\t&& (lower[%d] <= INDEX_%d && INDEX_%d <= upper[%d]) \\\n",
			i, i+1, i+1, i);

	fprintf(file_c, "))\n\n");
}

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Writes code to perform all the actions on the cellular 
		automata and its rules.  Including, but not limited to,
		variable declarations, I/O, cell initialization.
*/
void begin_rules(void) {
	int i;
	long int j;

	decl_time_and_cell_and_random();

	fprintf(file_c, "#include \"%s%s%d.hh\"\n", 
		TMP_DIR, BASE_FILE_NAME, pid);

	fprintf(file_c, "#define DIMS %d\n", dimension_count);

	if (dimension_count >= 1)
		fprintf(file_c, "#define DIM_1_SIZE %d\n", dim_size[0]);
	if (dimension_count >= 2)
		fprintf(file_c, "#define DIM_2_SIZE %d\n", dim_size[1]);
	if (dimension_count >= 3)
		fprintf(file_c, "#define DIM_3_SIZE %d\n", dim_size[2]);
	fprintf(file_c, "\n");

	/* Microvectors are incompatible with agents. */
	if (mv_bits > 0 && agent_declared)
		error("Microvectors cannot be used with agents", (char*) NULL);

	/* Declarations for using microvectors. */
	if (mv_bits > 0) {
		int i;
		int bits_per_word = sizeof(int)*8;
		int cells_per_word = bits_per_word/mv_bits;
		unsigned int mask = pow_2(mv_bits)-1;
		unsigned int reverse_mask = 0;

		for (i = 1; i < cells_per_word; i++)
			reverse_mask = (reverse_mask << mv_bits) | mask;

		fprintf(file_c, "#define BITS_PER_CELL %d\n", mv_bits);
		fprintf(file_c, "#define BITS_PER_WORD %d\n", bits_per_word);
		fprintf(file_c, "#define CELLS_PER_WORD %d\n", cells_per_word);

		if (dim_size[0] % cells_per_word != 0) {
			error("Microvectors attempting to pack %d cells per word",
				(char*) cells_per_word);
			error("Set the first dimension (-s option) to be a multiple of %d",
				(char*) cells_per_word);
		}

		fprintf(file_c, "#define NUM_OF_WORDS %d\n",
			dim_size[0]/cells_per_word);
		fprintf(file_c, "#define BIT_MASK %#x\n\n", mask);

		fprintf(file_c, "#define WORD_MASK (UV_S2VEC(1, %d))\n", mv_bits);

		fprintf(file_c, "#define guarded_copy(X,Y) \\\n");
		fprintf(file_c,
			"\t(X = (((X)&(UV_COMPL(guard,%d))) | ((Y) & guard)))\n\n",
			mv_bits);

		gen_right_shift("right_shift_copy_cell", true, true);
		gen_right_shift("right_shift_assign_cell", false, false);
		gen_right_shift("right_shift_assign_const", false, true);

		gen_left_shift("left_shift_copy_cell", true, true);
		gen_left_shift("left_shift_assign_cell", false, false);
		gen_left_shift("left_shift_assign_const", false, true);
	}

	/* Declare the cell universe(s) and their corresponding now and
	   next pointers.  Two MAX_OFFSETs are for the copying of boundaries,
	   while the third is for buffering between now and next values.

	   Each dimension of the cellular universe, except the first one,
	   has 2*MAX_OFFSET extra elements for boundary copying.  The first
	   dimension has an additional MAX_OFFSET+1 elements for the buffering
	   between now and next values.  The now and next positions are offset
	   MAX_OFFSET+1 with respect to one another in the first dimension.
	   The two variables are also declared here.

	   In 2 dimensions, the picture relating "cells", "next", and "now"
	   looks like:

		+-----------------------------------------------+
		|						|
		|   +---+-------------------------------+---+	|
		|   |	|				|   |	|
		|   |	|				|   |	|
		|   |	|				|   |	|
		|   |	|				|   |	|
		|   |	|				|   |	|
		|   |	| now/next			|   |	|
		|   +---+-------------------------------+---+	|
		|   next/now					|
		+-----------------------------------------------+
		cells

	*/
	if (cell_field_count > 0) {
		char dims[MAX_STRING_LENGTH];
		char tmp[MAX_STRING_LENGTH];

#if MAX_CPUS > 1 
		if (cpus > 1)
			if (mv_bits > 0)
				sprintf(tmp,
					"[NUM_OF_WORDS+MAX_OFFSET*3+1+(CPUS-1)*(2*MAX_OFFSET+1)]");
			else
				sprintf(tmp,
					"[DIM_1_SIZE+MAX_OFFSET*3+1+(CPUS-1)*(2*MAX_OFFSET+1)]");
		else
#endif
		if (mv_bits > 0)
			sprintf(tmp, "[NUM_OF_WORDS+MAX_OFFSET*3+1]");
		else
			sprintf(tmp, "[DIM_1_SIZE+MAX_OFFSET*3+1]");
		strcpy(dims, tmp);
		for (i = 1; i < dimension_count; i++) {
			sprintf(tmp, "[DIM_%d_SIZE+MAX_OFFSET*2]", i+1);
			strcat(dims, tmp);
		}

		gen_field_arrays(file_c, "static", "cells", dims, true);
	}
	if (agent_declared) {
		fprintf(file_c, "static struct agent_type *agent_cells");
#if MAX_CPUS > 1 
		if (cpus > 1)
			fprintf(file_c,
				"[%d+MAX_OFFSET*3+1+(CPUS-1)*(3*MAX_OFFSET+1)]",
				dim_size[0]);
		else
#endif
		fprintf(file_c, "[%d+MAX_OFFSET*3+1]", dim_size[0]);
		for (i = 1; i < dimension_count; i++)
			fprintf(file_c, "[%d+MAX_OFFSET*2]", dim_size[i]);
		fprintf(file_c, ";\n\n");
	}

	/* Declare and initialize (if possible) the variables containing
	   the bounds of indexing for the outermost for loop.
	*/
	for (i = 1; i <= dimension_count; i++) {
#if MAX_CPUS > 1 
		if (cpus > 1 && i == 1)
			;
		else
#endif
		{
			print_indent();
			fprintf(file_c,
				"static unsigned int dim_%d_lb = 0;\n", i);
			print_indent();
			if (mv_bits > 0 && i == 1)
				fprintf(file_c,
					"static unsigned int dim_%d_ub = NUM_OF_WORDS-1;\n",
					i);
			else
				fprintf(file_c,
					"static unsigned int dim_%d_ub = %d;\n",
					i, dim_size[i-1]-1);
		}
	}

	/* Size of a partition. */
#if MAX_CPUS > 1 
	if (cpus > 1) {
		print_indent();
		fprintf(file_c, "static unsigned int part_size;\n");
	}
#endif

	/* Increment direction for primary for loops. */
	print_indent();
	fprintf(file_c, "static int inc = 1;\n\n");

	fprintf(file_c, "#include \"%scommon.c\"\n\n", LIB_DIR);

	gen_in_range();

#if MAX_CPUS > 1
	/* Setup for doing multi-processing. */
	if (cpus > 1)
		fprintf(file_c, "#include \"%smulti.c\"\n\n", LIB_DIR);
	else
#endif
		fprintf(file_c, "#include \"%suni.c\"\n\n", LIB_DIR);

	/* Define the function for reading the report time file entries. */
	gen_get_report_time();

	/* Define the function for initializing the cell universe. */
	gen_init_cells();

	/* Start the "main" body of the program. */
#if MAX_CPUS > 1
	fprintf(file_c, "int cellang_main (int output, int view, int filter, int field, int id) {\n");
#else
	fprintf(file_c, "int cellang_main (int output, int view, int filter, int field) {\n");
#endif
	indent;

	gen_declarations();

	gen_begin_time_loop();

	/* Generate code to step through the ranges of all indices. */
	for (i = 1; i <= dimension_count; i++) {
		print_indent();
		fprintf(file_c, "dim_%d = dim_%d_lb;\n", i, i);
		print_indent();
		fprintf(file_c, "while (1) {\n");
		indent;

		/* Create the file, even if it will be empty. */
		{
			char file_name[MAX_STRING_LENGTH];
			FILE *file_dim_h;

			sprintf(file_name, "%s%s%dd%d.h", 
				TMP_DIR, BASE_FILE_NAME, pid, i);
			file_dim_h = fopen(file_name, "a");

			fclose(file_dim_h);
		}
		fprintf(file_c, "#include \"%s%s%dd%d.h\"\n", 
			TMP_DIR, BASE_FILE_NAME, pid, i);
	}

	/* Begin statements inside loops that range over dimensional ranges. */

	if (mv_bits > 0) fprintf(file_c, "register uvec guard;\n");

	/* Generate code to read in the variables include file. */
	fprintf(file_c, "#include \"%s%s%d.h\"\n", TMP_DIR, BASE_FILE_NAME, pid);

	/* Generate code to set the "next" cell to the current value of
	   the "now" cell, which acts as its default value.
	*/
	print_indent();
	fprintf(file_c, "assign_cell(next");
	gen_current_index("dim_1");
	fprintf(file_c, ", now");
	gen_current_index("dim_1");
	fprintf(file_c, ");\n");
}

/* SUPPORT ROUTINE
      ACTION - generates the definition file for MAX_OFFSET.
*/
void gen_define_maxoffset(void) {
	/* Define MAX_OFFSET. */
	fprintf(file_hh, "#define MAX_OFFSET %d\n", max_rel_index);
}

/* SUPPORT ROUTINE
      ACTION - generates the C code, which combines the agent gap values.
*/
void gen_combine_gap(void) {
	int i, j;

	print_indent();
	fprintf(file_c, "if ((inc>0 && id<CPUS-1) || (inc<0 && id>0))\n");
	indent;

	print_indent();
	fprintf(file_c,
		"for (dim_1=dim_1_ub-inc*(MAX_OFFSET-1);dim_1!=dim_1_ub+inc;dim_1+=inc)\n");
	for (i = 2; i <= dimension_count; i++) {
		print_indent();
		fprintf(file_c, "for (dim_%d=0;dim_%d<%d;dim_%d++)\n",
			i, i, dim_size[i-1], i);
	}
	print_indent();
	fprintf(file_c, "{\n");
	indent;

	print_indent();
	fprintf(file_c, "register struct agent_type *tmp_agent;\n");
	print_indent();
	fprintf(file_c, "register int from_index;\n");
	print_indent();
	fprintf(file_c, "register int to_index;\n");

	/* Generate the first looping combinator. */
	print_indent();
	fprintf(file_c, "to_index = ");
	for(i = 1; i <= dimension_count; i++) {
		fprintf(file_c, "+dim_%d", i);

		for (j = dimension_count; j > i; j--)
			fprintf(file_c, "*(%d+MAX_OFFSET*2)", dim_size[j-1]);
	}
	fprintf(file_c, ";\n");
	print_indent();
	fprintf(file_c, "tmp_agent = agent_next[to_index];\n");
	print_indent();
	fprintf(file_c, "from_index = ");
	for(i = 1; i <= dimension_count; i++) {
		if (i == 1)
			fprintf(file_c,
				"+(inc*(3*MAX_OFFSET+1)+dim_%d)", i);
		else
			fprintf(file_c, "+dim_%d", i);

		for (j = dimension_count; j > i; j--)
			fprintf(file_c, "*(%d+MAX_OFFSET*2)", dim_size[j-1]);
	}
	fprintf(file_c, ";\n");
	print_indent();
	fprintf(file_c, "if (tmp_agent) {\n");
	indent;
	print_indent();
	fprintf(file_c, "while (tmp_agent->agent) tmp_agent = tmp_agent->agent;\n");
	print_indent();
	fprintf(file_c, "tmp_agent->agent = agent_next[from_index];\n");
	undent;
	print_indent();
	fprintf(file_c, "} else\n");
	indent;
	print_indent();
	fprintf(file_c, "agent_next[to_index] = agent_next[from_index];\n");
	undent;

	/* Generate the second looping combinator. */
	print_indent();
	fprintf(file_c, "to_index = ");
	for(i = 1; i <= dimension_count; i++) {
		if (i == 1)
			fprintf(file_c, "(inc*(4*MAX_OFFSET+1)+dim_%d)", i);
		else
			fprintf(file_c, "+dim_%d", i);

		for (j = dimension_count; j > i; j--)
			fprintf(file_c, "*(%d+MAX_OFFSET*2)", dim_size[j-1]);
	}
	fprintf(file_c, ";\n");
	print_indent();
	fprintf(file_c, "tmp_agent = agent_next[to_index];\n");
	print_indent();
	fprintf(file_c, "from_index = ");
	for(i = 1; i <= dimension_count; i++) {
		if (i == 1)
			fprintf(file_c, "+(dim_%d+inc*MAX_OFFSET)", i);
		else
			fprintf(file_c, "+dim_%d", i);

		for (j = dimension_count; j > i; j--)
			fprintf(file_c, "*(%d+MAX_OFFSET*2)", dim_size[j-1]);
	}
	fprintf(file_c, ";\n");


	print_indent();
	fprintf(file_c, "if (tmp_agent) {\n");
	indent;
	print_indent();
	fprintf(file_c, "while (tmp_agent->agent) tmp_agent = tmp_agent->agent;\n");
	print_indent();
	fprintf(file_c, "tmp_agent->agent = agent_next[from_index];\n");
	undent;
	print_indent();
	fprintf(file_c, "} else\n");
	indent;
	print_indent();
	fprintf(file_c, "agent_next[to_index] = agent_next[from_index];\n");
	undent;

	undent;
	print_indent();
	fprintf(file_c, "}\n");

	undent;
}

/* SUPPORT ROUTINE
      ACTION - generates the C code, which combines the agent boundary values.
*/
void gen_combine_boundary(void) {
	int i, j, k, l;

	print_indent();
#if MAX_CPUS > 1
	if (cpus > 1) {
		/* Only the first (0th) process does the combining. */
		fprintf(file_c, "if (!id) ");
	}
#endif
	fprintf(file_c, "{\n");
	indent;
	print_indent();
	fprintf(file_c, "register struct agent_type *tmp_agent;\n");
	print_indent();
	fprintf(file_c, "register int from_index;\n");
	print_indent();
	fprintf(file_c, "register int to_index;\n");

	/* A surface for each dimension must be copied. */
	for (i = 1; i <= dimension_count; i++) {
		for (j = 1; j <= dimension_count; j++) {
			print_indent();
			if (i == j)
				fprintf(file_c,
					"for (dim_%d=0;dim_%d<MAX_OFFSET;dim_%d++)\n",
					j, j, j);
			else if (j < i)
				fprintf(file_c,
					"for (dim_%d=0;dim_%d<%d;dim_%d++)\n",
					j, j, dim_size[j-1], j);
#if MAX_CPUS > 1
			else if (cpus > 1 && j == 1)
				fprintf(file_c,
					"for (dim_%d=-MAX_OFFSET;dim_%d<%d+(CPUS-1)*(3*MAX_OFFSET+1)+MAX_OFFSET;dim_%d++)\n",
					j, j, dim_size[j-1], j);
#endif
			else
				fprintf(file_c,
					"for (dim_%d=-MAX_OFFSET;dim_%d<%d+MAX_OFFSET;dim_%d++)\n",
					j, j, dim_size[j-1], j);
		}
		print_indent();
		fprintf(file_c, "{\n");
		indent;

		/* Generate the first looping combinator. */
		print_indent();
		fprintf(file_c, "to_index = ");
		for(k = 1; k <= dimension_count; k++) {
			fprintf(file_c, "+dim_%d", k);

			for (l = dimension_count; l > k; l--)
				fprintf(file_c,
					"*(%d+MAX_OFFSET*2)", dim_size[l-1]);
		}
		fprintf(file_c, ";\n");
		print_indent();
		fprintf(file_c, "tmp_agent = agent_next[to_index];\n");
		print_indent();
		fprintf(file_c, "from_index = ");
		for(k = 1; k <= dimension_count; k++) {
#if MAX_CPUS > 1
			if (cpus > 1 && i == k && k == 1)
				fprintf(file_c,
					"+(%d+(CPUS-1)*(3*MAX_OFFSET+1)+dim_%d)",
					dim_size[k-1], k);
			else
#endif
			if (i == k)
				fprintf(file_c, "+(%d+dim_%d)", dim_size[k-1], k);
			else
				fprintf(file_c, "+dim_%d", k);

			for (l = dimension_count; l > k; l--)
				fprintf(file_c,
					"*(%d+MAX_OFFSET*2)", dim_size[l-1]);
		}
		fprintf(file_c, ";\n");
		print_indent();
		fprintf(file_c, "if (tmp_agent) {\n");
		indent;
		print_indent();
		fprintf(file_c, "while (tmp_agent->agent) tmp_agent = tmp_agent->agent;\n");
		print_indent();
		fprintf(file_c, "tmp_agent->agent = agent_next[from_index];\n");
		undent;
		print_indent();
		fprintf(file_c, "} else\n");
		indent;
		print_indent();
		fprintf(file_c, "agent_next[to_index] = agent_next[from_index];\n");
		undent;

		/* Generate the second looping combinator. */
		print_indent();
		fprintf(file_c, "to_index = ");
		for(k = 1; k <= dimension_count; k++) {
#if MAX_CPUS > 1
			if (cpus > 1 && i == k && k == 1)
				fprintf(file_c,
					"+(%d-MAX_OFFSET+(CPUS-1)*(3*MAX_OFFSET+1)+dim_%d)",
					dim_size[k-1], k);
			else
#endif
			if (i == k)
				fprintf(file_c, "+(%d-MAX_OFFSET+dim_%d)",
					dim_size[k-1], k);
			else
				fprintf(file_c, "+dim_%d", k);

			for (l = dimension_count; l > k; l--)
				fprintf(file_c,
					"*(%d+MAX_OFFSET*2)", dim_size[l-1]);
		}
		fprintf(file_c, ";\n");
		print_indent();
		fprintf(file_c, "tmp_agent = agent_next[to_index];\n");
		print_indent();
		fprintf(file_c, "from_index = ");
		for(k = 1; k <= dimension_count; k++) {
			if (i == k)
				fprintf(file_c, "+(dim_%d-MAX_OFFSET)", k);
			else
				fprintf(file_c, "+dim_%d", k);

			for (l = dimension_count; l > k; l--)
				fprintf(file_c,
					"*(%d+MAX_OFFSET*2)", dim_size[l-1]);
		}
		fprintf(file_c, ";\n");


		print_indent();
		fprintf(file_c, "if (tmp_agent) {\n");
		indent;
		print_indent();
		fprintf(file_c, "while (tmp_agent->agent) tmp_agent = tmp_agent->agent;\n");
		print_indent();
		fprintf(file_c, "tmp_agent->agent = agent_next[from_index];\n");
		undent;
		print_indent();
		fprintf(file_c, "} else\n");
		indent;
		print_indent();
		fprintf(file_c, "agent_next[to_index] = agent_next[from_index];\n");
		undent;

		undent;
		print_indent();
		fprintf(file_c, "}\n");
	}

	undent;
	print_indent();
	fprintf(file_c, "}\n");
}

/* SUPPORT ROUTINE
      ACTION - generates the C code to free the storage being used by any 
	       agents which might be associated with a cell from this (the
	       last) time step.
*/
void gen_free_agents(void) {
	int i, j;

	/* Remove all of the agents associated with the current cell. */
	print_indent();
	fprintf(file_c, "while (agent_now");
	gen_current_index("dim_1");
	fprintf(file_c, ") {\n");
	indent;

	print_indent();
	fprintf(file_c, "tmp_agent = agent_now");
	gen_current_index("dim_1");
	fprintf(file_c, ";\n");
	print_indent();
	fprintf(file_c, "agent_now");
	gen_current_index("dim_1");
	fprintf(file_c, " = agent_now");
	gen_current_index("dim_1");
	fprintf(file_c, "->agent;\n");
	print_indent();
	fprintf(file_c, "free(tmp_agent);\n");

	undent;
	print_indent();
	fprintf(file_c, "}\n");

	/* Make sure that the agent field in the cells array is set to NULL. */
	print_indent();
	fprintf(file_c, "agent_now[");
	for (i = 1; i <= dimension_count; i++) {
		fprintf(file_c, "+dim_%d", i);

		for (j = dimension_count; j > i; j--)
			fprintf(file_c, "*(%d+MAX_OFFSET*2)", dim_size[j-1]);
	}
	fprintf(file_c, "] = NULL;\n");
}


/* SUPPORT ROUTINE
      ACTION - Generates code to reference changed cell field value.
*/
void gen_field_value() {
	if (field_given && field_given <= cell_field_count)
		if (field_list[field_given-1].alias) {
			char tmp[MAX_STRING_LENGTH];
			char *name, *index;
			strcpy(tmp, field_list[field_given-1].name);
			name = strtok(tmp, "[");
			index = strtok((char*) NULL, "]");
			fprintf(file_c, "(%s) %s_next",
				(mv_bits > 0 ? "uvec" : cast_type()), name);
			gen_current_index("dim_1");
			fprintf(file_c, "[%s]", index);
		} else {
			fprintf(file_c, "(%s) %s_next",
				(mv_bits > 0 ? "uvec" : cast_type()),
				field_list[field_given-1].name);
			gen_current_index("dim_1");
		}
	else if (cell_fields_declared || agent_declared)
		fprintf(file_c, "(%s) field_value",
			(mv_bits > 0 ? "uvec" : cast_type()));
	else {
		fprintf(file_c, "(%s) next",
			(mv_bits > 0 ? "uvec" : cast_type()));
		gen_current_index("dim_1");
	}
}

/* SUPPORT ROUTINE
      ACTION - Generates code to display/print changed cell values.
*/
void gen_show_cell(void) {
	int i;

	fprintf(file_c, "\n");
	print_indent();
	fprintf(file_c, "if (1 == report_time_step || (report_time_step != 1 && (_time_step == next_report_time || _time_step == next_read_time))) {\n");
	indent;
	if (!cell_fields_declared && agent_field_count == 0) {
		print_indent();
		fprintf(file_c,
			"if (view && (report_time_step != 1 || next");
		gen_current_index("dim_1");
		fprintf(file_c, " != now");
		gen_current_index("dim_1");
		fprintf(file_c, ")) {\n");

		indent;
	} else if (field_given && field_given <= cell_field_count) {
		print_indent();
		fprintf(file_c,
			"if (view && (report_time_step != 1 || %s_next",
			field_list[field_given-1].name);
		gen_current_index("dim_1");
		fprintf(file_c, " != %s_now",
			field_list[field_given-1].name);
		gen_current_index("dim_1");
		fprintf(file_c, ")) {\n");

		indent;
	} else if (!field_given) {
		print_indent();
		fprintf(file_c, "if (view && field < %d) {\n",
			cell_field_count);

		indent;

		if (cell_fields_declared) {
			print_indent();
			fprintf(file_c, "switch (field) {\n");
			indent;
			for (i = 0; i < cell_field_count; i++) {
				print_indent();
				if (field_list[i].alias) {
					char tmp[MAX_STRING_LENGTH];
					char *name, *index;
					strcpy(tmp, field_list[i].name);
					name = strtok(tmp, "[");
					index = strtok((char*) NULL, "]");
					fprintf(file_c,
						"case %d: { field_value = %s_",
						i, name);

					if (field_list[i].constant)
						fprintf(file_c, "now");
					else
						fprintf(file_c, "next");
					gen_current_index("dim_1");

					fprintf(file_c, "[%s]", index);
				} else {
					fprintf(file_c,
						"case %d: { field_value = %s_",
						i, field_list[i].name);

					if (field_list[i].constant)
						fprintf(file_c, "now");
					else
						fprintf(file_c, "next");
					gen_current_index("dim_1");
				}
				fprintf(file_c, "; break; }\n");
			}

			if (!agent_declared) {
				/* An error since there were no agent fields. */
				print_indent();
				fprintf(file_c, "default: {\n");
				indent;

				print_indent();
				fprintf(file_c,
					"fprintf(stderr, \"There is no field %%d\\n\", field+1);\n");
				print_indent();
				fprintf(file_c, "exit_signaled(1);\n");

				undent;
				print_indent();
				fprintf(file_c, "}\n");
			}

			undent;
			print_indent();
			fprintf(file_c, "}\n");
		}
	}

	if (!field_given ||
		(field_given && field_given <= cell_field_count)) {

		if (mv_bits > 0) {
			print_indent();
			fprintf(file_c, "{\n");
			indent;

			print_indent();
			fprintf(file_c, "register int i;\n");
			print_indent();
			fprintf(file_c, "register uvec word = ");
			gen_field_value();
			fprintf(file_c, ";\n");

			print_indent();
			fprintf(file_c, "for (i = NUM_OF_WORDS*(CELLS_PER_WORD-1)+dim_1; i >= dim_1 ; i -= NUM_OF_WORDS) {\n");
			indent;
		}

		print_indent();
		if (mv_bits > 0)
			gen_set_cell_prefix("i");
		else
			gen_set_cell_prefix("dim_1");

		if (mv_bits > 0)
			fprintf(file_c, "(long int) %s(word,%d)",
				(mv_signed ? "UVO_FSVAL" : "UV_FVAL"), mv_bits);
		else
			gen_field_value();
		fprintf(file_c, ");\n");

		if (mv_bits > 0) {
			print_indent();
			fprintf(file_c,
				"if (i > dim_1) word >>= BITS_PER_CELL;\n");
			undent;
			print_indent();
			fprintf(file_c, "}\n");

			undent;
			print_indent();
			fprintf(file_c, "}\n");
		}

		undent;

		print_indent();
		fprintf(file_c, "}\n");
	}

	print_indent();
	fprintf(file_c, "if (output) inrange = in_range(dim_1");
	for (i = 2; i <= dimension_count; i++)
		fprintf(file_c, ",dim_%d", i);
	fprintf(file_c, ");\n");

	print_indent();
	fprintf(file_c, "if (output && inrange && (report_time_step != 1 || changed(next");
	gen_current_index("dim_1");
	fprintf(file_c, ", now");
	gen_current_index("dim_1");
	fprintf(file_c, "))) {\n");
	indent;

	if (mv_bits > 0) {
		print_indent();
		fprintf(file_c, "register int i;\n");
		for (i = 0; i < cell_field_count; i++) {
			print_indent();
			fprintf(file_c, "register uvec word_%d = (uvec) ", i+1);

			if (cell_fields_declared)
				if (field_list[i].alias) {
					char tmp[MAX_STRING_LENGTH];
					char *name, *index;
					strcpy(tmp, field_list[i].name);
					name = strtok(tmp, "[");
					index = strtok((char*) NULL, "]");

					if (field_list[i].constant)
						fprintf(file_c, "%s_now", name);
					else
						fprintf(file_c, "%s_next", name);
					gen_current_index("dim_1");
					fprintf(file_c, "[%s]", index);
				} else {
					if (field_list[i].constant)
						fprintf(file_c, "%s_now",
							field_list[i].name);
					else
						fprintf(file_c, "%s_next",
							field_list[i].name);
					gen_current_index("dim_1");
				}
			else {
				fprintf(file_c, "next");
				gen_current_index("dim_1");
			}
			fprintf(file_c, ";\n");
		}
		print_indent();
		fprintf(file_c, "for (i = NUM_OF_WORDS*(CELLS_PER_WORD-1)+dim_1; i >= dim_1 ; i -= NUM_OF_WORDS) {\n");
		indent;
	}

	gen_lock_display();

	print_indent();
	fprintf(file_c, "printf(\"[%%d");
	for (i = 2; i <= dimension_count; i++)
		fprintf(file_c, ", %%d");
	fprintf(file_c, "] = \"");

	if (mv_bits > 0)
		fprintf(file_c, ", i");
	else
		fprintf(file_c, ", dim_1");
	for (i = 2; i <= dimension_count; i++)
		fprintf(file_c, ", dim_%d", i);
	fprintf(file_c, ");\n");

	if (mv_bits > 0)
		for (i = 0; i < cell_field_count; i++) {
			print_indent();
			fprintf(file_c, "printf(\"");
			if (i > 0)
				fprintf(file_c, ", ");
			fprintf(file_c, "%s\"", format_type());

			fprintf(file_c, ", (int) %s(word_%d,%d));\n",
				(mv_signed ? "UVO_FSVAL" : "UV_FVAL"),
				i+1, mv_bits);
		}
	else if (cell_fields_declared)
		for (i = 0; i < cell_field_count; i++) {
			print_indent();
			fprintf(file_c, "printf(\"");
			if (i > 0)
				fprintf(file_c, ", ");
			fprintf(file_c, "%s\"", format_type());
			if (field_list[i].alias) {
				char tmp[MAX_STRING_LENGTH];
				char *name, *index;
				strcpy(tmp, field_list[i].name);
				name = strtok(tmp, "[");
				index = strtok((char*) NULL, "]");
				fprintf(file_c, ", %s_", name);

				if (field_list[i].constant)
					fprintf(file_c, "now");
				else
					fprintf(file_c, "next");
				gen_current_index("dim_1");

				fprintf(file_c, "[%s]", index);
			} else {
				fprintf(file_c, ", %s_", field_list[i].name);

				if (field_list[i].constant)
					fprintf(file_c, "now");
				else
					fprintf(file_c, "next");
				gen_current_index("dim_1");
			}
			fprintf(file_c, ");\n");
		}
	else {
		print_indent();
		fprintf(file_c, "printf(\"%s\"", format_type());
		fprintf(file_c, ", next");
		gen_current_index("dim_1");
		fprintf(file_c, ");\n");
	}
	print_indent();
	fprintf(file_c, "printf(\"\\n\");\n");

	gen_unlock_display();

	if (mv_bits > 0) {
		print_indent();
		fprintf(file_c, "if (i > dim_1) {\n");
		indent;
		for (i = 0; i < cell_field_count; i++) {
			print_indent();
			fprintf(file_c, "word_%d >>= BITS_PER_CELL;\n", i+1);
		}
		undent;
		print_indent();
		fprintf(file_c, "}\n");

		undent;
		print_indent();
		fprintf(file_c, "}\n");
	}

	undent;
	print_indent();
	fprintf(file_c, "}\n");

	undent;
	print_indent();
	fprintf(file_c, "}\n");
}

/* SUPPORT ROUTINE
      ACTION - Generates code to display/print agent values.
*/
void gen_show_agent(void) {
	int i;

	/* Only do this if the field to be printed requires it. */
	if (!field_given) {
		print_indent();
		fprintf(file_c, "if (field > %d || output)\n",
			cell_field_count-1);
		indent;
	}

	print_indent();
	fprintf(file_c, "if (1 == report_time_step || (report_time_step != 1 && (_time_step == next_report_time || _time_step == next_read_time))) {\n");
	indent;

	gen_barrier();

	/* Generate code to step through the ranges of all indices. */
	print_indent();
#if MAX_CPUS > 1
	if (cpus > 1)
		fprintf(file_c, "for(dim_1=part_lb;dim_1<part_ub;dim_1++)");
	else
#endif
	fprintf(file_c, "for(dim_1 = 0; dim_1 < %d; dim_1++)", dim_size[0]);
	for (i = 2; i <= dimension_count; i++) {
		fprintf(file_c, "\n");
		print_indent();
		fprintf(file_c, "for(dim_%d = 0; dim_%d < %d; dim_%d++)",
			i, i, dim_size[i-1], i);
	}
	fprintf(file_c, " {\n");
	indent;

	print_indent();
	fprintf(file_c, "if (output) inrange = in_range(dim_1");
	for (i = 2; i <= dimension_count; i++)
		fprintf(file_c, ",dim_%d", i);
	fprintf(file_c, ");\n");

	print_indent();
	fprintf(file_c, "if (output && inrange && agent_next");
	gen_current_index("dim_1");
	fprintf(file_c, ") {\n");
	indent;

	gen_lock_display();

	/* Generate the cell location. */
	print_indent();
	fprintf(file_c, "printf(\"[%%d");
	for (i = 2; i <= dimension_count; i++)
		fprintf(file_c, ", %%d");
	fprintf(file_c, "] = \"");

	for (i = 1; i <= dimension_count; i++)
		fprintf(file_c, ", dim_%d", i);

	fprintf(file_c, ");\n");

 	/* Skip over the cell field(s). */
	for (i = 0; i < cell_field_count; i++) {
		print_indent();
		fprintf(file_c, "printf(\", \");\n");
	}

	/* Print the fields of any associated agents. */
/* DANA - note that this code will NOT print out a 0 value indicating the
          abscense of an agent that was indicated during a previous time step.
*/
	print_indent();
	fprintf(file_c, "tmp_agent = agent_next");
	gen_current_index("dim_1");
	fprintf(file_c, ";\n");
	print_indent();
	fprintf(file_c, "while (tmp_agent) {\n");

	indent;

	for (i = cell_field_count; i < field_count; i++) {
		print_indent();
		if (i < field_count-1)
			fprintf(file_c, "printf(\"%s, \", ", format_type());
		else
			fprintf(file_c, "printf(\"%s\", ", format_type());
		fprintf(file_c, "tmp_agent->%s);\n", field_list[i].name);
	}

	print_indent();
	fprintf(file_c, "tmp_agent = tmp_agent->agent;\n");

	print_indent();
	fprintf(file_c, "if (tmp_agent) printf(\", \");\n");

	undent;
	print_indent();
	fprintf(file_c, "}\n");

	print_indent();
	fprintf(file_c, "printf(\"\\n\");\n");

	gen_unlock_display();

	undent;
	print_indent();
	fprintf(file_c, "}");

	if (!field_given || (field_given && field_given > cell_field_count)) {
		fprintf(file_c, " else if (view) {\n");
		indent;

		/* Skip over enough agents to find the desired field. */
		print_indent();
		fprintf(file_c, "count = %d;\n", field_count);
		print_indent();
		fprintf(file_c, "tmp_agent = agent_next");
		gen_current_index("dim_1");
		fprintf(file_c, ";\n");
		print_indent();

		if (field_given)
			fprintf(file_c, "while (tmp_agent && count <= %d) {\n",
				field_given-1);
		else
			fprintf(file_c,
				"while (tmp_agent && count <= field) {\n");

		indent;

		print_indent();
		fprintf(file_c, "tmp_agent = tmp_agent->agent;\n");
		print_indent();
		fprintf(file_c, "count += %d;\n", agent_field_count);

		undent;
		print_indent();
		fprintf(file_c, "}\n");

		/* Get the desired field of the found agent. */
		print_indent();
		fprintf(file_c, "if (!tmp_agent) field_value = 0;\n");
		print_indent();

		if (field_given)
			fprintf(file_c, "else switch (%d-count) {\n",
				field_given-1+agent_field_count);
		else
			fprintf(file_c,
				"else switch (field-count+%d) {\n",
				agent_field_count);

		indent;

		for (i = cell_field_count; i < field_count; i++) {
			print_indent();
			fprintf(file_c,
				"case %d: { field_value = tmp_agent->%s; break; }\n", 
				i-cell_field_count, field_list[i].name);
		}

		undent;
		print_indent();
		fprintf(file_c, "}\n");

		print_indent();
		gen_set_cell_prefix("dim_1");
		fprintf(file_c, "(%s) field_value);\n", cast_type());

		undent;
		print_indent();
		fprintf(file_c, "}");
	}
	fprintf(file_c, "\n");

	/* Close off for loops that range over dimensional ranges. */
	undent;
	print_indent();
	fprintf(file_c, "}\n");

	undent;
	print_indent();
	fprintf(file_c, "}\n");

	if (!field_given) undent;
}

/* This is used to count the number of constant array value reader functions
   that have been created.  Each function is labeled with this number as
   part of the name to make their generation easier.
*/
int num_of_array_read_funs = 0;

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Writes code to end the grouping of cellular automata rules.
*/
void end_rules(void) {
	int i;

	if (agent_declared) gen_free_agents();

	if (cell_field_count > 0) gen_show_cell();

	/* Close off for loops that range over dimensional ranges. */
	for (i = dimension_count; i > 0; i--) {
		print_indent();
		fprintf(file_c, "if (dim_%d == dim_%d_ub) break;\n", i, i);
		print_indent();
		fprintf(file_c, "dim_%d += inc;\n", i);
		undent;
		print_indent();
		fprintf(file_c, "}\n");
	}

	if (agent_declared) {
		/* Must synchronize BEFORE combining the boundary! */
		gen_barrier();

		gen_combine_boundary();
#if MAX_CPUS > 1
		if (cpus > 1) {
			/* Must synchronize BEFORE combining the gaps! */
			gen_barrier();

			gen_combine_gap();
		}
#endif
	}

	if (agent_declared) gen_show_agent();

	gen_barrier();

	gen_swap_now_next();

#if MAX_CPUS > 1
	if (cpus > 1) {
		/* Generate code to exchange the values for each of the 
		   partitioned *_lb *_ub pairs.
		*/
		print_indent();
		fprintf(file_c, "{\n");
		indent;
		print_indent();
		fprintf(file_c, "register unsigned int tmp;\n");
		print_indent();
		fprintf(file_c, "tmp = dim_1_lb;\n");
		print_indent();
		fprintf(file_c, "dim_1_lb = dim_1_ub;\n");
		print_indent();
		fprintf(file_c, "dim_1_ub = tmp;\n");
		undent;
		print_indent();
		fprintf(file_c, "}\n");

		print_indent();
		fprintf(file_c, "if (id != CPUS-1) continue;\n");
	}
#endif

	/* Negate/reverse the value of inc. */
	print_indent();
	fprintf(file_c, "inc = -inc;\n");

	if (filter) {
		print_indent();
		fprintf(file_c,
			"if (filter && do_filter()) exit_signaled(0);\n");
	}

	/* Update the time variables which determine when cell values
	   should be reported.
	*/
	print_indent();
	fprintf(file_c, "if (!view) {\n");
	indent;

	print_indent();
	fprintf(file_c, "if (time_file != NULL && _time_step == next_read_time)\n");
	indent;
	print_indent();
	fprintf(file_c, "get_report_time(time_file);\n");
	undent;
	print_indent();
	fprintf(file_c, "else if (report_time_step > 0 && _time_step >= next_report_time)\n");
	indent;
	print_indent();
	fprintf(file_c, "next_report_time = _time_step + report_time_step;\n");
	undent;

	/* Go to the next time step. */
	print_indent();
	fprintf(file_c, "_time_step++;\n");

	undent;
	print_indent();
	fprintf(file_c, "}\n");

	/* Generate code to exchange the values for each of the non-partitioned
	   *_lb *_ub pairs.
	*/
	print_indent();
	fprintf(file_c, "{\n");
	indent;
	print_indent();
	fprintf(file_c, "register unsigned int tmp;\n");
	for (i = 1; i <= dimension_count; i++) {
#if MAX_CPUS > 1
		if (cpus > 1 && i == 1) continue;
#endif
		print_indent();
		fprintf(file_c, "tmp = dim_%d_lb;\n", i);
		print_indent();
		fprintf(file_c, "dim_%d_lb = dim_%d_ub;\n", i, i);
		print_indent();
		fprintf(file_c, "dim_%d_ub = tmp;\n", i);
	}
	undent;
	print_indent();
	fprintf(file_c, "}\n");

	gen_end_time_loop();

	/* Close off the "main" body. */
	print_indent();
	fprintf(file_c, "return 0;\n");
	undent;
	print_indent();
	fprintf(file_c, "}\n");

	gen_define_maxoffset();

	/* Generate the function that calls the constant array reader
	   functions.
	*/
	fprintf(file_hh, "\nvoid const_array_readers(void) {\n");
	for (i = 1; i <= num_of_array_read_funs; i++)
		fprintf(file_hh, "   const_array_reader_%d();\n", i);
	fprintf(file_hh, "}\n");

	/* Generate the max_value_width declaration for allowing the
	   neighborhood window cell size to adjust to the range of
	   field values of the field being viewed.
	*/
	fprintf(file_c, "\nint max_value_width[] = {");
	for (i = 0; i < field_count; i++) {
		unsigned int max_width;
		char string[MAX_STRING_LENGTH];

		if (use_floats)
			sprintf(string, "%f",
				field_list[i].range.lb.float_lower_bound);
		else
			sprintf(string, "%d",
				field_list[i].range.lb.lower_bound);
		max_width = strlen(string);

		if (use_floats)
			sprintf(string, "%f",
				field_list[i].range.ub.float_upper_bound);
		else
			sprintf(string, "%d",
				field_list[i].range.ub.upper_bound);
		if (max_width < strlen(string))
			max_width = strlen(string);

		fprintf(file_c, "%u", max_width);

		if (i < field_count-1) fprintf(file_c, ", ");
	}
	fprintf(file_c, "};\n");

	/* Generate the declarations and values for doing cell field 
	   examinations.
	*/
	fprintf(file_c, "\n");
	if (cell_fields_declared) {
		fprintf(file_c, "char field_names[%d] = \"",
			cell_field_count * longest_field_name + 1);
		for (i = 0; i < cell_field_count; i++) {
			int j = longest_field_name - strlen(field_list[i].name);
			for (; j > 0 ; j--) fprintf(file_c, " "); 
			fprintf(file_c, "%s", field_list[i].name);
		}
	} else
		fprintf(file_c, "char field_names[] = \"NONE");
	fprintf(file_c, "\";\n");

	fprintf(file_c, "int field_name_length = %d;\n", longest_field_name);
	fprintf(file_c, "int field_count = %d;\n\n", cell_field_count);

	/* Generate code for get_field. */
	fprintf(file_c, "%s get_field (int field", cast_type());
	for (i = 1; i <= dimension_count; i++)
		fprintf(file_c, ", int dim_%d", i);
	fprintf(file_c, ") {\n");
	indent;

	/* These extra things are needed when using microvectors. */
	if (mv_bits > 0) {
		print_indent();
		fprintf(file_c, "long int word;\n", i);
		print_indent();
		fprintf(file_c, "register int which = dim_1/NUM_OF_WORDS;\n");
		print_indent();
		fprintf(file_c, "dim_1 = dim_1%%NUM_OF_WORDS;\n");
	}

	print_indent();
	if (cell_fields_declared) {
		fprintf(file_c, "switch (field) {\n");
		indent;

		for (i = 0; i < cell_field_count; i++) {
			print_indent();
			fprintf(file_c, "case %d: ", i);

			if (mv_bits > 0)
				fprintf(file_c, "word = ");
			else
				fprintf(file_c, "return ");

			if (field_list[i].constant)
				gen_cell_field_name(field_list[i],
						    "dim_1+MAX_OFFSET");
			else
#if MAX_CPUS > 1
			if (cpus > 1)
				gen_cell_field_name(field_list[i],
						    "((inc>0?(MAX_OFFSET<<1)+1:MAX_OFFSET)+dim_1+(2*MAX_OFFSET+1)*(dim_1/part_size))");
			else
#endif
			gen_cell_field_name(field_list[i],
					    "(inc>0?(MAX_OFFSET<<1)+1:MAX_OFFSET)+dim_1");
			fprintf(file_c, ";");
			if (mv_bits > 0) fprintf(file_c, " break;");
			fprintf(file_c, "\n");
		}
		print_indent();
		fprintf(file_c, "default:\n");
		indent;
		print_indent();
		fprintf(file_c, "fprintf(stderr, \"%%s: Field %%d doesn't exist\\n\", command, field);\n");
		print_indent();
		fprintf(file_c, "exit_signaled(1);\n");
		undent;

		undent;
		print_indent();
		fprintf(file_c, "}\n");
	} else if (cell_field_count == 1) {
		if (mv_bits > 0) 
			fprintf(file_c, "word = cells");
		else
			fprintf(file_c, "return cells");

#if MAX_CPUS > 1
		if (cpus > 1)
			fprintf(file_c,
				"[(inc>0?(MAX_OFFSET<<1)+1:MAX_OFFSET)+dim_1+(2*MAX_OFFSET+1)*(dim_1/part_size)]");
		else
#endif
		fprintf(file_c, "[(inc>0?(MAX_OFFSET<<1)+1:MAX_OFFSET)+dim_1]");
		for (i = 2; i <= dimension_count; i++)
			fprintf(file_c, "[dim_%d+MAX_OFFSET]", i);
		fprintf(file_c, ";\n");
	} else
		fprintf(file_c, "return 0;\n");

	/* If microvectors are being used, then the value to be
	   returned must be stripped out of the word.
	*/
	if (mv_bits > 0 && cell_field_count > 0) {
		print_indent();
		fprintf(file_c,
			"return (long int) %s(word >> ((CELLS_PER_WORD-which-1)*BITS_PER_CELL), %d);\n", 
			(mv_signed ? "UVO_FSVAL" : "UV_FVAL"), mv_bits);
	}

	undent;
	fprintf(file_c, "}\n\n");

	/* Generate an "empty" function body for the version NOT being used. */
	if (use_floats)
		fprintf(file_c, "long int cell_field_value");
	else
		fprintf(file_c, "float cell_field_float_value");
	fprintf(file_c,
		" (int field, int dim_1, int dim_2, int dim_3) {\n\treturn 0;\n}\n\n");

	/* Generate the REAL version of this function. */
	fprintf(file_c, "%s ", cast_type());
	if (use_floats)
		fprintf(file_c, "cell_field_float_value");
	else
		fprintf(file_c, "cell_field_value");
	fprintf(file_c,
		" (int field, int dim_1, int dim_2, int dim_3) {\n");
	indent;

	print_indent();
	fprintf(file_c, "return get_field(field, dim_1");
	if (dimension_count > 1) fprintf(file_c, ", dim_2");
	if (dimension_count > 2) fprintf(file_c, ", dim_3");
	for (i = 4; i <= dimension_count; i++)
		fprintf(file_c, ", 0");
	fprintf(file_c, ");\n");

	undent;
	print_indent();
	fprintf(file_c, "}\n\n");

	/* Generate code for the functions to be used by filters. */
	if (filter) {
		print_indent();
		fprintf(file_c,
			"unsigned long int get_time(void) { return _time_step; }\n\n");

		print_indent();
		fprintf(file_c, "int num_dims(void) { return %d; }\n\n",
			dimension_count);

		print_indent();
		fprintf(file_c, "int dim_size(int dim) {\n");

		indent;
		print_indent();
		fprintf(file_c, "switch (dim) {\n");
		indent;

		for (i = 0; i < dimension_count; i++) {
			print_indent();
			fprintf(file_c, "case %d: return %d;\n",
				i, dim_size[i]);
		}

		undent;
		print_indent();
		fprintf(file_c, "}\n");
		undent;
		print_indent();
		fprintf(file_c, "}\n");
	}
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr, StackToken
      LEAVES  - n/a
      ACTION  - Places the necessary declaration in the symbol table
		(including the value).  The StackToken must not have
		been previously declared as either a constant or variable.
*/
void declare_simple_const(void) {
	attr_rec attribute;
	char *const_name;
	int i;

	/* If there is an error, pop the stack and return. */
	if (stack_error(2)) {
		stack_clear(2);
		top -= 2;
		return;
	}

	attribute.assignable = false;
	attribute.field_name = CONSTANT;
	attribute.structured = false;
	attribute.agent = false;
	attribute.array = false;
	attribute.index = false;
	attribute.cell_ref = false;

	/* The "address" is used to store the actual value which should
	   be substituted for the constant.
	*/
	attribute.address = stack[top--].data.expr.addr;

	/* If the value contains a "." then it must be a float value. */
	attribute.is_float = (int) strchr(attribute.address, '.');

	const_name = stack[top--].data.token;

	/* Place the identifier into the symbol table. */
	if (intable(const_name, CONSTANT))
		error("Redeclaration of named constant, '%s'", const_name);
	else if (intable(const_name, VARIABLE))
		error("Redeclaration of variable, '%s'", const_name);
	else
		enterid(const_name, attribute);

	myfree(const_name);
}

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr, StackToken
      LEAVES  - StackToken
      ACTION  - Declares a C array variable in the .hh file (so that it
		is shared in the multi-processor case).  The declaration
		isn't terminated however, since a value might still be
		given.  The StackToken must not have been previously
		declared as either a constant or variable and the StackExpr
		must be a constant with a value greater than zero (0).
*/
void declare_array_const(void) {
	attr_rec attribute;
	char *token;

	/* If there is an error, pop the stack and return. */
	if (stack_error(2)) {
		stack_clear(2);
		top--;
		stack[top].kind = StackError;
		return;
	}

	attribute.assignable = false;
	attribute.is_float = use_floats;
	attribute.field_name = CONSTANT;
	attribute.structured = false;
	attribute.agent = false;
	attribute.array = true;

	attribute.size = atoi(stack[top].data.expr.addr);
	myfree(stack[top--].data.expr.addr);

	attribute.index = false;
	attribute.cell_ref = false;

	token = stack[top].data.token;
	attribute.address = mymalloc(strlen(token)+2);
	sprintf(attribute.address, "_%s", token);

	if (attribute.size <= 0) {
		char msg[MAX_STRING_LENGTH];
		sprintf(msg, "Constant array, '%s', declared with %d elements",
			token, attribute.size);
		error(msg, (char*) NULL);
	}

	/* Place the identifier into the symbol table. */
	if (intable(token, CONSTANT))
		error("Redeclaration of named constant, '%s'", token);
	else if (intable(token, VARIABLE))
		error("Redeclaration of variable, '%s'", token);
	else
		enterid(token, attribute);

	/* Write the array declaration to the .hh file. */
	fprintf(file_hh, "\n%s %s[%d]",
		cast_type(), attribute.address, attribute.size);
}

/* SEMANTIC ROUTINE
      EXPECTS - StackToken(1), StackToken(2)
      LEAVES  - n/a
      ACTION  - StackToken(1) gives the name of a file from which to
		read the values of the constant array at startup time.
		StackToken(2) gives the name of the array into which
		the read values must be stored.  The C declaration of
		the array must be finished, and then the function to
		read the value declared.  The array read functions are
		numbered so as to make it easy to generate calls to them.
*/
void read_from_file(void) {

	/* If there is an error, pop the stack and return. */
	if (stack_error(2)) {
		stack_clear(2);
		top -= 2;
		return;
	}

	/* Finish the array declaration. */
	fprintf(file_hh, ";\n\n");

	/* Write the array value reader function. */
	fprintf(file_hh, "void const_array_reader_%d(void) {\n",
		++num_of_array_read_funs);

	fprintf(file_hh, "   int i;\n");
	fprintf(file_hh, "   FILE *array_file = fopen(\"%s\", \"r\");\n",
		stack[top].data.token);

	fprintf(file_hh, "   if (array_file == NULL) {\n");
	fprintf(file_hh, "         fprintf(stderr, \"Error opening file '%s'\\n\");\n",
		stack[top].data.token);
	fprintf(file_hh, "         exit_signaled(1);\n");
	fprintf(file_hh, "   }\n");

	fprintf(file_hh, "   for (i = 0; i < %d; i++)\n",
		get_attribute(stack[top-1].data.token, CONSTANT)->size);

	fprintf(file_hh,
		"      if (1 != fscanf(array_file, \" %s\", &_%s[i])) {\n",
		format_type(),
		stack[top-1].data.token);
	fprintf(file_hh, "         fprintf(stderr, \"Error reading values for '%s' from '%s'\\n\");\n",
		stack[top-1].data.token, stack[top].data.token);
	fprintf(file_hh, "         exit_signaled(1);\n");
	fprintf(file_hh, "      }\n");

	fprintf(file_hh, "   fclose(array_file);\n");

	fprintf(file_hh, "}\n\n");

	myfree(stack[top--].data.token);
	myfree(stack[top--].data.token);
}

/* SEMANTIC ROUTINE
      EXPECTS - StackExpr, ..., StackExpr, StackMark, StackToken
      LEAVES  - n/a
      ACTION  - Finishes the declaration of the constant array (named by
		StackToken) with the assignment of the StackExpr values.
		The number of StackExpr must match the size of the
		const array being declared.
*/
void array_value(void) {
	int mark;
	char *token;
	attr_rec *attr;

	fprintf(file_hh, " = {");

	/* Go down to the mark. */
        mark = top;
	while (stack[mark].kind != StackMark) mark--;

	token = stack[mark-1].data.token;
	attr = get_attribute(token, CONSTANT);

	/* The parser should force the token to represent a constant array,
	   but it's not a bad idea to check anyway.
	*/
	if (!attr->array) {
		fprintf(stderr, "INTERNAL ERROR: token in array_value not an array.\n");
		exit(1);
	}

	/* The number of expressions must be the same as the size of the
	   constant array.
	*/
	if (top-mark < attr->size)
		error("Too few values given for const array '%s'", token);
	else if (top-mark > attr->size)
		error("Too many values given for const array '%s'", token);

	/* Work back to the top of the stack, listing each value in turn. */ 
	while (++mark <= top) {
		char *token = stack[mark].data.token;

		fprintf(file_hh, "%s", token);

		if (mark < top) fprintf(file_hh, ", ");
	}

	fprintf(file_hh, "};\n");

	/* Clean stuff off the stack. */
	while (stack[top].kind != StackMark)
		myfree(stack[top--].data.token);
	top -= 2;
}

/************************************************************************/

/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Places the forward declarations for predefined symbols 
                into the symbol table.  This semantic routine is called 
		from main as opposed to being called from the grammar.
*/
void start_up(void) {
	char file_name[MAX_STRING_LENGTH];

	/* Open the C files. */
	sprintf(file_name, "%s%s%d.hh", TMP_DIR, BASE_FILE_NAME, pid);
	file_hh = fopen(file_name, "w");
	sprintf(file_name, "%s%s%d.h", TMP_DIR, BASE_FILE_NAME, pid);
	file_h = fopen(file_name, "w");
	sprintf(file_name, "%s%s%d.c", TMP_DIR, BASE_FILE_NAME, pid);
	file_c = fopen(file_name, "w");

	/* Initialize the symbol table. */
	initsymboltable();

	/* Generate code to read in the standard include file(s). */
	fprintf(file_c, "#include<stdio.h>\n");
	fprintf(file_c, "#include<stdlib.h>\n");
	fprintf(file_c, "#include<string.h>\n");
	fprintf(file_c, "#include<signal.h>\n\n");
	if (mv_bits > 0) {
		fprintf(file_c, "#include \"%suvector.h\"\n", LIB_DIR);
		fprintf(file_c,
			"#define UVO_GTE(x,y,bits) (UVO_NOT(UVO_LT(x,y,bits),bits))\n");
		fprintf(file_c,
			"#define UVO_LTE(x,y,bits) (UVO_NOT(UVO_GT(x,y,bits),bits))\n");
		fprintf(file_c,
			"#define UV_NEQUAL(x,y,bits) (UV_NOT(UV_EQUAL(x,y,bits),bits))\n");
		fprintf(file_c,
			"#define UVO_NEQUAL(x,y,bits) (UVO_NOT(UVO_EQUAL(x,y,bits),bits))\n");
		fprintf(file_c,
			"#define UVO_MOD(x,y,bits) (UVO_MINUS(x,UVO_MUL(UVO_DIV(x,y,bits),y,bits),bits))\n");
		fprintf(file_c,
			"#define UVO_FSVAL(x,bits) ((signed)(UVO_FSIGN(x,bits)?(-(UVO_FVAL(UVO_COMPL(x,bits),bits)+1)):UVO_FVAL(x,bits)))\n\n");
	}

#if MAX_CPUS > 1
	if (cpus > 1) {
#if SUN
		fprintf(file_c, "#include<thread.h>\n\n");
#elif SGI
		fprintf(file_c, "#include<ulocks.h>\n");
		fprintf(file_c, "#include<task.h>\n\n");
#endif
	}
#endif

	/* Generate the external references to use a cellviewer. */
	fprintf(file_c, "extern void exit_signaled(int);\n");
	fprintf(file_c, "extern void set");
	if (use_floats) fprintf(file_c, "_float");
	fprintf(file_c, "_cell_entry(long int, long int, long int, %s);\n",
		cast_type());

	/* Generate the external references to use a filter. */
	if (filter) fprintf(file_c, "extern int do_filter(void);\n");

	/* The declarations for the following variables are given outside
	   the body of functions so that when the software is made in
	   combination with the viewer, these variables are visible and/or
	   get initialized only once.
	*/

	/* command should be extern since it will be set by the viewer
	   portion of the software.
	*/
	fprintf(file_c, "extern char *command;\n\n");

	/* Write code to initialize the variable used for reading
	   the input to the cellular program.
	*/
	fprintf(file_c, "static long int next_time;\n");

	/* _time_step should NOT be static, since it will be used by the
	   viewer portion of the software.
	*/
	fprintf(file_c, "unsigned long int _time_step = 0;\n");

	/* Declare the time variables that determine when reports of 
	   cell values should be made. 
	*/
	fprintf(file_c, "unsigned long int report_time_step = 1;\n");
	fprintf(file_c, "unsigned long int next_report_time = 0;\n");
	fprintf(file_c, "long int next_read_time = 0;\n");

	/* time_file should NOT be static, since it will be used by the
	   viewer portion of the software.
	*/
	fprintf(file_c, "FILE *time_file = NULL;\n\n");

	/* use_floats should NOT be static, since it will be used by the
	   viewer portion of the software.
	*/
	fprintf(file_c, "int use_floats = %d;\n\n", use_floats);

	fprintf(file_c, "#define CPUS %d\n\n", cpus);

	/* This macro is used for range checks to make sure that array
	   indices are within bounds.  NOTE:  This macro reuses the
	   read_value variable.
	*/
	if (rng_chk)
		fprintf(file_c,
			"#define RNG_CHK(X,UB) ((((read_value=(X))<0)||(read_value>UB))?exit_signaled(1),0:(int)read_value)\n\n");
	else
		fprintf(file_c, "#define RNG_CHK(X,UB) X\n\n");
}


/* SEMANTIC ROUTINE
      EXPECTS - n/a
      LEAVES  - n/a
      ACTION  - Do some clean up actions.
*/
void finish_up(void) {
	/* Check to make sure that the semantic stack is empty. */
	if (top != -1)
		error("COMPILER ERROR -- %d entry(ies) left on the semantic stack",
		      (char*) top+1);

	/* Check to make sure that the mv_guard list is empty if microvectors
	   were used.
	*/
	if (mv_bits > 0 && mv_guards)
		error("COMPILER ERROR -- entry(ies) were left on the microvector guard list",
		      (char*) NULL);

	/* Check to make sure that the array_nesting list is empty. */
	if (array_nesting)
		error("COMPILER ERROR -- entry(ies) were left on the array nesting list",
		      (char*) NULL);

	/* Flush and close the output files. */
	fflush(file_hh);
	fclose(file_hh);
	fflush(file_h);
	fclose(file_h);
	fflush(file_c);
	fclose(file_c);
}
